/////////////////////////////////////////////////////////////////////////////
// Name:        Configure.cpp
// Purpose:    Configuration
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3
/////////////////////////////////////////////////////////////////////////////
// For compilers that support precompilation, includes "wx/wx.h".  
#include "wx/wxprec.h" 

#ifndef WX_PRECOMP
    #include "wx/log.h"
    #include "wx/wx.h"
    #include "wx/app.h"
    #include "wx/frame.h" 
	#include "wx/config.h"
	#include "wx/dirctrl.h"
	#include "wx/xrc/xmlres.h"
	#include "wx/spinctrl.h"
	#include "wx/tooltip.h"
	#include "wx/tokenzr.h"
	#include "wx/statline.h"
	#include "wx/colordlg.h"
	#include "wx/fontdlg.h"
#endif

#include "Externs.h"
#include "Tools.h"
#include "MyGenericDirCtrl.h"
#include "MyDirs.h"
#include "MyFiles.h"
#include "MyFrame.h"
#include "Misc.h"
#include "Accelerators.h"
#include "Filetypes.h"
#include "Configure.h"

wxArrayString PossibleTerminals, PossibleTerminalCommands, PossibleToolsTerminalCommands;  // Global, used in ConfigureTerminals


int SHCUT_OldIniBrowse = SHCUT_START - 1;	// Wizard button ids. It's pretty 'local' here so it should be safe
int SHCUT_ResourcesBrowse = SHCUT_START - 2;
int SHCUT_ResourcesAbort = SHCUT_START - 3;
int SHCUT_Configure = SHCUT_START - 4;


FirstPage::FirstPage(wxWizard *parent, wxWizardPage* Page2, wxWizardPage* Page3)    :  MyWizardPage(parent), m_Page2(Page2), m_Page3(Page3)
{
wxBoxSizer* mainsizer = new wxBoxSizer( wxVERTICAL );
wxBoxSizer* sizer1 = new wxBoxSizer( wxVERTICAL );
wxStaticText* welcome = new wxStaticText( this, wxID_ANY,  _("Welcome to 4Pane.") );
wxFont font = welcome->GetFont(); font.SetPointSize(30); welcome->SetFont( font );
int x, y; welcome->GetTextExtent(_("Welcome to 4Pane."), &x,&y);
wxSize size( x+10,y ); 
#if wxVERSION_NUMBER > 2502
welcome->SetMinSize( size );
#else
welcome->SetSize( size );
#endif
sizer1->Add( welcome, 0, wxTOP, 80 );
	wxBoxSizer* sizer2 = new wxBoxSizer( wxVERTICAL );
		wxBoxSizer* sizer2_1 = new wxBoxSizer( wxVERTICAL );
		wxStaticText* FewMoments = new wxStaticText( this, wxID_ANY,  _("4Pane is a File Manager that aims to be fast and full-featured without bloat.") );
		wxFont font1 = FewMoments->GetFont(); font1.SetPointSize(font1.GetPointSize() + 2); FewMoments->SetFont( font1 );
		sizer2_1->Add( FewMoments, 0, wxBOTTOM, 40 );
		wxStaticText* FourPaneNeeds = new wxStaticText( this, wxID_ANY,  _("Please click Next to configure 4Pane for your system.") ); FourPaneNeeds->SetFont( font1 );
		sizer2_1->Add( FourPaneNeeds, 0 );
			wxBoxSizer* sizer2_1_1 = new wxBoxSizer( wxHORIZONTAL );
			wxStaticText* CopyAConfigFile = new wxStaticText( this, wxID_ANY,  _("Or if there's already a configuration file that you'd like to copy, click") ); CopyAConfigFile->SetFont( font1 );
			sizer2_1_1->Add( CopyAConfigFile, 0, wxALIGN_CENTRE );
			wxButton* OldIniBrowse = new wxButton( this, SHCUT_OldIniBrowse,  _("Here") ); OldIniBrowse->SetFont( font1 );
			sizer2_1_1->Add( OldIniBrowse, 0, wxLEFT | wxALIGN_CENTRE, 5 );
			sizer2_1->Add( sizer2_1_1, 0, wxTOP, 40 );
	sizer2->Add( sizer2_1, 1, wxTOP | wxBOTTOM, 25 );
mainsizer->Add( sizer1, 0, wxTOP | wxBOTTOM, 40 );
mainsizer->Add( sizer2, 0, wxEXPAND | wxALIGN_CENTRE );
this->SetSizer( mainsizer );

RCFound = Configure::LocateResourcesSilently();						// Have a (slightly tentative) go at finding resources. If it succeeds, skip page 2
}

SecondPage::SecondPage(wxWizard *dad,  wxWizardPage* Page3)    :  MyWizardPage(dad), parent(dad), m_Page3(Page3)
{
wxBoxSizer* m2ainsizer = new wxBoxSizer( wxVERTICAL );
wxBoxSizer* s2izer1 = new wxBoxSizer( wxVERTICAL );
wxStaticText* resourcestext = new wxStaticText( this, wxID_ANY,  _("Resources successfully located. Please press Next") );
wxFont font = resourcestext->GetFont(); font.SetPointSize(font.GetPointSize() + 2); resourcestext->SetFont( font );
int x, y; resourcestext->GetTextExtent( resourcestext->GetLabel(), &x,&y );
wxSize size( x+10,y ); 
#if wxVERSION_NUMBER > 2502
resourcestext->SetMinSize( size );
#else
resourcestext->SetSize( size );
#endif
s2izer1->Add( resourcestext, 0, wxTOP | wxBOTTOM | wxALIGN_CENTRE, 60 );

m2ainsizer->Add( s2izer1, 0, wxEXPAND | wxALL | wxALIGN_CENTRE, 50 );
this->SetSizer( m2ainsizer );
}

void SecondPage::OnPageDisplayed()
{
if ( ! Configure::LocateResources() )  return parent->EndModal( wxID_CANCEL);
}

ThirdPage::ThirdPage(NoConfigWizard *parent, const wxString& configFP)    :  MyWizardPage(parent), configFPath(configFP)
{
wxBoxSizer* m3ainsizer = new wxBoxSizer( wxVERTICAL );
wxBoxSizer* s3izer1 = new wxBoxSizer( wxVERTICAL );
wxStaticText* hadalooktext = new wxStaticText( this, wxID_ANY,  _("I've had a look around your system, and created a configuration that should work.") );
wxFont font = hadalooktext->GetFont(); font.SetPointSize(font.GetPointSize() + 2); hadalooktext->SetFont( font );
s3izer1->Add( hadalooktext, 0, wxTOP | wxBOTTOM | wxALIGN_CENTRE, 30 );
wxStaticText* fullerconfigtext = new wxStaticText( this, wxID_ANY,  _("You can do much fuller configuration at any time from Options > Configure 4Pane.") ); fullerconfigtext->SetFont( font );
s3izer1->Add(  fullerconfigtext, 0, wxBOTTOM | wxALIGN_CENTRE, 30 );

bool DoShortcut = true;								// Lastly, ensure there's a 4Pane shortcut if the user wants one (not all distros' binaries manage to add one)
wxString Desktop( wxGetHomeDir() + wxT("/Desktop") ); FileData shortcut( Desktop + wxT("/4Pane.desktop") );
if ( shortcut.IsValid() ) DoShortcut = false;	// There's already a shortcut
  else
	{	FileData dtop( Desktop ); if ( ! dtop.CanTHISUserWrite() ) DoShortcut = false; }	// There isn't, but there isn't a ~/Desktop either, or else it's not writable by its owner?!
if ( DoShortcut == true )
	{	parent->shortcut = new wxCheckBox( this, wxID_ANY,  _("Put a 4Pane shortcut on the desktop") ); parent->shortcut->SetValue( true );
		 s3izer1->Add( parent->shortcut, 0, wxTOP, 120 );
		 parent->SetFocus();
	}

m3ainsizer->Add( s3izer1, 0, wxEXPAND | wxALL | wxALIGN_CENTRE, 50 );
this->SetSizer( m3ainsizer );
}

void ThirdPage::OnPageDisplayed()
{
delete wxConfigBase::Set((wxConfigBase *) NULL);		// Delete any old wxConfig
Configure::CopyConf( configFPath );								// Copy any master conf file into the new ini
wxConfig* localconfig = new wxConfig( wxGetApp().GetAppName(), wxT(""), configFPath );	//  & make a new, correct one
wxConfigBase::Set( localconfig );	// Needed to make it happen. There will already have been a ctor call, from the static wxConfigBase::Get()s (?), which might have created a null config

wxBusyCursor busy;						// Hardly worth bothering, but maybe on a very old, slow machine...
Configure::FindPrograms();
Configure::SaveDefaultsToIni();
Configure::DetectFixedDevices();	// Now there's an ini, add to it any detectable fixed drives
}
	
#include "bitmaps/include/wizardbitmap.xpm"

NoConfigWizard::NoConfigWizard(wxWindow* parent, int id, const wxString& title, const wxString& configFPath)	:	wxWizard( parent, id, title, wizardbitmap  )
{
shortcut = NULL;											// This might be the 3rd page ?desktop checkbox

ThirdPage* Page3 = new ThirdPage(this, configFPath);
SecondPage* Page2 = new SecondPage(this, Page3);
Page1 = new FirstPage(this, Page2, Page3 );

#if wxVERSION_NUMBER > 2500
	GetPageAreaSizer()->Add( Page1 );
#else
	FitToPage( Page1 );
#endif
}

void NoConfigWizard::OnButton(wxCommandEvent& event)
{
wxString filepath; bool flag = false;

FileDirDlg fdlg( this, FD_SHOWHIDDEN | FD_RETURN_FILESONLY, HOME, _("Browse for the Configuration file to copy") );
do
	{	if ( fdlg.ShowModal() != wxID_OK ) return;
		
		filepath = fdlg.GetFilepath();
		FileData fd( filepath, true );												// Use the dereferencing form of FileData: we want any symlink's target
		if ( ! fd.IsValid() || ! fd.IsRegularFile() ) continue;
		if ( ! fd.CanTHISUserPotentiallyCopy() )
			{	wxMessageDialog dialog(this, _("I'm afraid you don't have permission to Copy this file.\nDo you want to try again?"),
																																										_("No Exit!"), wxYES_NO | wxICON_QUESTION );
				if ( dialog.ShowModal() != wxID_YES )		return;
				continue;
			}
		
		flag = true;
	}
	while ( ! flag );		

wxLogNull NoFilesystemErrorsThanks;
wxCopyFile( filepath, HOME + wxT("/.") + wxGetApp().GetAppName(), true );	// Copy it. No need for subtlety: we've already checked it's a file, have permissions

delete wxConfigBase::Set((wxConfigBase *) NULL);		// Delete any old wxConfig
wxConfig* localconfig = new wxConfig( wxGetApp().GetAppName(), wxT(""), HOME + wxT("/.") + wxGetApp().GetAppName() );	//  & make a new, correct one
wxConfigBase::Set( localconfig );	// Needed to make it happen. There will already have been a ctor call, from the static wxConfigBase::Get()s (?), which might have created a null config

EndModal( wxID_OK);
}

BEGIN_EVENT_TABLE(NoConfigWizard, wxWizard)
	EVT_WIZARD_PAGE_CHANGED(wxID_ANY, NoConfigWizard::OnPageChanged)
	EVT_BUTTON(SHCUT_OldIniBrowse, NoConfigWizard::OnButton)
END_EVENT_TABLE()

//static
bool Configure::FindIni()	// If there's not a valid ini file in HOME (or provided), use a wizard to find/create one
{
wxString ConfigFilepath = ((MyApp&)wxGetApp()).ConfigFilepath;	// See if there's a user-provided filepath-to-config
if ( ConfigFilepath.IsEmpty() ) ConfigFilepath = HOME+wxT("/.") + wxGetApp().GetAppName(); // If not, look in HOME

while (true)
	{	FileData ini( ConfigFilepath );
		if ( ini.IsValid() )	
			{	wxConfigBase* config = wxConfigBase::Get(); 									// See if there's (now) a valid config
				if ( config != NULL )
					{	wxString test = config->Read( wxT("Misc/Identifier") );
						if ( test == wxT("Genuine config file") )	break;							// It's the real thing...
					}
			}

		if ( ! Wizard( ConfigFilepath ) ) return false;												// Otherwise run the wizard to create one. False means failure, user abort
		wxConfigBase::Get()->Write(wxT("Misc/ResourceDir"), RCDIR);				// Since there's a chance the user had to locate these, store the results
		wxConfigBase::Get()->Write(wxT("Misc/BitmapsDir"), BITMAPSDIR);
		wxConfigBase::Get()->Write(wxT("Misc/HelpDir"), HELPDIR);
	}

delete wxConfigBase::Set((wxConfigBase *) NULL);		// Delete any old wxConfig
wxConfig* localconfig = new wxConfig( wxGetApp().GetAppName(), wxT(""), ConfigFilepath );	//  & make a new, correct one
wxConfigBase::Set( localconfig );	// Needed to make it happen. There will already have been a ctor call, from the static wxConfigBase::Get()s (?), which might have created a null config
	
bool ans = Configure::LocateResources();														// We now need to locate /rc, /bitmaps, /docs
if ( ans )
	{	wxConfigBase::Get()->Write(wxT("Misc/ResourceDir"), RCDIR);				// Since there's a chance the user had to locate these, store the results
		wxConfigBase::Get()->Write(wxT("Misc/BitmapsDir"), BITMAPSDIR);
		wxConfigBase::Get()->Write(wxT("Misc/HelpDir"), HELPDIR);
	}
return ans;
}

//static
bool Configure::LocateResources()	// Where is everything?
{
wxString extra;

while ( ! FindRC( extra ) )
	{	wxString msg( wxT("Can't find 4Pane's resource files. There must be something wrong with your installation. :(\nDo you want to try to locate them manually?") );
		if ( wxMessageBox( msg, wxT("Eek! Can't find resources."), wxYES_NO | wxICON_EXCLAMATION ) != wxYES ) return false;		
							// If the correct dir wasn't found, see if the user can find it
		wxDialog pseudo( NULL, wxID_ANY, wxT("") );	// We need a valid wxWindow to parent the wxDirDialog. Shouldn't need to Show() it
		#if wxVERSION_NUMBER >2604 && defined(__WXGTK20__)
		int style = wxDD_DEFAULT_STYLE | wxDD_DIR_MUST_EXIST;				// the gtk2 dirdlg uses different flags :/
		#else
		int style = wxDEFAULT_DIALOG_STYLE;
		#endif
		wxDirDialog dlg( &pseudo,_("Browse for  ..../4Pane/rc/"), HOME, style );
		if ( dlg.ShowModal() != wxID_OK ) return false;
		extra = dlg.GetPath();										// This is the selected dir. Putting it into 'extra' means that the next iteration of FindRC will look there too
		if ( extra.Right(1) == wxFILE_SEP_PATH ) extra = extra.BeforeLast(wxFILE_SEP_PATH);
		if ( extra.Right(3) == wxT("/rc") ) extra = extra.BeforeLast(wxFILE_SEP_PATH);	// The idea is that we get the Path to /rc/, which should be correct for /bitmaps/ & /docs/ too
	}

while ( ! FindBitmaps( extra ) )
	{	wxString msg( wxT("Can't find 4Pane's bitmaps. There must be something wrong with your installation. :(\nDo you want to try to locate them manually?") );
		if ( wxMessageBox( msg, wxT("Eek! Can't find bitmaps."), wxYES_NO | wxICON_EXCLAMATION ) != wxYES ) return false;		

		wxDialog pseudo( NULL, wxID_ANY, wxT("") );
		#if wxVERSION_NUMBER >2604 && defined(__WXGTK20__)
		int style = wxDD_DEFAULT_STYLE | wxDD_DIR_MUST_EXIST;				// the gtk2 dirdlg uses different flags :/
		#else
		int style = wxDEFAULT_DIALOG_STYLE;
		#endif
		wxDirDialog dlg( &pseudo,_("Browse for ..../4Pane/bitmaps/"), HOME, style );
		if ( dlg.ShowModal() != wxID_OK ) return false;
		extra = dlg.GetPath();
		if ( extra.Right(1) == wxFILE_SEP_PATH ) extra = extra.BeforeLast(wxFILE_SEP_PATH);
		if ( extra.Right(8) == wxT("/bitmaps") ) extra = extra.BeforeLast(wxFILE_SEP_PATH);
	}

while ( ! FindHelp( extra ) )
	{	wxString msg( wxT("Can't find 4Pane's Help files. There must be something wrong with your installation. :(\nDo you want to try to locate them manually?") );
		if ( wxMessageBox( msg, wxT("Eek! Can't find Help files."), wxYES_NO | wxICON_EXCLAMATION ) != wxYES ) return true;	// NB we can just about cope without help, so return true this time

		wxDialog pseudo( NULL, wxID_ANY, wxT("") );
		#if wxVERSION_NUMBER >2604 && defined(__WXGTK20__)
		int style = wxDD_DEFAULT_STYLE | wxDD_DIR_MUST_EXIST;				// the gtk2 dirdlg uses different flags :/
		#else
		int style = wxDEFAULT_DIALOG_STYLE;
		#endif
		wxDirDialog dlg( &pseudo,_("Browse for ..../4Pane/doc/"), HOME, style );
		if ( dlg.ShowModal() != wxID_OK ) return true;
		extra = dlg.GetPath();
		if ( extra.Right(1) == wxFILE_SEP_PATH ) extra = extra.BeforeLast(wxFILE_SEP_PATH);
		if ( extra.Right(4) == wxT("/doc") ) extra = extra.BeforeLast(wxFILE_SEP_PATH);
	}

return true;
}

//static
bool Configure::LocateResourcesSilently()	// Where is everything? This is just a rough first check, so no error messages/try again
{
wxString extra;

if ( ! FindRC( extra ) ) return false;
if ( ! FindBitmaps( extra ) ) return false;
if ( ! FindHelp( extra ) ) return false;
return true;
}

//static
bool Configure::FindRC( wxString& extra )	// Find at least 1 valid source of XRC files
{
wxConfigBase::Get()->Read(wxT("Misc/ResourceDir"), &RCDIR);
if ( ! RCDIR.IsEmpty() )	{ FileData stat( RCDIR + wxT("/dialogs.xrc")); if ( stat.IsValid() )  return true; }	// Look in config to see if the user has put things elsewhere

FileData stat1( (RCDIR = extra+wxT("/rc/")) + wxT("dialogs.xrc") ); if ( stat1.IsValid() ) return true;								// If we were passed a user-located dir, try there
FileData stat2( (RCDIR = wxGetApp().StdDataDir+wxT("/rc/")) + wxT("dialogs.xrc") ); if ( stat2.IsValid() ) return true;	// If not, look in (probably) /usr/local/share/4Pane/rc
FileData stat3( (RCDIR = wxT("/usr/local/share/4Pane/rc/")) + wxT("dialogs.xrc") ); if ( stat3.IsValid() ) return true;		// Or hard-wire this (might be needed in a livecd situation)
FileData stat4( (RCDIR = wxT("/usr/share/4Pane/rc/")) + wxT("dialogs.xrc") ); if ( stat4.IsValid() ) return true;				// Ditto in debianish systems
FileData stat5( (RCDIR = wxGetCwd()+wxT("/rc/")) + wxT("dialogs.xrc") ); if ( stat5.IsValid() ) return true;					// If still not, look in ./rc
FileData stat6( (RCDIR = HOME+wxT("/rc/")) + wxT("dialogs.xrc") ); if ( stat6.IsValid() ) return true;								// 1 last chance. Try in HOME

return false;
}

//static
bool Configure::FindBitmaps( wxString& extra )
{
wxConfigBase::Get()->Read(wxT("Misc/BitmapsDir"), &BITMAPSDIR);
if ( ! BITMAPSDIR.IsEmpty() )	{ FileData stat( BITMAPSDIR + wxT("/back.xpm")); if ( stat.IsValid() )  return true; }	// Look in config to see if the user has put things elsewhere

FileData stat1( (BITMAPSDIR = extra+wxT("/bitmaps/")) + wxT("back.xpm") ); if ( stat1.IsValid() ) return true;								// If we were passed a user-located dir, try there
FileData stat2( (BITMAPSDIR = wxGetApp().StdDataDir+wxT("/bitmaps/")) + wxT("back.xpm") ); if ( stat2.IsValid() ) return true;	// If not, look in (probably) /usr/local/share/4Pane/bitmaps
FileData stat3( (BITMAPSDIR = wxT("/usr/local/share/4Pane/bitmaps/")) + wxT("back.xpm") ); if ( stat3.IsValid() ) return true;		// Or hard-wire this (might be needed in a livecd situation)
FileData stat4( (BITMAPSDIR = wxT("/usr/share/4Pane/bitmaps/")) + wxT("back.xpm") ); if ( stat4.IsValid() ) return true;				// Ditto in debianish systems
FileData stat5( (BITMAPSDIR = wxGetCwd()+wxT("/bitmaps/")) + wxT("back.xpm") ); if ( stat5.IsValid() ) return true;					// If still not, look in ./bitmaps
FileData stat6( (BITMAPSDIR = HOME+wxT("/bitmaps/")) + wxT("back.xpm") ); if ( stat6.IsValid() ) return true;							// 1 last chance. Try in HOME

return false;
}

//static
bool Configure::FindHelp( wxString& extra )
{
wxConfigBase::Get()->Read(wxT("Misc/HelpDir"), &HELPDIR);
if ( ! HELPDIR.IsEmpty() )	{ FileData stat( HELPDIR + wxT("/About.htm")); if ( stat.IsValid() )  return true; }	// Look in config to see if the user has put things elsewhere

FileData stat1( (HELPDIR = extra+wxT("/About.htm")) ); if ( stat1.IsValid() ) return true;									// If we were passed a user-located dir, try there
FileData stat2( (HELPDIR = extra+wxT("/doc/")) + wxT("About.htm") ); if ( stat2.IsValid() ) return true;				// or nearby
FileData stat3( (HELPDIR = wxGetApp().StdDocsDir) + wxT("/About.htm") ); if ( stat3.IsValid() ) return true;		// If not, look in (probably) /usr/local/share/doc/4Pane
FileData stat4( (HELPDIR = wxT("/usr/local/share/4Pane/doc/")) + wxT("About.htm") ); if ( stat4.IsValid() ) return true;// Or hard-wire this (might be needed in a livecd situation)
FileData stat5( (HELPDIR = wxT("/usr/share/doc/4pane/")) + wxT("About.htm") ); if ( stat5.IsValid() ) return true;			// Ditto in debianish systems
FileData stat6( (HELPDIR = wxT("/usr/share/doc/4Pane/")) + wxT("About.htm") ); if ( stat6.IsValid() ) return true;			//  or almost...
FileData stat7( (HELPDIR = wxT("/usr/doc/4Pane/")) + wxT("About.htm") ); if ( stat7.IsValid() ) return true;		// Ditto in slackware (Sigh! And even that's a symlink)
FileData stat8( (HELPDIR = wxGetCwd()+wxT("/doc/")) + wxT("About.htm") ); if ( stat8.IsValid() ) return true;	// If still not, look in ./docs
FileData stat9( (HELPDIR = HOME+wxT("/doc/")) + wxT("About.htm") ); if ( stat9.IsValid() ) return true;			// 1 last chance. Try in HOME

return false;
}

void Configure::FindPrograms()
{
wxString name( wxT("kdesu") ); KDESU_PRESENT = TestExistence( name );
name =  wxT("gnomesu"); XSU_PRESENT = TestExistence( name );
name = wxT("gksu"); GKSU_PRESENT = TestExistence( name );

name = wxT("konsole"); KONSOLE_PRESENT = TestExistence( name );
name = wxT("gnome-terminal"); GNOME_TERM_PRESENT = TestExistence( name );
name = wxT("xterm"); XTERM_PRESENT = TestExistence( name );

if ( TestExistence( wxT("/sbin/showmount") ) ) SHOWMOUNTFPATH = wxT("/sbin/showmount"); // This is normally /USR/sbin/showmount, but in debian...
}

//static
void Configure::DetectFixedDevices()
{
ArrayofDeviceStructs* DeviceStructArray = new ArrayofDeviceStructs;
DeviceManager::GetFloppyInfo( DeviceStructArray, 0, NULL, true );
DeviceManager::GetCdromInfo( DeviceStructArray, 0, NULL, true );

	// The rest is adapted from DeviceManager::SaveFixedDevices(). Unfortunately, esp. within a static function, it wasn't reusable
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL) return;
config->DeleteGroup( wxT("Devices/ToLoad") );																// Out with the old . . .

wxString Rootname( wxT("Devices/ToLoad/") );
size_t count = DeviceStructArray->GetCount();

for ( size_t n = 0; n < count; ++n )																// Go thru the list of cdrom drives etc
	{	DevicesStruct* tempstruct = DeviceStructArray->Item( n );
		if ( tempstruct == NULL ) continue;
		wxString subgrp = CreateSubgroupName( n, count );

		config->Write( Rootname+subgrp+wxT("/Device"), tempstruct->devicenode );
		config->Write( Rootname+subgrp+wxT("/Name"), tempstruct->name1 );
		wxString mtpt; if ( tempstruct->mountpts.GetCount() ) mtpt = tempstruct->mountpts[0];
		config->Write( Rootname+subgrp+wxT("/Mountpoint"), mtpt );			// Save any fstab mountpt
		config->Write( Rootname+subgrp+wxT("/Readonly"), tempstruct->readonly );
		config->Write( Rootname+subgrp+wxT("/Devicetype"), tempstruct->devicetype );
		config->Write( Rootname+subgrp+wxT("/Ignore"), tempstruct->ignore );	// If true, don't actually use this one
	}
	
config->Write( Rootname + wxT("NoOfDevices"), (long)count );				// How many to reload next time
config->Flush();

delete DeviceStructArray;
}

//static
enum distro Configure::DeduceDistro( double& Version )		// Try to work out which distro & version we're on
{
enum distro Distro = gok;

if ( wxFileExists( wxT("/etc/SuSE-release") ) )											// SuSE
	{	Distro = suse;
		wxTextFile tf( wxT("/etc/SuSE-release") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(1); 
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;	// This should put the first valid float into Version, or -1 if there isn't one
		return Distro;
	}

if ( wxFileExists( wxT("/etc/pclinuxos-release") ) )										// PCLinuxOS
	{	Distro = pclos;
		wxTextFile tf( wxT("/etc/pclinuxos-release") );									// Could also have used /etc/issue --- same contents:  PCLinuxOS release 2007
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0); 
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;			// This should put the first valid float into Version
		return Distro;
	}

wxString filename;																						// Temp store for filename, see below
if ( wxFileExists( filename=wxT("/etc/mandrakelinux-release") ) 				// Mandrake/Mandriva  (watch out: PCLinuxOS has symlinks with these names)
					||  wxFileExists( filename=wxT("/etc/mandriva-release") ) )
	{	Distro = mandrake;
		wxTextFile tf( filename );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0); 
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;
		return Distro;
	}

if ( wxFileExists( wxT("/etc/fedora-release") ) )											// Fedora
	{	Distro = fedora;
		wxTextFile tf( wxT("/etc/fedora-release") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0); 
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;
		return Distro;
	}

if ( wxFileExists( wxT("/etc/zenwalk-version") ) )											// Zenwalk. Must come before Slackware as it also has an /etc/slackware-version
	{	Distro = zenwalk;
		wxTextFile tf( wxT("/etc/zenwalk-version") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0);																// Should be e.g. Zenwalk 5.0
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;
		return Distro;
	}

if ( wxFileExists( wxT("/etc/slackware-version") ) )										// Slackware
	{	Distro = slackware;
		wxTextFile tf( wxT("/etc/slackware-version") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0); 
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;			// Using GetFloat() copes with foo bar 1.2.3, returning 1.2
		return Distro;
	}

if ( wxFileExists( wxT("/etc/sabayon-release") ) )										// Sabayon, which we must do before gentoo
	{	Distro = sabayon;
		wxTextFile tf( wxT("/etc/sabayon-release") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0);
		if ( ! GetFloat( line, &Version, 1 ) ) Version = wxNOT_FOUND;			// This should put the second valid float into Version (Sabayon Linux x86-64 3.4)
		return Distro;
	}

if ( wxFileExists( wxT("/etc/gentoo-release") ) )											// Gentoo
	{	Distro = gentoo;
		wxTextFile tf( wxT("/etc/gentoo-release") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0);
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;
		return Distro;
	}
	
if ( wxFileExists( wxT("/etc/puppyversion") ) )												// Puppy Linux
	{	Distro = puppy;
		wxTextFile tf( wxT("/etc/puppyversion") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0); 
		line.ToDouble( &Version ); return Distro;											// Puppy is easy: it's just 2.00 (or whatever)
	}


if ( wxFileExists( wxT("/etc/redhat-release") ) )											// Red Hat (watch out: PCLinuxOS has a symlink with this name
	{	Distro = redhat;
		wxTextFile tf( wxT("/etc/redhat-release") );
		if ( ! tf.Open() )  return Distro;
		wxString line = tf.GetLine(0);
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;
		return Distro;
	}

if ( wxFileExists( wxT("/etc/debian_version") ) )											// It's some type of Debian, but which?
	{	if ( wxFileExists( wxT("/etc/knoppix-version") ) )	
			{	Distro = knoppix;																			// Knoppix
				wxTextFile tf( wxT("/etc/knoppix-version") );
				if ( ! tf.Open() )  return Distro;
				wxString line = tf.GetLine(0); 
		if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;
				return Distro;
			}

			
		if ( wxFileExists( wxT("/etc/issue") ) )	
			{	Distro = mepis;																			// Mepis
				wxTextFile tf( wxT("/etc/issue") );
				if ( ! tf.Open() )  return Distro;
				wxString line = tf.GetLine(0); 
				if ( line.Contains(wxT("MEPIS")) )					// If it *doesn't* it's probably ubuntu proper. (In Mepis 3.1 MEPIS started the string; in 6.0 it's near the end)
					{	if ( wxFileExists( wxT("/etc/lsb-release") ) )							// In 6.0, it's come over all ubuntu-ish, so we have to use lsb-release
							{	wxTextFile tf( wxT("/etc/lsb-release") );
								if ( ! tf.Open() )  return Distro;
								wxString line = tf.GetLine(1); 
								if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;	// This should put the first valid float into Version
								return Distro;	// Though Mepis calls itself 6.0, its ubuntu version returns 6.06 :(
							}
						 else if ( wxFileExists( wxT("/etc/debian_version") ) )
							{	wxTextFile debf( wxT("/etc/debian_version") );			// Mepis <6 doesn't store a version no, so use the debian one if it exists (as in Mepis 3.1)
								if ( ! debf.Open() )  return Distro;
								line = debf.GetLine(0); 
								if ( line.ToDouble( &Version ) ) return Distro;
							}

						Version = -1; return Distro;	// (Backstop)
					}
			}

		if ( wxFileExists( wxT("/etc/lsb-release") ) )											// Ubuntu
			{	Distro = debian;		// Temporarily
				wxTextFile tf( wxT("/etc/lsb-release") );
				if ( ! tf.Open() )  return Distro;
				wxString line = tf.GetLine(0); 
				if ( line == wxT("DISTRIB_ID=Ubuntu") ) Distro = ubuntu;
				line = tf.GetLine(1);
				if ( ! GetFloat( line, &Version, 0 ) ) Version = wxNOT_FOUND;	// This should put the first valid float into Version
				return Distro;
			}

		else																										// Plain vanilla debian
			{	Distro = debian;
				wxTextFile tf( wxT("/etc/debian_version") );
				if ( ! tf.Open() )  return Distro;
				wxString line = tf.GetLine(0); 
				if ( line.ToDouble( &Version ) ) return Distro;
				Version = -1; return Distro;
			}
	}

Version = -1; return (Distro=other);															// If we're here, nothing of note was found
}

//static
void Configure::SaveProbableHAL_Sys_etcInfo( enum distro Distro, double Version )		// Using either the discovered distro, or stat + guesswork, save the probable discovery/loading setup
{
FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = true;		// Make the commonest answer the default as it saves a lot of typing
FLOPPY_DEVICES_SUPERMOUNT=CDROM_DEVICES_SUPERMOUNT=USB_DEVICES_SUPERMOUNT = false; // Ditto
DEVICE_DISCOVERY_TYPE = SysHal; 																														// Ditto

if ( Distro==suse )
   { if ( Version < 9.1 ) 	// I'm guessing about 9.0, but this is certainly true for 8.2) 
      {	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT= false;
        DEVICE_DISCOVERY_TYPE = OriginalMethod; SCSIUSBDIR = wxT("/proc/scsi/usb-storage-"); SCSISCSI =  wxT("/proc/scsi/scsi");		// (which they probably were anyway)
        FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert; USB_DEVICES_DISCOVERY_INTO_FSTAB = FstabInsert;
      }
     else if (Version > 9.0 && Version < 9.3 )		// SuSE's oddity: mounting to mtab without HAL/sys
      {	DEVICE_DISCOVERY_TYPE = MtabMethod;
        FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
      }
     else if ( Version >= 10.1 )									// SuSE 10.1-2 seems to automount only via kde, and only usb devices! HAL/sys discovery
      {	DEVICE_DISCOVERY_TYPE = SysHal; 
        FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert; USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
        FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=false;
      }
   }

if ( Distro==slackware )
	{	if ( Version < 11.1 )
			{	if ( KernelLaterThan( wxT("2.5") ) )		// (I doubt if anyone's using precisely 2.5.0)
					{	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = false; 		// True for 10.0 & 10.2, & 11.0 with the 2.6 kernel
						FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
					}
			 	else
					{	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT= false;		// True for 10.0 & 10.2, & 11.0 with the 2.4 kernel
						DEVICE_DISCOVERY_TYPE = OriginalMethod; SCSIUSBDIR = wxT("/proc/scsi/usb-storage-"); SCSISCSI =  wxT("/proc/scsi/scsi");		// (which they probably were anyway)
						FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
					}
			}
		 else if ( Version >= 12.0 )
					{	CD_DEVICES_DISCOVERY_INTO_FSTAB=FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;  USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
						FLOPPY_DEVICES_AUTOMOUNT= false;		// Floppies don't AUTOMOUNT
					}
	}

if ( Distro==sabayon )		// I've only started with 3.4, which does the newfangled, leave it to the desktop manager, thing; except for floppies which don't
	{	FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;  CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
		FLOPPY_DEVICES_AUTOMOUNT= false;		// Floppies don't AUTOMOUNT
	}

if ( Distro==gentoo )		// Can't really say with gentoo, as it's up to the individual how he configures. Make minimal guesses
	{	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = false;
		FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
		DEVICE_DISCOVERY_TYPE =  KernelLaterThan( wxT("2.5") ) ? SysHal : OriginalMethod; 
	}

if ( Distro==ubuntu )
  { if ( Version < 6 )
      {	FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = FstabInsert;
        FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT= false;		// Only USB_DEVICES_AUTOMOUNT
      }
    else if ( Version >= 6.06 )
      {	FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;  CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
        FLOPPY_DEVICES_AUTOMOUNT= false;		// Floppies don't AUTOMOUNT
      }
  }

if ( Distro==mepis )
	{ if ( Version < 6 )
      {	FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = FstabInsert;
        FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT= false;
      }
    else if ( Version == 6.06 )	// Mepis now returns a ubuntu version :(
      {	FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = CD_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;  USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
        FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT= false;		// Only USB_DEVICES_AUTOMOUNT
      } 
    else if ( Version >= 7.0 )		// Mepis no longer returns a ubuntu version!
      {	FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert; CD_DEVICES_DISCOVERY_INTO_FSTAB = FstabInsert; USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
        FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT= false;		// Both usbs and dvds can automount, but it's better not to for dvds as they -> fstab anyway
      } 
  }

if ( Distro==debian )
	{ if ( Version==3.1 )																			// Sarge
      {	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = false; 
        FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
      }
    else if ( Version > 3.1 )																	// Etch==4.0
      {	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = false;
        FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB = USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
      }
  }

if ( Distro==knoppix )
	{ if ( Version >= 5.3 )																		// Uses Etch
      {	FLOPPY_DEVICES_AUTOMOUNT = false; CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = true;
        FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
      }
  }

if ( Distro==mandrake && Version == 10.1 )		
	{	FLOPPY_DEVICES_SUPERMOUNT = true;
		FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
		FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT= false;			// Only USB_DEVICES_AUTOMOUNT, and then only sometimes!
	}
	
if ( Distro==mandrake && Version == 10.2 )		
	{	FLOPPY_DEVICES_SUPERMOUNT = true;
		FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT= false;	// Nothing properly automounts
		CD_DEVICES_DISCOVERY_INTO_FSTAB=FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert; USB_DEVICES_DISCOVERY_INTO_FSTAB=FstabInsert;
	}
	
if ( Distro==mandrake && Version == 2007)		
	{	FLOPPY_DEVICES_SUPERMOUNT = true;	
		CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = false;
		FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
	}
	
if ( Distro==mandrake && Version == 2008)	// Assuming we allow Konq to do it	
	{	FLOPPY_DEVICES_AUTOMOUNT = false;	
		FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
		CD_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
	}
	
if ( Distro==mandrake && Version >= 2008.1)	// There are now sabayon-like icons next to the clock
	{	FLOPPY_DEVICES_AUTOMOUNT = false;	
		FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=NoInsert;
		CD_DEVICES_DISCOVERY_INTO_FSTAB = USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
	}

if ( Distro==fedora )
	{	if ( Version < 4 )	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT= false;		// Only USB_DEVICES_AUTOMOUNT
		 else if ( Version < 5 )
			{	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT= false;
				FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert; USB_DEVICES_DISCOVERY_INTO_FSTAB = FstabInsert;			
			}
		 else if ( Version < 7 ) // This works for 5 & 6
			{	FLOPPY_DEVICES_AUTOMOUNT = false; FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert; 
				CD_DEVICES_DISCOVERY_INTO_FSTAB = USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
			}
		else //if ( Version >= 7 )	// It all depends on whether you want konq to do it for you
			{	FLOPPY_DEVICES_AUTOMOUNT = false; CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = true;
				FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert; CD_DEVICES_DISCOVERY_INTO_FSTAB = USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
			}
	}

if ( Distro==pclos )
	{	if ( Version >= 2007 )		// It all depends on whether you want konq to do it for you
			{	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = false;
				FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB = USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
			}
	}

if ( Distro==puppy /*&& Version < 9.1*/) 	// I've only looked at 2.0. No automount, & I can't get usb devices to be recognised, tho' I think they should be
	{	FLOPPY_DEVICES_AUTOMOUNT=CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT= false;
		DEVICE_DISCOVERY_TYPE = OriginalMethod; SCSIUSBDIR = wxT("/proc/scsi/usb-storage-"); SCSISCSI =  wxT("/proc/scsi/scsi");		// (which they probably were anyway)
		FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB=CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert;
	}

if ( Distro==zenwalk) 	// I've only looked at 5.0. Standard version has automounting cdroms/usbs, but not floppies. There's a 'core' version that doesn't, though
	{	FLOPPY_DEVICES_AUTOMOUNT = false; CDROM_DEVICES_AUTOMOUNT=USB_DEVICES_AUTOMOUNT = true;
		FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert; CD_DEVICES_DISCOVERY_INTO_FSTAB = USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
	}

	
if ( Distro==gok || Distro==other )
  { FLOPPY_DEVICES_AUTOMOUNT = false;	// Use what are the commonest values nowadays
    FLOPPY_DEVICES_DISCOVERY_INTO_FSTAB = NoInsert; CD_DEVICES_DISCOVERY_INTO_FSTAB=USB_DEVICES_DISCOVERY_INTO_FSTAB = MtabInsert;
    if ( ! KernelLaterThan( wxT("2.5") ) )		// (I doubt if anyone's using precisely 2.5.0)
      { DEVICE_DISCOVERY_TYPE = OriginalMethod; SCSIUSBDIR = wxT("/proc/scsi/usb-storage-"); SCSISCSI =  wxT("/proc/scsi/scsi"); }		// (which they probably were anyway)
  }

DeviceManager::SaveMiscData();																						// These effectively save what we've just done
DeviceBitmapButton::SaveDefaults();
}

//static
void Configure::SaveDefaultSus()		// Save a useful value for WHICH_SU
{
if ( WHICH_SU == dunno )
	{	if ( KDESU_PRESENT )  WHICH_SU = kdesu;
		 else if ( GKSU_PRESENT )  WHICH_SU = gksu;
		 else if ( XSU_PRESENT )  WHICH_SU = gnomesu;
	}
wxConfigBase::Get()->Write( wxT("/Misc/Which_Su"),  (long)WHICH_SU );
}

//static
void Configure::SaveDefaultTerminal( enum distro Distro, double Version )		// Save a useful value for TERMINAL etc
{
PossibleTerminals.Add( wxT("konsole") ); PossibleTerminals.Add( wxT("gnome-terminal" )); PossibleTerminals.Add( wxT("xterm") );
PossibleTerminalCommands.Add( wxT("konsole -e") ); PossibleTerminalCommands.Add( wxT("gnome-terminal -x") ); PossibleTerminalCommands.Add( wxT("xterm -e") );
PossibleToolsTerminalCommands.Add( wxT("konsole --noclose -e") ); PossibleToolsTerminalCommands.Add( wxT("xterm -hold -e") );

if ( KONSOLE_PRESENT || GNOME_TERM_PRESENT || XTERM_PRESENT )			// We need at least 1 to be available
	{	if ( KONSOLE_PRESENT ) 
			{ TERMINAL = wxT("konsole"); TERMINAL_COMMAND = wxT("konsole -e "); 
				if ( Distro==fedora && Version >= 9 )											// This comes with a konsole that doesn't --noclose !
								TOOLS_TERMINAL_COMMAND = wxT("xterm -hold -e ");
					else	TOOLS_TERMINAL_COMMAND = wxT("konsole --noclose -e "); 
			}
		 else if ( GNOME_TERM_PRESENT ) 
		 			{	TERMINAL = wxT("gnome-terminal"); TERMINAL_COMMAND = wxT("gnome-terminal -x ");
						if ( XTERM_PRESENT )	TOOLS_TERMINAL_COMMAND = wxT("xterm -hold -e ");	// Use xterm if it's present, as gnome-terminal lacks a 'stay-open' flag
								else TOOLS_TERMINAL_COMMAND = wxT("gnome-terminal -x ");
					}
		 else if ( XTERM_PRESENT )  { TERMINAL = wxT("xterm"); TERMINAL_COMMAND = wxT("xterm -e "); TOOLS_TERMINAL_COMMAND = wxT("xterm -hold -e "); }

	}

ConfigureTerminals::Save();							// Save all of this
}

//static
void Configure::SaveDefaultEditors( enum distro Distro, double Version )
{
wxConfigBase::Get()->Write( wxT("Editors/Icons/NoOfKnownIcons"),  8l );						// Save the icons first, in case no editors are found (whereupon we'd abort)
wxConfigBase::Get()->Write( wxT("Editors/Icons/a/Icon"),  wxT("kwrite.xpm") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/a/Icontype"),  9l );
wxConfigBase::Get()->Write( wxT("Editors/Icons/b/Icon"),  wxT("gedit.xpm") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/b/Icontype"),  9l );
wxConfigBase::Get()->Write( wxT("Editors/Icons/c/Icon"),  wxT("kedit.xpm") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/c/Icontype"),  9l );
if ( (Distro==debian  && Version>=4.0) || (Distro==zenwalk && Version>=5.0) )
	wxConfigBase::Get()->Write( wxT("Editors/Icons/d/Icon"),  wxT("iceweasel.png") );
else
	wxConfigBase::Get()->Write( wxT("Editors/Icons/d/Icon"),  wxT("firefox.png") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/d/Icontype"),  15l );
wxConfigBase::Get()->Write( wxT("Editors/Icons/e/Icon"),  wxT("editor.xpm") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/e/Icontype"),  9l );
wxConfigBase::Get()->Write( wxT("Editors/Icons/f/Icon"),  wxT("mousepad.png") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/f/Icontype"),  15l );
wxConfigBase::Get()->Write( wxT("Editors/Icons/g/Icon"),  wxT("openoffice.png") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/g/Icontype"),  15l );
wxConfigBase::Get()->Write( wxT("Editors/Icons/h/Icon"),  wxT("unknown.xpm") );
wxConfigBase::Get()->Write( wxT("Editors/Icons/h/Icontype"),  9l );

wxConfigBase::Get()->Write(wxT("/Editors/NoOfEditors"),  0l );			// Save a zero count first, just so that it's stored first (& so can be more easily seen)
long count=0; wxString path( wxT("/Editors/") );

if ( TestExistence( wxT("kwrite"), true ) )											// Now see which of the likely possibilities are available
	{	wxString key( path + wxChar( wxT('a') + count++ ) ); key += wxT('/');
		wxConfigBase::Get()->Write( key+wxT("Editor"),  wxT("kwrite") ); wxConfigBase::Get()->Write( key+wxT("LaunchString"), wxT("kwrite") );
	}	// We can skip the rest for kwrite, as the zero defaults are correct
 else
   if ( TestExistence( wxT("kate"), true ) )  // I doubt if anyone would want icons for both kwrite and kate
    {	wxString key( path + wxChar( wxT('a') + count++ ) ); key += wxT('/');
      wxConfigBase::Get()->Write( key+wxT("Editor"),  wxT("kate") ); wxConfigBase::Get()->Write( key+wxT("LaunchString"), wxT("kate") );
      wxConfigBase::Get()->Write( key+wxT("CanUseTabs"),  true ); // Reuse the kwrite icon, which I think kate does anyway
    }

if ( TestExistence( wxT("gedit"), true ) )
	{	wxString key( path + wxChar( wxT('a') + count++ ) ); key += wxT('/');
		wxConfigBase::Get()->Write( key+wxT("Editor"),  wxT("gedit") ); wxConfigBase::Get()->Write( key+wxT("LaunchString"), wxT("gedit") );
		wxConfigBase::Get()->Write( key+wxT("CanUseTabs"),  true ); wxConfigBase::Get()->Write( key+wxT("IconNo"), 1l );
	}

if ( TestExistence( wxT("mousepad"), true ) )
	{	wxString key( path + wxChar( wxT('a') + count++ ) ); key += wxT('/');
		wxConfigBase::Get()->Write( key+wxT("Editor"),  wxT("mousepad") ); wxConfigBase::Get()->Write( key+wxT("LaunchString"), wxT("mousepad") );
		wxConfigBase::Get()->Write( key+wxT("CanUseTabs"),  false ); wxConfigBase::Get()->Write( key+wxT("IconNo"), 5l );
	}

if ( TestExistence( wxT("firefox"), true ) )
	{	wxString key( path + wxChar( wxT('a') + count++ ) ); key += wxT('/');
		wxConfigBase::Get()->Write( key+wxT("Editor"),  wxT("firefox") ); wxConfigBase::Get()->Write( key+wxT("LaunchString"), wxT("firefox") );
		wxConfigBase::Get()->Write( key+wxT("CanUseTabs"),  false ); wxConfigBase::Get()->Write( key+wxT("IconNo"), 3l );
	}

if ( TestExistence( wxT("openoffice"), true ) )
	{	wxString key( path + wxChar( wxT('a') + count++ ) ); key += wxT('/');
		wxConfigBase::Get()->Write( key+wxT("Editor"),  wxT("openoffice") ); wxConfigBase::Get()->Write( key+wxT("LaunchString"), wxT("openoffice") );
		wxConfigBase::Get()->Write( key+wxT("CanUseTabs"),  false ); wxConfigBase::Get()->Write( key+wxT("IconNo"), 6l );
	}
  else if ( TestExistence( wxT("ooffice"), true ) )
	{	wxString key( path + wxChar( wxT('a') + count++ ) ); key += wxT('/');
		wxConfigBase::Get()->Write( key+wxT("Editor"),  wxT("openoffice") ); wxConfigBase::Get()->Write( key+wxT("LaunchString"), wxT("ooffice") );
		wxConfigBase::Get()->Write( key+wxT("CanUseTabs"),  false ); wxConfigBase::Get()->Write( key+wxT("IconNo"), 6l );
	}

wxConfigBase::Get()->Write(wxT("/Editors/NoOfEditors"),  count );			// Now save the actual count
}

//static
void Configure::SaveDefaultUDTools()
{
wxConfigBase* config = wxConfigBase::Get();
bool kw=false, ge=false; long count=4; 					// If we don't use kwrite or gedit, there'll be 4 items in the main menu
wxString name( wxT("kwrite") ); 
if ( TestExistence( name ) )	{ ++count; kw=true; }		// See if kwrite &/or gedit are available. If so, inc count
 else { name = wxT("gedit");
			if ( TestExistence( name ) )	 { ++count; ge=true; }
		}

wxString su_string;
if ( KDESU_PRESENT ) su_string << wxT("kdesu ");
	else if ( GKSU_PRESENT ) su_string << wxT("gksu ");
		else if ( XSU_PRESENT ) su_string << wxT("gnomesu  --title "" -c ");
			else su_string << wxT("sudo ");						// If none of the gui su.s are available, sudo might still work


config->Write( wxT("Tools/Launch/Label"),  wxT("Run a Program") );
config->Write( wxT("Tools/Launch/Count"),  count );

config->Write(  wxT("Tools/Launch/label_a"), wxT("df") );
config->Write( wxT("Tools/Launch/data_a"), wxT("df -h") );
config->Write( wxT("Tools/Launch/terminal_a"), true );
config->Write( wxT("Tools/Launch/persists_a"), true );

config->Write(  wxT("Tools/Launch/label_b"), wxT("du") );
config->Write( wxT("Tools/Launch/data_b"), wxT("du -h") );
config->Write( wxT("Tools/Launch/terminal_b"), true );
config->Write( wxT("Tools/Launch/persists_b"), true );

config->Write(  wxT("Tools/Launch/label_c"), wxT("md5sum") );
config->Write( wxT("Tools/Launch/data_c"), wxT("md5sum %f") );
config->Write( wxT("Tools/Launch/terminal_c"), true );
config->Write( wxT("Tools/Launch/persists_c"), true );

config->Write(  wxT("Tools/Launch/label_d"), wxT("touch %a") );
config->Write( wxT("Tools/Launch/data_d"), wxT("touch %a") );
config->Write( wxT("Tools/Launch/terminal_d"), false );
config->Write( wxT("Tools/Launch/persists_d"), false );

if ( kw )
	{	wxString command( su_string );
		command << wxT("kwrite");
		config->Write(  wxT("Tools/Launch/label_e"), command );
		command << wxT(" %f"); config->Write( wxT("Tools/Launch/data_e"), command );
		config->Write( wxT("Tools/Launch/terminal_e"), (su_string==wxT("sudo ") ? true : false) );
		config->Write( wxT("Tools/Launch/persists_e"), false );
	}
 else if ( ge )
	{	wxString command( su_string );
		command << wxT("gedit");
		config->Write(  wxT("Tools/Launch/label_e"), command );
		command << wxT(" %a"); config->Write( wxT("Tools/Launch/data_e"), command );
		config->Write( wxT("Tools/Launch/terminal_e"), (su_string==wxT("sudo ") ? true : false) );
		config->Write( wxT("Tools/Launch/persists_e"), false );
	}

if ( ! wxFileExists( wxT("/etc/debian_version") ) )											// Debian folk won't want rpm stuff
	{	config->Write( wxT("Tools/Launch/a/Label"),  wxT("RPM") );
		config->Write( wxT("Tools/Launch/a/Count"),  3 );
		
		config->Write(  wxT("Tools/Launch/a/label_a"), wxT("Install: rpm -Uvh") );
		config->Write( wxT("Tools/Launch/a/data_a"), wxT("sudo rpm -Uvh %f") );
		config->Write( wxT("Tools/Launch/a/terminal_a"), true );
		config->Write( wxT("Tools/Launch/a/persists_a"), true );
		
		config->Write(  wxT("Tools/Launch/a/label_b"), wxT("Query rpm") );
		config->Write( wxT("Tools/Launch/a/data_b"), wxT("sudo rpm -qpi %f") );
		config->Write( wxT("Tools/Launch/a/terminal_b"), true );
		config->Write( wxT("Tools/Launch/a/persists_b"), true );
		
		config->Write(  wxT("Tools/Launch/a/label_c"), wxT("List files in rpm") );
		config->Write( wxT("Tools/Launch/a/data_c"), wxT("sudo rpm -qpl %f") );
		config->Write( wxT("Tools/Launch/a/terminal_c"), true );
		config->Write( wxT("Tools/Launch/a/persists_c"), true );
	}
}

//static
void Configure::SaveDefaultFileAssociations()
{
wxConfigBase* config = wxConfigBase::Get();
bool kw=false, ge=false, kt=false, mp=false, im=false, gm=false, kp=false, gp=false, ev=false, ac=false, ff=false, oo=false, kf=false, tm=false ;
long cumcount=0;  wxArrayString txtarray, pdfarray, mmarray, graflabelarray, grafcmdarray;
wxString name( wxT("kwrite") ); 
if ( TestExistence( name ) )	kw=true;			// See if kwrite, gedit. ImageMagick etc are available. If so, inc count
name = wxT("gedit");	if ( TestExistence( name ) )	 ge=true;
name = wxT("kate");	if ( TestExistence( name ) )	 kt=true;
name = wxT("mousepad");	if ( TestExistence( name ) )	 mp=true;

name = wxT("kpdf");	if ( TestExistence( name ) )	 kp=true;
name = wxT("gpdf");	if ( TestExistence( name ) )	 gp=true;
name = wxT("evince");	if ( TestExistence( name ) )	 ev=true;
name = wxT("acroread");	if ( TestExistence( name ) )	 ac=true;
name = wxT("ooffice");	if ( TestExistence( name ) )	 oo=true;

name = wxT("firefox");	if ( TestExistence( name ) )	 ff=true;

name = wxT("display");	if ( TestExistence( name ) )	 im=true;
name = wxT("gimp");	if ( TestExistence( name ) )	 gm=true;

name = wxT("kaffeine");	if ( TestExistence( name ) )	 kf=true;
name = wxT("totem");	if ( TestExistence( name ) )	 tm=true;

													// First the data for the OpenWith dialog
config->Write( wxT("FileAssociations/Audio/dummy"), wxT("") );
config->Write( wxT("FileAssociations/Development/dummy"), wxT("") );
if ( kw )
	{	config->Write( wxT("FileAssociations/Editors/KWrite/Ext"), wxT("txt,TXT,tex,README") );
		config->Write( wxT("FileAssociations/Editors/KWrite/Filepath"), wxT("kwrite") );
		config->Write( wxT("FileAssociations/Editors/KWrite/Command"), wxT("kwrite %s") );
		txtarray.Add( wxT("kwrite") );
	}
if ( ge )
	{	config->Write( wxT("FileAssociations/Editors/gedit/Ext"), wxT("txt,TXT,tex,README") );
		config->Write( wxT("FileAssociations/Editors/gedit/Filepath"), wxT("gedit") );
		config->Write( wxT("FileAssociations/Editors/gedit/Command"), wxT("gedit %s") );
		txtarray.Add( wxT("gedit") );
	}
if ( kt )
	{	config->Write( wxT("FileAssociations/Editors/kate/Ext"), wxT("txt,TXT,tex,README") );
		config->Write( wxT("FileAssociations/Editors/kate/Filepath"), wxT("kate") );
		config->Write( wxT("FileAssociations/Editors/kate/Command"), wxT("kate %s") );
		txtarray.Add( wxT("kate") );
	}
if ( mp )
	{	config->Write( wxT("FileAssociations/Editors/mousepad/Ext"), wxT("txt,TXT,tex,README") );
		config->Write( wxT("FileAssociations/Editors/mousepad/Filepath"), wxT("mousepad") );
		config->Write( wxT("FileAssociations/Editors/mousepad/Command"), wxT("mousepad %s") );
		txtarray.Add( wxT("mousepad") );
	}
if ( ! txtarray.GetCount() )  config->Write( wxT("FileAssociations/Editors/dummy"), wxT("") );		// If none is present, we still need a dummy entry

if ( im )
	{	config->Write( wxT("FileAssociations/Graphics/ImageMagick/Ext"), wxT("gif,xpm,png,jpg,jpeg,JPG,JPEG,bmp") );
		config->Write( wxT("FileAssociations/Graphics/ImageMagick/Filepath"), wxT("display") );
		config->Write( wxT("FileAssociations/Graphics/ImageMagick/Command"), wxT("display %s") );
		graflabelarray.Add( wxT("ImageMagick") ); grafcmdarray.Add( wxT("display") );
	}
if ( gm )
	{	config->Write( wxT("FileAssociations/Graphics/The Gimp/Ext"), wxT("gif,xpm,png,jpg,jpeg,JPG,JPEG,bmp") );
		config->Write( wxT("FileAssociations/Graphics/The Gimp/Filepath"), wxT("gimp") );
		config->Write( wxT("FileAssociations/Graphics/The Gimp/Command"), wxT("gimp %s") );
		graflabelarray.Add( wxT("The Gimp") ); grafcmdarray.Add( wxT("gimp") );
	}
if ( ! graflabelarray.GetCount() )   config->Write( wxT("FileAssociations/Graphics/dummy"), wxT("") );

if ( ff )
	{	config->Write( wxT("FileAssociations/Internet/firefox/Ext"), wxT("html,htm,HTML,HTM") );
		config->Write( wxT("FileAssociations/Internet/firefox/Filepath"), wxT("firefox") );
		config->Write( wxT("FileAssociations/Internet/firefox/Command"), wxT("firefox %s") );
	}
 else config->Write( wxT("FileAssociations/Internet/dummy"), wxT("") );

if ( gp )
	{	config->Write( wxT("FileAssociations/Office/gpdf/Ext"), wxT("pdf,PDF") );
		config->Write( wxT("FileAssociations/Office/gpdf/Filepath"), wxT("gpdf") );
		config->Write( wxT("FileAssociations/Office/gpdf/Command"), wxT("gpdf %s") );
		pdfarray.Add( wxT("gpdf") );	// This is used later
	}

if ( ev )
	{	config->Write( wxT("FileAssociations/Office/evince/Ext"), wxT("pdf,PDF") );
		config->Write( wxT("FileAssociations/Office/evince/Filepath"), wxT("evince") );
		config->Write( wxT("FileAssociations/Office/evince/Command"), wxT("evince %s") );
		pdfarray.Add( wxT("evince") );
	}

if ( kp )
	{	config->Write( wxT("FileAssociations/Office/kpdf/Ext"), wxT("pdf,PDF") );
		config->Write( wxT("FileAssociations/Office/kpdf/Filepath"), wxT("kpdf") );
		config->Write( wxT("FileAssociations/Office/kpdf/Command"), wxT("kpdf %s") );
		pdfarray.Add( wxT("kpdf") );
	}

if ( ac )
	{	config->Write( wxT("FileAssociations/Office/acroread/Ext"), wxT("pdf,PDF") );
		config->Write( wxT("FileAssociations/Office/acroread/Filepath"), wxT("acroread") );
		config->Write( wxT("FileAssociations/Office/acroread/Command"), wxT("acroread %s") );
		pdfarray.Add( wxT("acroread") );
	}

if ( oo )
	{	config->Write( wxT("FileAssociations/Office/oocalc/Ext"), wxT("sxc,xls,ods") );
		config->Write( wxT("FileAssociations/Office/oowriter/Ext"), wxT("sxw,doc,odt,rtf") );
		config->Write( wxT("FileAssociations/Office/oocalc/Filepath"), wxT("oocalc") );
		config->Write( wxT("FileAssociations/Office/oowriter/Filepath"), wxT("oowriter") );
		config->Write( wxT("FileAssociations/Office/oocalc/Command"), wxT("oocalc %s") );
		config->Write( wxT("FileAssociations/Office/oowriter/Command"), wxT("oowriter %s") );
	}

if ( ! (oo || kp || gp || ev || ac) )  config->Write( wxT("FileAssociations/Office/dummy"), wxT("") );

if ( kf )
	{	config->Write( wxT("FileAssociations/Multimedia/kaffeine/Ext"), wxT("avi,mov,mpeg,mpg") );
		config->Write( wxT("FileAssociations/Multimedia/kaffeine/Filepath"), wxT("kaffeine") );
		config->Write( wxT("FileAssociations/Multimedia/kaffeine/Command"), wxT("kaffeine %s") );
		mmarray.Add( wxT("kaffeine") );
	}
	
if ( tm )
	{	config->Write( wxT("FileAssociations/Multimedia/totem/Ext"), wxT("avi,mov,mpeg,mpg") );
		config->Write( wxT("FileAssociations/Multimedia/totem/Filepath"), wxT("totem") );
		config->Write( wxT("FileAssociations/Multimedia/totem/Command"), wxT("totem %s") );
		mmarray.Add( wxT("totem") );
	}
	
if ( ! (kf || tm ) )  config->Write( wxT("FileAssociations/Multimedia/dummy"), wxT("") );

config->Write( wxT("FileAssociations/Other/dummy"), wxT("") );
config->Write( wxT("FileAssociations/Utilities/dummy"), wxT("") );

							// Now the Default App for Ext thing, and which apps to display in the Open menu
wxChar group=wxT('a');																	// This will provide e.g. the 'b' etc in LaunchExt/b
config->Write( wxT("LaunchExt/count"), 0 );									// Do an initial write, just to make it the first entry

if ( txtarray.GetCount() )
	{	const int txtcount = 4; long appcount = txtarray.GetCount();
		wxString defaultlabel, defaultcmd;
		if ( appcount ) { defaultlabel = txtarray[0]; defaultcmd = defaultlabel << wxT(" %s"); } // Use the first of any commands as the default

		const wxChar* ext[txtcount] = { wxT("txt"), wxT("TXT"), wxT("tex"), wxT("README") };
		for ( int n=0; n < txtcount; ++n, ++group )
			{	wxString key = wxT("LaunchExt/"); key << group;
				config->Write( key + wxT("/Ext"), wxString(ext[n]) );
				config->Write( key + wxT("/appcount"), appcount );
				config->Write( key + wxT("/DefaultLabel"), defaultlabel );
				config->Write( key + wxT("/DefaultCommand"), defaultcmd );
				
				wxChar which=wxT('a');
				for ( int app=0; app < (int)appcount; ++app, ++which )
					{	wxString appkey = key + wxT('/'); appkey << which;					// Thus making appkey LaunchExt/b/a, LaunchExt/b/b etc
						wxString command( txtarray[app] ); command << wxT(" %s");
						config->Write( appkey + wxT("/Label"), txtarray[app] );
						config->Write( appkey + wxT("/Command"), command );
					}
			}

		cumcount += txtcount;
	}

if ( pdfarray.GetCount() )
	{	const int pdfcount = 2; long appcount = pdfarray.GetCount();
		wxString defaultlabel, defaultcmd;
		if ( appcount ) { defaultlabel = pdfarray[0]; defaultcmd = defaultlabel << wxT(" %s"); } // Use the first of any commands as the default

		const wxChar* ext[pdfcount] = { wxT("pdf"), wxT("PDF") };
		for ( int n=0; n < pdfcount; ++n, ++group )
			{	wxString key = wxT("LaunchExt/"); key << group;
				config->Write( key + wxT("/Ext"), wxString(ext[n]) );
				config->Write( key + wxT("/appcount"), appcount );
				config->Write( key + wxT("/DefaultLabel"), defaultlabel );
				config->Write( key + wxT("/DefaultCommand"), defaultcmd );
				
				wxChar which=wxT('a');
				for ( int app=0; app < (int)appcount; ++app, ++which )
					{	wxString appkey = key + wxT('/'); appkey << which;					// Thus making appkey LaunchExt/b/a, LaunchExt/b/b etc
						wxString command( pdfarray[app] ); command << wxT(" %s");
						config->Write( appkey + wxT("/Label"), pdfarray[app] );
						config->Write( appkey + wxT("/Command"), command );
					}
			}

		cumcount += pdfcount;
	}

if ( oo )
	{	const int oocount = 7;
		const wxChar* ext[oocount] = { wxT("sxw"), wxT("doc"), wxT("odt"), wxT("rtf"), wxT("xls"), wxT("ods"), wxT("sxc") };
		for ( int n=0; n < 4; ++n, ++group )										// First the 4 writer exts
			{	wxString key = wxT("LaunchExt/"); key << group;
				config->Write( key + wxT("/Ext"), wxString(ext[n]) );
				config->Write( key + wxT("/appcount"), 1l );
				config->Write( key + wxT("/DefaultLabel"), wxT("Openoffice writer") );
				config->Write( key + wxT("/DefaultCommand"), wxT("oowriter %s") );
				
				config->Write( key + wxT("/a/Label"), wxT("Openoffice writer") );
				config->Write( key + wxT("/a/Command"), wxT("oowriter %s") );
			}
		for ( int n=4; n < 7; ++n, ++group )										// Then the 3 calc ones
			{	wxString key = wxT("LaunchExt/"); key << group;
				config->Write( key + wxT("/Ext"), wxString(ext[n]) );
				config->Write( key + wxT("/appcount"), 1l );
				config->Write( key + wxT("/DefaultLabel"), wxT("Openoffice calc") );
				config->Write( key + wxT("/DefaultCommand"), wxT("oocalc %s") );
				
				config->Write( key + wxT("/a/Label"), wxT("Openoffice calc") );
				config->Write( key + wxT("/a/Command"), wxT("oocalc %s") );
			}
		cumcount += oocount;
	}

if ( ff )
	{	const int ffcount = 4; 
		const wxChar* ext[ffcount] = { wxT("html"), wxT("htm"), wxT("HTML"), wxT("HTM") };
		for ( int n=0; n < ffcount; ++n, ++group )
			{	wxString key = wxT("LaunchExt/"); key << group;
				config->Write( key + wxT("/Ext"), wxString(ext[n]) );
				config->Write( key + wxT("/appcount"), 1l );
				config->Write( key + wxT("/DefaultLabel"), wxT("firefox") );
				config->Write( key + wxT("/DefaultCommand"), wxT("firefox %s") );
				
				config->Write( key + wxT("/a/Label"), wxT("firefox") );
				config->Write( key + wxT("/a/Command"), wxT("firefox %s") );
			}
		cumcount += ffcount;
	}

if ( im || gm )
	{	const int grafcount = 4;  long appcount = im + gm;
		wxString defaultlabel, defaultcmd;
		if ( im ) { defaultlabel = wxT("ImageMagick"); defaultcmd = wxT("display %s"); }
		 else  { defaultlabel = wxT("The Gimp"); defaultcmd = wxT("gimp %s"); }
		const wxChar* ext[grafcount] = { wxT("jpg"), wxT("png"), wxT("gif"), wxT("bmp") };
		for ( int n=0; n < grafcount; ++n, ++group )
			{	wxString key = wxT("LaunchExt/"); key << group;
				config->Write( key + wxT("/Ext"), wxString(ext[n]) );
				config->Write( key + wxT("/appcount"), appcount );
				config->Write( key + wxT("/DefaultLabel"), defaultlabel );
				config->Write( key + wxT("/DefaultCommand"), defaultcmd );
		
				wxChar which=wxT('a');
				for ( int app=0; app < (int)appcount; ++app, ++which )
					{	wxString appkey = key + wxT('/'); appkey << which;					// Thus making appkey LaunchExt/e/a, LaunchExt/e/b etc
						wxString command( grafcmdarray[app] ); command << wxT(" %s");
						config->Write( appkey + wxT("/Label"), graflabelarray[app] );
						config->Write( appkey + wxT("/Command"), command );
					}
			}
		cumcount += grafcount;
	}

if ( kf || tm )
	{	const int mmcount = 4; long appcount = kf + tm;
		wxString defaultlabel, defaultcmd;
		if ( kf ) { defaultlabel = wxT("kaffeine"); defaultcmd = wxT("kaffeine %s"); }
		 else  { defaultlabel = wxT("totem"); defaultcmd = wxT("totem %s"); }

		const wxChar* ext[mmcount] = { wxT("avi"), wxT("mov"), wxT("mpeg"), wxT("mpg") };
		for ( int n=0; n < mmcount; ++n, ++group )
			{	wxString key = wxT("LaunchExt/"); key << group;
				config->Write( key + wxT("/Ext"), wxString(ext[n]) );
				config->Write( key + wxT("/appcount"), appcount );
				config->Write( key + wxT("/DefaultLabel"), defaultlabel );
				config->Write( key + wxT("/DefaultCommand"), defaultcmd );
				
				wxChar which=wxT('a');
				for ( int app=0; app < (int)appcount; ++app, ++which )
					{	wxString appkey = key + wxT('/'); appkey << which;
						wxString command( mmarray[app] ); command << wxT(" %s");
						config->Write( appkey + wxT("/Label"), mmarray[app] );
						config->Write( appkey + wxT("/Command"), command );
					}
			}

		cumcount += mmcount;
	}

config->Write( wxT("LaunchExt/count"), cumcount );							// Write the final count
}

//static
void Configure::SaveDefaultDirCtrlTBButtons()
{
const long NoOfKnownIcons = 5;
wxConfigBase* config = wxConfigBase::Get();
													// Start by creating the available-icon list
const wxChar* Icons[] = { wxT("gohome.xpm"), wxT("MyDocuments.xpm"), wxT("bm1_button.xpm"), wxT("bm2_button.xpm"), wxT("bm3_button.xpm") };
config->Write( wxT("/Misc/DirCtrlTB/Icons/NoOfKnownIcons"), NoOfKnownIcons );
for ( long n=0; n < NoOfKnownIcons; ++n )
	{	config->Write( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icon"), Icons[n] );
		config->Write( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icontype"), 9 ); // They're all xpms, as it happens
	}

config->Write( wxT("Misc/DirCtrlTB/Buttons/button_a/IconNo"),  0l );
config->Write( wxT("Misc/DirCtrlTB/Buttons/button_a/tooltip"),  wxT("Go to Home directory") );
config->Write( wxT("Misc/DirCtrlTB/Buttons/button_a/destination"),  HOME );

wxString name( wxT("/Documents") ); wxString docs = HOME + name;
FileData stat( docs );															// Let's see if there's a dir called Documents
if ( ! stat.IsValid() )
	{	name = wxT("/documents"); docs = HOME + name;
		FileData stat( docs ); if ( ! stat.IsValid() ) return;		// If not, & there isn't one called documents either, forget it
	}

config->Write( wxT("Misc/DirCtrlTB/Buttons/button_b/IconNo"),  1l );
config->Write( wxT("Misc/DirCtrlTB/Buttons/button_b/tooltip"),  wxT("Go to Documents directory") );
config->Write( wxT("Misc/DirCtrlTB/Buttons/button_b/destination"),  docs );
}

//static
void Configure::SaveDefaultMisc()
{
int width, height;
wxDisplaySize(&width, &height); width -= 30; height -= 100;		// Get the screen size, then chop off a bit for taskbar & contingencies
long vertsplit = width * 4 / 10, halfvertsplit = vertsplit / 2, horizsplit = height / 2;

wxConfigBase* config = wxConfigBase::Get();
config->Write( wxT("/Misc/AppHeight"), height );									// Save 4Pane's initial size
config->Write( wxT("/Misc/AppWidth"), width );

config->Write( wxT("/Tabs/tab_a/verticalsplitpos"), vertsplit );				// Now the initial pane sizes etc
config->Write( wxT("/Tabs/tab_a/horizontalsplitpos"), horizsplit );
config->Write( wxT("/Tabs/tab_a/leftverticalsashpos"), halfvertsplit );
config->Write( wxT("/Tabs/tab_a/rightverticalsashpos"), halfvertsplit );	// This actually allows more for the Rt fileview, which will have more cols
config->Write( wxT("/Tabs/tab_a/tophorizontalsashpos"), vertsplit);
config->Write( wxT("/Tabs/tab_a/bottomhorizontalsashpos"), vertsplit );
config->Write( wxT("/Tabs/tab_a/unsplitsashpos"), vertsplit );


wxString Lkey = wxT("/Tabs/tab_a/Lwidthcol"), Rkey = wxT("/Tabs/tab_a/Rwidthcol");
wxString LDeskey = wxT("LDesiredwidthcol"), RDeskey = wxT("RDesiredwidthcol");

config->Write( wxT("/Tabs/tab_a/Lwidthcol0"), halfvertsplit );
config->Write( wxT("/Tabs/tab_a/LDesiredwidthcol0"), halfvertsplit );
config->Write( wxT("/Tabs/tab_a/Rwidthcol0"), halfvertsplit );
config->Write( wxT("/Tabs/tab_a/RDesiredwidthcol0"), halfvertsplit );
config->Write( wxT("/Tabs/tab_a/Rwidthcol1"), halfvertsplit/4 );		// Ext
config->Write( wxT("/Tabs/tab_a/RDesiredwidthcol1"), halfvertsplit/4 );
config->Write( wxT("/Tabs/tab_a/Rwidthcol2"), halfvertsplit/4 );		// Size
config->Write( wxT("/Tabs/tab_a/RDesiredwidthcol2"), halfvertsplit/4 );
config->Write( wxT("/Tabs/tab_a/Rwidthcol3"), halfvertsplit/2 );		// Time
config->Write( wxT("/Tabs/tab_a/RDesiredwidthcol3"), halfvertsplit/2 );

}

//static
void Configure::CopyConf( const wxString& iniFilepath )	// Copy any master conf file into the local ini
{
wxString conf; bool ConfAvailable = Configure::CheckForConfFile( conf );
if ( ConfAvailable )
	wxCopyFile( conf, iniFilepath );
}

//static
void Configure::SaveDefaultsToIni()	// Make sure there is a valid config file, with reasonably-sensible defaults
{
wxConfigBase* config = wxConfigBase::Get(); 
if (config==NULL)    
	{	wxString msg( _("If you are seeing this message (and you shouldn't!), it's because the configuration file couldn't be saved.\nThe least-unlikely reasons would be incorrect permissions or a Read Only partition."));
		wxMessageBox( msg ); return;
	}

config->SetPath(wxT("/"));
config->Write( wxT("Aardvark"),  wxT("Dummy first line") );		// Since wxFileConfig tends to be alphabetical in its Group order, and sometimes crashes when it tries to insert at line 0

double Version; enum distro Distro = DeduceDistro( Version );	
if ( ! config->Exists( wxT("Devices/Misc/FloppyDevicesAutomount") ) )	SaveProbableHAL_Sys_etcInfo( Distro, Version );
SaveDefaultSus();
if ( ! config->Exists( wxT("/Misc/Terminal/TERMINAL") ) )  SaveDefaultTerminal(Distro, Version);
if ( ! config->Exists( wxT("/Editors/NoOfEditors") ) ) SaveDefaultEditors( Distro, Version );
if ( ! config->Exists( wxT("Tools/Launch/Label") ) )  SaveDefaultUDTools();
if ( ! config->Exists( wxT("FileAssociations/Audio") ) )  SaveDefaultFileAssociations();
SaveDefaultDirCtrlTBButtons();
SaveDefaultMisc();
config->Write( wxT("/Network/ShowmountFilepath"), SHOWMOUNTFPATH );			// In case we found it in /sbin/

config->Write( wxT("Misc/Identifier"),  wxT("Genuine config file") );		// Accept no substitutes
config->Flush();
}

//static
bool Configure::CheckForConfFile( wxString& conf )	// Is there a 4Pane.conf file anywhere, from which to load some config data?
{
wxString filepath = wxGetHomeDir() + wxT("/4Pane.conf"); FileData home( filepath );
if ( home.IsValid() ) { conf = home.GetUltimateDestination(); return true; }	// If there's a 4Pane.conf at  home, return it (allowing for symlinkage)

filepath = wxGetHomeDir() + wxT("/4pane.conf"); FileData home1( filepath );
if ( home1.IsValid() ) { conf = home1.GetUltimateDestination(); return true; }	// Ditto lowercase

filepath = wxGetCwd() + wxT("/4Pane.conf"); FileData Cwd( filepath );
if ( Cwd.IsValid() ) { conf = Cwd.GetUltimateDestination(); return true; }		// Now look in CWD

filepath = wxGetCwd() + wxT("/4pane.conf"); FileData Cwd1( filepath );
if ( Cwd1.IsValid() ) { conf = Cwd1.GetUltimateDestination(); return true; }

filepath = wxT("/etc/4Pane/4Pane.conf"); FileData etc( filepath );					// And finally,  /etc/4[Pp]ane/4[Pp]ane.conf
if ( etc.IsValid() ) { conf = etc.GetUltimateDestination(); return true; }

filepath = wxT("/etc/4Pane/4pane.conf"); FileData etc1( filepath );
if ( etc1.IsValid() ) { conf = etc1.GetUltimateDestination(); return true; }

filepath = wxT("/etc/4pane/4Pane.conf"); FileData etc2( filepath );
if ( etc2.IsValid() ) { conf = etc2.GetUltimateDestination(); return true; }

filepath = wxT("/etc/4pane/4pane.conf"); FileData etc3( filepath );
if ( etc3.IsValid() ) { conf = etc3.GetUltimateDestination(); return true; }

return false;
}

//static
bool Configure::Wizard( const wxString& configFPath )
{
NoConfigWizard* oz = new NoConfigWizard(NULL, wxID_ANY, wxT("4Pane configuration wizard"), configFPath );
bool result = oz->RunWizard( oz->Page1 );

if  ( result && oz->shortcut && oz->shortcut->IsChecked() )						// Provide a desktop shortcut if appropriate
	{	wxLogNull shh; wxCopyFile( wxString( RCDIR + wxT("/4Pane.desktop") ), wxString( wxGetHomeDir() + wxT("/Desktop/4Pane.desktop") ) ); }
	
oz->Destroy();
return result;
}

bool Configure::TestExistence( wxString name, bool executable/*=false*/ )		// Is 'name' in $PATH, and optionally is it executable by us
{
if ( name.IsEmpty() ) return false;

if ( wxIsAbsolutePath( name ) )													// First try it neat, if it's an absolute filepath (which it seldom will be)
	{	FileData stat( name );
		if ( stat.IsValid() )
			return ( executable ? stat.CanTHISUserExecute() : true );	// Aha!
	}

FileData stat( name );																	// Next look in the cwd, just on the off-chance
if ( stat.IsValid() )
	return ( executable ? stat.CanTHISUserExecute() : true );

wxString Path(wxT("PATH")), PATH;
wxGetEnv( Path, &PATH );															// Get $PATH into PATH

wxStringTokenizer tkn( PATH, wxT(":"), wxTOKEN_STRTOK );
size_t count = tkn.CountTokens();
for ( size_t c=0; c < count; ++c )												// Go thru the PATH, looking for name in the filepath
	{	wxString next = tkn.GetNextToken();
		FileData stat( next + wxFILE_SEP_PATH + name );
		if ( ! stat.IsValid() ) continue;
		return (executable ? stat.CanTHISUserExecute() : true);	// Found it. If we were asked just for existence, return true. Else return true if we can execute.
	}

return false;																				// If we're here, it wasn't found
}

void Configure::LoadConfiguration()
{
ToolbarIconsDialog::LoadDirCtrlTBinfo( DirCtrlTBinfoarray );		// Load info on the desired extra buttons to load to each Dirview mini-toolbar
LoadMisc();
}

void Configure::SaveConfiguration()
{
SaveMisc();
wxConfigBase::Get()->Flush();
}

void Configure::LoadMisc()		// Load the one-off things like Filter history
{
ConfigureDisplay::LoadDisplayConfig();              // Load the 'Display' stuff e.g. ASK_ON_DELETE
ConfigureTerminals::LoadTerminalsConfig();          // Ditto terminals
ConfigureNet::LoadNetworkConfig();                  // and network
ConfigureMisc::LoadMiscConfig();                    // and Misc

wxConfigBase* config = wxConfigBase::Get();	if (config==NULL)  return; 

MAX_COMMAND_HISTORY = (size_t)config->Read( wxT("/Misc/MaxCommandHistory"), 15l );
config->Read( wxT("/Misc/RETAIN_REL_TARGET"), &RETAIN_REL_TARGET, 1 );	// Whether, when a relative symlink is moved, its target remains unchanged

config->SetPath( wxT("/History/FilterHistory/") );
size_t count = config->GetNumberOfEntries();        // Count the entries
FilterHistory.Clear();

wxString item, key, histprefix(wxT("hist_"));	
for ( size_t n=0; n < count; ++n )                  // Create the key for every item, from hist_a to hist_('a'+count-1)
	{	key.Printf(wxT("%c"), wxT('a') + (wxChar)n );   // Make key hold a, b, c etc
		config->Read( histprefix+key, &item );
		if ( item.IsEmpty() )   continue;
		FilterHistory.Add( item );
	}

WHICHGREP = (greptype)config->Read( wxT("/Misc/whichgrep"), (long)PREFER_QUICKGREP ); // Which flavour of grep does the user prefer

config->SetPath(wxT("/"));
}

void Configure::SaveMisc()		// Save the one-off things like Filter history
{
wxConfigBase* config = wxConfigBase::Get();
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return; }

config->Write( wxT("/Misc/RETAIN_REL_TARGET"), RETAIN_REL_TARGET );	// Whether, when a relative symlink is moved, its target remains unchanged

config->DeleteGroup( wxT("/History/FilterHistory") );   // Delete current info, otherwise we'll end up with duplicates or worse
config->SetPath( wxT("/History/FilterHistory/") );

size_t count = wxMin( FilterHistory.GetCount(), MAX_COMMAND_HISTORY );  // Count the entries to be saved: not TOO many

wxString item, key, histprefix(wxT("hist_"));           // Create the key for every item, from hist_a to hist_('a'+count-1)
for ( size_t n=0;  n < count; ++n )                     // The newest command is stored at the beginning of the array
	{	key.Printf(wxT("%c"), wxT('a') + (wxChar)n );       // Make key hold a, b, c etc
		config->Write( histprefix+key, FilterHistory[n] );  // Save the entry
	}

config->Write(wxT("/Misc/whichgrep"), (long)WHICHGREP); // Which flavour of grep does the user prefer

config->SetPath(wxT("/"));
}

void Configure::ConfigureTooltips()	// Globally turn tooltips on/off, set their delay
{
#if defined __WXGTK__
wxConfigBase* config = wxConfigBase::Get();					// Find the config data
if (config==NULL)  return;

unsigned long delay = config->Read( wxT("/Misc/TooltipDelay"), 2000l );
if ( delay > 600000 ) delay = 2000;										// If delay is more than 10 mins, it's probably an error!

wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, wxWindow::FindFocus(), wxT("TooltipDlg"));
if ( ! delay)	((wxCheckBox*)dlg.FindWindow( wxT("Check") ))->SetValue( true );	// If delay is 0, the 'Turned off' checkbox is set
	else ((wxSpinCtrl*)dlg.FindWindow( wxT("Spin") ))->SetValue( delay / 1000 );

if ( dlg.ShowModal() != wxID_OK )		return;					// If user cancelled, abort

if ( ((wxCheckBox*)dlg.FindWindow( wxT("Check") ))->IsChecked() )
			delay = 0;
 else	delay = ((wxSpinCtrl*)dlg.FindWindow( wxT("Spin") ))->GetValue() * 1000;

config->Write(wxT("/Misc/TooltipDelay"), (long)delay);		// Save the info
config->Flush();

wxToolTip::Enable( delay );												// Implement any change
if ( delay )  wxToolTip::SetDelay( delay );
#endif
}

void Configure::ApplyTooltipPreferences()	// Loads tooltip choices, and makes it happen. Don't do this too early in the app, or it won't 'take'
{
#if defined __WXGTK__
unsigned long delay = wxConfigBase::Get()->Read( wxT("/Misc/TooltipDelay"), 2000l );
if ( delay > 600000 ) delay = 2000;										// If delay is more than 10 mins, it's probably an error!

wxToolTip::Enable( delay );
if ( delay )  wxToolTip::SetDelay( delay );
#endif
}

void Configure::Configure4Pane()		// The main Configuration
{
MyConfigureDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("ConfigureDlg"));

CallersHelpcontext = HelpContext;																						// Save the global to local store
HelpContext = HCconfigUDC;																								//   & set it to UDCommands as this comes first

((ConfigureNotebook*)dlg.FindWindow( wxT("Notebook") ))->Init( this );

dlg.ShowModal();

HelpContext = CallersHelpcontext;																						// Revert to the original value
}


IMPLEMENT_DYNAMIC_CLASS(ConfigureNotebook, wxNotebook)

void MyConfigureDialog::EndModal(int retCode)
{
if ( ! ((ConfigureNotebook*)FindWindow( wxT("Notebook") ))->Dirty() )								// If the notebook has unsaved data, give user a chance to save
	wxDialog::EndModal(retCode);																						// If there isn't, or he doesn't, pass to base-class EndModal
}

void ConfigureNotebook::Init( Configure* dad )
{
parent = dad;

long startpage = wxConfigBase::Get()->Read( wxT("/Configure/Startpage"), 0l );

	#if wxVERSION_NUMBER < 2500
if ( startpage < GetPageCount() )  SetSelection( startpage );
	#else
if ( startpage < (int)GetPageCount() )  SetSelection( startpage );
	#endif
 else SetSelection( 0 );


UDCPage = UDCNew;																										// For the User-defined Command page, start with New Command
panel = (wxPanel*)FindWindow( wxT("UDCAddTool") );													//  & so initialise panel to point to this subpanel

for ( int n=0; n < NotSelected; ++n )   
	{	InitPage( (pagename)n );
		dirty[n] = false; 
	}
}

void ConfigureNotebook::InitPage( enum pagename page )
{
switch( page )
	{	case UDCommand:	{LMT = MyFrame::mainframe->Layout->m_notebook->LaunchFromMenu;
											FolderArray = LMT->GetFolderArray(); ToolArray = LMT->GetToolarray();
											if ( UDCPage == UDCNew )
												{	UDCFilepath = (wxTextCtrl*)FindWindow( wxT("UDCFilepath") );
													UDCCombo = (wxComboBox*)FindWindow( wxT("UDCMenus") );
													UDCCombo->Clear(); size_t count=0; FillUDCMenusCombo( count, 0 );		// Recursively fill the folders combo
													#if wxVERSION_NUMBER >= 2701 && (defined(__WXGTK20__) || defined(__WXX11__))
													if ( ! UDCCombo->IsEmpty() )  UDCCombo->SetSelection(0);	// Show the first item in the combo textctrl: needed in >2.7.0.1
													#endif
													((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->SetValue(false);
													((wxCheckBox*)FindWindow( wxT("UDCTerminalPersists") ))->SetValue(false);
												}
											  else 													// Edit & Delete.  Reuse the ptrs, as there won't be any conflict
												{	UDCFilepath = (wxTextCtrl*)FindWindow( wxT("UDCFilepath") );
													UDCCombo = (wxComboBox*)FindWindow( wxT("UDCLabelCombo") );
													UDCCombo->Clear();																				// Clear & fill the Labels combo
													for ( size_t n=0; n < ToolArray->GetCount(); ++n )  UDCCombo->Append( ToolArray->Item(n)->Label );
													#if wxVERSION_NUMBER >= 2701 && (defined(__WXGTK20__) || defined(__WXX11__))
													if ( ! UDCCombo->IsEmpty() )  UDCCombo->SetSelection(0);
													#endif
													wxCommandEvent event; event.SetInt(0); OnUDCLabelComboSelectionChanged(event);	// Fake an event to obtain the Filepath
												}

											editing = false;																								// Used by Edit, but needs to false in others
											return;}
											
		case UDShortcuts:	{wxNotebookPage* page = GetPage( UDShortcuts );		// Get a ptr to this page.  Then add the relevant panel, shared with AcceleratorList
											MyShortcutPanel* panel = (MyShortcutPanel*)wxXmlResource::Get()->LoadPanel( page, wxT("ConfigureShortcuts") );
											panel->Init( MyFrame::mainframe->AccelList, true );
											
											wxBoxSizer* sizer = new wxBoxSizer( wxVERTICAL ); sizer->Add( panel, 1, wxEXPAND );
											#if defined(__WXGTK20__)
											F10check = new wxCheckBox( page, wxID_ANY, _(" Prevent the F10 key from opening the File menu") );
											F10check->SetToolTip(_("If this is unticked, gtk2 grabs the F10 key-press, so you can't use it as a shortcut (it's the default one for  \"New file or dir\")."));
											panel->F10check = F10check; 										// It needs 2b stored in MyShortcutPanel too, for later access
											F10check->SetValue( USE_F10_KEY );
											page->GetSizer()->Add( F10check, 0, wxALIGN_CENTRE | wxTOP|wxBOTTOM, 10 ); 
											#endif									
											page->GetSizer()->Add( sizer, 1, wxEXPAND ); page->GetSizer()->Layout();
											
											wxNotifyEvent event(MyListCtrlEvent);	// Send a custom event to panel to size listctrl columns.  Can't just tell it: list won't be properly made yet
											wxPostEvent(panel, event);

											return;}
											
		case ConfigDevicesEditors:	return;
		case ConfigDisplay:		ConfDisplay = new ConfigureDisplay; return;
		case ConfigTerminals:	ConfTerminals = new ConfigureTerminals; return;
		case ConfigNet:				ConfNet = new ConfigureNet; return;
		case ConfigMisc:			ConfMisc = new ConfigureMisc; return;
	 	  default: 					return;
	}
}

void ConfigureNotebook::OnButtonPressed(wxCommandEvent& event)
{
int id = event.GetId();

if ( id == XRCID("UDCAddToolBtn") || id == XRCID("UDCEditToolBtn") || id == XRCID("UDCDeleteToolBtn") )	// The New/Edit/Delete-Tool buttons
	{	if ( (UDCPage==UDCNew && id == XRCID("UDCAddToolBtn")) || (UDCPage==UDCEdit && id == XRCID("UDCEditToolBtn"))
							 || (UDCPage==UDCDelete && id == XRCID("UDCDeleteToolBtn")) ) 	return;						// Already there
		
		wxPanel* placeholder = (wxPanel*)FindWindow( wxT("UDCPlaceholderPanel") );
		panel->Destroy(); 
		if (  id == XRCID("UDCAddToolBtn") )
						{	panel = (wxPanel*)wxXmlResource::Get()->LoadPanel( placeholder, wxT("UDCAddTool") ); UDCPage = UDCNew; }
		  else if ( id == XRCID("UDCEditToolBtn") )
						{	panel = (wxPanel*)wxXmlResource::Get()->LoadPanel( placeholder, wxT("UDCEditTool") ); UDCPage = UDCEdit; }
			 else	{	panel = (wxPanel*)wxXmlResource::Get()->LoadPanel( placeholder, wxT("UDCDeleteTool") ); UDCPage = UDCDelete; }
		
		panel->Centre(); InitPage( UDCommand ); return;
	}

if ( id == XRCID("UDCNewSubmenu") )																					// The Add LaunchCommand subfolder button
	{	bool dup=0; wxString newfolderlabel; size_t current=UDCCombo->GetSelection();		// Get the index of the parent folder
		if ( FolderArray->Item(current)->Subs > 25 )
			{	wxMessageBox(_("Sorry, there's no room for another submenu here\nI suggest you try putting it elsewhere") ); return; }

		do
			{	wxString currentfolder = UDCCombo->GetValue(); currentfolder.Trim(false); if ( currentfolder.IsEmpty() ) return;
				wxString msg(wxT("Enter the name of the new submenu to add to "));
				newfolderlabel = wxGetTextFromUser( msg+currentfolder );  if ( newfolderlabel.IsEmpty() ) return;
				
				for ( size_t c=0;  c < FolderArray->Item( current )->Subs; ++c )							// Check the name isn't already taken
						if ( FolderArray->Item( current+1+c )->Label == newfolderlabel )  { dup=1; break; }
				if ( dup )
					{	wxMessageDialog dialog(this, _("Sorry, a menu with this name already exists\n                Try again?"), _("Oops!"), wxYES_NO | wxICON_ERROR );
						if ( dialog.ShowModal() != wxID_YES )        return;
    				}
			}
		 while (dup);
		
		UDToolFolder* newfolder = new UDToolFolder;															// Make a UDToolFolder to hold the new folder
		newfolder->Label = newfolderlabel;
		

		FolderArray->Item( current )->Subs++; 																			// Inc parent's count of subfolders
		if ( current == FolderArray->GetCount()-1 )   FolderArray->Add( newfolder );				// If parent is the last item, Append rather than Insert
		  else   FolderArray->Insert( newfolder, current+1 );														// Otherwise InsertBefore current + 1

		UDCCombo->Clear(); size_t count=0; FillUDCMenusCombo( count, 0 ); 						// The combo doesn't have an Insert, so empty & refill
		UDCCombo->SetSelection(current+1); dirty[UDCommand] = true; return;					// Set as the selection
	}

if ( id == XRCID("UDCDeleteSubmenu") )																				// The Delete LaunchCommand subfolder button
	{	size_t DelFolder = UDCCombo->GetSelection();
		if ( ! DelFolder )
			{	wxMessageDialog dialog(this, _("Sorry, you're not allowed to delete the root menu"), _("Oops!"), wxICON_ERROR );
				dialog.ShowModal(); return;
			}

		wxString msg, currentfolder = UDCCombo->GetValue(); currentfolder.Trim(false);
		msg.Printf(_("Delete menu \"%s\" and all its contents?"), currentfolder.c_str() );
		wxMessageDialog dialog(this, msg, _("Are you sure?"), wxYES_NO | wxICON_QUESTION  );
		if ( dialog.ShowModal() != wxID_YES )		return;
		
		size_t entrycount=0, parentfolder=0; bool done=false;
		for ( size_t n=0; n < DelFolder; ++n )
			{	entrycount += FolderArray->Item(n)->Entries;														// For every folder below the one 2b deleted, add its entries to entrycount
				if ( ! done )																												// If we haven't found the parent-folder yet,
					{	if	( (parentfolder + FolderArray->Item(n)->Subs) < DelFolder )						//   & this isn't it,
									parentfolder += FolderArray->Item(n)->Subs;									//		 add this folder's subfolders to the folder-count
						  else done = true;
					}
			}
																				// We now have enough info to do the deletion
		LMT->DeleteMenu( DelFolder, entrycount );
		FolderArray->Item(parentfolder)->Subs--;																		// Dec parent's sub-count
		UDCCombo->Clear(); size_t count=0; FillUDCMenusCombo( count, 0 ); 						// Instead of calculating how many menus to delete, it's easier to empty & refill
		dirty[UDCommand] = true; return;
	}

if ( id == XRCID("UDCAddBtn") )																							// The UDC 'Add new tool' button
	{	wxString command = UDCFilepath->GetValue();
		if ( command.IsEmpty() )  return;
		
		int folder = UDCCombo->GetSelection();  if ( folder==-1 ) folder = 0;							// Add to this menu.  If not found, use root	
			
		UDCFilepath->SetValue(wxT(""));
		wxString label = ((wxTextCtrl*)FindWindow( wxT("UDCLabel") ))->GetValue();
		if ( label.IsEmpty() )	label = command;																			// Suitable for simple tools like df -h
		  else ((wxTextCtrl*)FindWindow( wxT("UDCLabel") ))->SetValue( wxT("") );
		bool interminal = ((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->IsChecked();
		bool terminalpersists = ((wxCheckBox*)FindWindow( wxT("UDCTerminalPersists") ))->IsChecked();
		
		wxString commandstart = command.BeforeFirst(wxT(' '));												// Get the real command, without parameters etc
		if ( commandstart.GetChar(0) == wxT('~') ) commandstart = wxGetHomeDir() + commandstart.Mid(1);
		if ( ! parent->TestExistence( commandstart, true ) )														// Check there is such a command
			{	wxString msg;	
				if ( wxIsAbsolutePath(	commandstart ) )	msg.Printf(_("I can't find an executable command \"%s\".\nContinue anyway?"), commandstart.c_str() );
					else	msg.Printf(_("I can't find an executable command \"%s\" in your PATH.\nContinue anyway?"), commandstart.c_str() );
				wxMessageDialog dialog(this, msg, _("Are you sure?"), wxYES_NO | wxICON_QUESTION  );
				if ( dialog.ShowModal() != wxID_YES )		return;
			}
		
		QueryWrapWithShell( command );																					// Wraps command with sh -c if appropriate e.g. if pipes, or ~/foo
		bool result = LMT->AddTool( folder, command, interminal, terminalpersists, label );		// Do the addition
		
		if ( dirty[UDCommand] != true ) dirty[UDCommand] = result;										// If there was no previous dirt, store this result
		return;
	}

if ( id == XRCID("UDCEditBtn") )																							// The UDC 'Edit this tool' button
	{	if ( ! editing )																													// editing flags whether we're switching the edit on or off
			{	((wxButton*)FindWindow( wxT("UDCEditBtn") ))->SetLabel(_(" Click When Finished "));
				UDCCombo->Enable(false);																					// Switch off the combo so that the selection can't be changed
				UDCFilepath->Enable(true); ((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->Enable(true); // Make editable
				editing = true; return;
			}
											// Otherwise, this is the 2nd press to finish the editing, so process and revert the UI changes
		((wxButton*)FindWindow(wxT("UDCEditBtn")))->SetLabel(_("  Edit This Command  "));
		UDCCombo->Enable(true);	UDCFilepath->Enable(false); ((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->Enable(false);

		wxString newfpath = UDCFilepath->GetValue(); 
		if ( newfpath.IsEmpty() || m_sel == wxNOT_FOUND ) { editing = false; return; }
		QueryWrapWithShell( newfpath );																					// Wraps command with sh -c if appropriate e.g. if pipes, or ~/foo
		if ( newfpath != ToolArray->Item( m_sel )->Filepath
					|| ((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->IsChecked() != ToolArray->Item( m_sel )->InTerminal
					|| ((wxCheckBox*)FindWindow( wxT("UDCTerminalPersists") ))->IsChecked() != ToolArray->Item( m_sel )->TerminalPersists )
			dirty[UDCommand] = true;																							// If the data's changed . . .
			
		ToolArray->Item( m_sel )->Filepath = newfpath;
		ToolArray->Item( m_sel )->InTerminal = ((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->IsChecked();
		ToolArray->Item( m_sel )->TerminalPersists = ((wxCheckBox*)FindWindow( wxT("UDCTerminalPersists") ))->IsChecked();

		editing = false; return;
	}

if ( id == XRCID("UDCDeleteBtn") )																						// The UDC 'Delete this tool' button
	{	int sel = UDCCombo->GetSelection(); if ( sel == -1 ) return;
		wxString msg; msg.Printf(_("Delete command \"%s\"?"), UDCCombo->GetStringSelection().c_str() );
		wxMessageDialog dialog(this, msg, _("Are you sure?"), wxYES_NO | wxICON_QUESTION  );
		if ( dialog.ShowModal() != wxID_YES )		return;
		
		LMT->DeleteTool(sel); 																									// Remove the tool
		UDCCombo->Delete(sel);																								// & from the combo
		if ( ToolArray->GetCount() ) UDCCombo->SetSelection( wxMax(sel-1,0) );
		  else	UDCCombo->SetValue(wxT(""));																		// If none left, need to clear the 'textctrl' bit
		wxCommandEvent event; event.SetInt(0); OnUDCLabelComboSelectionChanged(event);	// Fake an event to update the Filepath
		
		dirty[UDCommand] = true; return;
	}

if ( id == XRCID("UDCBrowse") )																							// The UDCBrowse button
	{	static bool firsttime = true;					// The first time we run this, a default dir is needed.  After that, the last used dir is remembered
		wxString startdir; if ( firsttime ) startdir = HOME; firsttime = false;
		#if wxVERSION_NUMBER > 2501
		#if wxVERSION_NUMBER <2700
		wxFileDialog fdlg( this, _("Choose a file"), startdir, wxT(""), wxT("*"), wxOPEN | wxCHANGE_DIR );
		#else
		wxFileDialog fdlg( this, _("Choose a file"), startdir, wxT(""), wxT("*"), wxFD_OPEN | wxFD_CHANGE_DIR );
		#endif //wxVERSION_NUMBER <2700
		#else
		wxFileDialog fdlg( this, _("Choose a file"), startdir, wxT(""), wxT("*"), wxOPEN | wxHIDE_READONLY | wxCHANGE_DIR );
		#endif
		if ( fdlg.ShowModal()  == wxID_OK )  UDCFilepath->SetValue( fdlg.GetPath() ); return ;
	}

if ( id == XRCID("UDCApply") )																								// The UDC 'OK' button
	{	if ( UDCPage==UDCEdit )																								// If this is the 'Edit' module, check the Label hasn't been altered
			{	if ( m_sel == wxNOT_FOUND ) { dirty[UDCommand] = false; return; }
				wxString newstring = UDCCombo->GetValue();
				if (  !newstring.IsEmpty() && newstring != ToolArray->Item(m_sel)->Label )			// If the label has been changed, implement it
					 ToolArray->Item(m_sel)->Label = newstring;
				if (  !newstring.IsEmpty() && newstring != ToolArray->Item(m_sel)->Label )			// If the label has been changed, implement it
					 ToolArray->Item(m_sel)->Label = newstring;
				m_sel = wxNOT_FOUND;
			}
			
		if ( ! dirty[UDCommand] ) return;
		LMT->SaveMenu();																											// Save the new data
		MyFrame::mainframe->AccelList->Init();																			// Now reload shortcut data, in case any UDtool shortcuts were just altered
		LMT->LoadMenu();																											// Now we can reload the UDtools, with any shortcuts correct
		InitPage( UDCommand ); dirty[UDCommand] = false; return;
	}

if ( id == XRCID("UDCCancel") )																								// The UDC 'Cancel' button
	{	if ( UDCPage==UDCNew )  UDCFilepath->Clear();
		if ( ! dirty[UDCommand] ) return;
		LMT->LoadArrays();	// Reload the arrays with data from config, in case of alterations
		InitPage( UDCommand ); dirty[UDCommand] = false; return; 
	}
																																				// Configure Devices, Editors
DeviceManager* pDM = MyFrame::mainframe->Layout->m_notebook->DeviceMan->deviceman; if ( pDM==NULL ) return;
if ( id == XRCID("Devices General") )	return pDM->ConfigureDevices();
if ( id == XRCID("Devices") )	return pDM->ConfigureFixedorUsbDevices(CDT_fixed);
if ( id == XRCID("USB") )	return pDM->ConfigureFixedorUsbDevices(CDT_removable);
if ( id == XRCID("Editors") )	return pDM->ConfigureEditors();

if ( id == XRCID("Display_OK") ) { ConfDisplay->OnOK(); return; }										// Display page stuff
if ( id == XRCID("Display_CANCEL") ) { ConfDisplay->InitialiseData(); ConfDisplay->InsertData(); return; } // Reverts, then refills
if ( id == XRCID("SelectColoursBtn") ) { ConfDisplay->OnSelectColours(); return; } 			// Calls a colour dialog
if ( id == XRCID("TreeFontBtn") ) { ConfDisplay->OnChangeFont(); return; } 
if ( id == XRCID("TreeDefaultFontBtn") ) { ConfDisplay->OnDefaultFont(); return; }
if ( id == XRCID("TBIcons") ) { ConfDisplay->OnConfigureTBIcons(); return; }
if ( id == XRCID("Tooltips") ) { parent->ConfigureTooltips(); return; }

if ( id == XRCID("Terminals_OK") ) { ConfTerminals->OnOK(); return; }								// Terminal page stuff
if ( id == XRCID("Terminals_CANCEL") ) { ConfTerminals->OnCancel(); return; }
if ( id == XRCID("TerminalAccept") || id == XRCID("TerminalCommandAccept") || id == XRCID("ToolsTerminalCommandAccept") ) 
																{ ConfTerminals->OnAccept( id ); return; }
if ( id == XRCID("TerminalFontBtn") ) { ConfTerminals->OnChangeFont(); return; } 
if ( id == XRCID("TerminalDefaultFontBtn") ) { ConfTerminals->OnDefaultFont(); return; }

if ( id == XRCID("Net_OK") ) { ConfNet->OnOK(); return; }													// Network page stuff
if ( id == XRCID("Net_CANCEL") ) { ConfNet->InitialiseData(); ConfNet->InsertData(); return; }
if ( id == XRCID("Net_Add") ) { ConfNet->AddServer(); return; }
if ( id == XRCID("Net_Delete") ) { ConfNet->DeleteServer(); return; }

if ( id == XRCID("Misc_OK") ) { ConfMisc->OnOK(); return; }												// Misc page stuff
if ( id == XRCID("Misc_CANCEL") ) { ConfMisc->InitialiseData(); ConfMisc->InsertData(); return; }
if ( id == XRCID("DnD") ) { ConfMisc->OnConfigureDnD(); HelpContext = HCconfigMisc; return; }
if ( id == XRCID("Status") ) { ConfMisc->OnConfigureStatusbar(); return; }
if ( id == XRCID("Export") ) { ConfMisc->OnExportBtn( this ); return; }
}

void ConfigureNotebook::OnUDCLabelComboSelectionChanged(wxCommandEvent& event)
{
if ( UDCCombo==NULL ) return;

int sel = event.GetInt(); 		// I originally used GetSelection() here, but on a slower machine the selection didn't have time to update
if ( sel == -1 ) { UDCFilepath->SetValue(wxT("")); return; }												// No data
UDCFilepath->SetValue( ToolArray->Item(sel)->Filepath );

if ( UDCPage==UDCEdit )																									// As Delete doesn't have a terminal checkbox
	{	((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->SetValue( ToolArray->Item(sel)->InTerminal );	// Is this command run in a terminal?
		((wxCheckBox*)FindWindow( wxT("UDCTerminalPersists") ))->SetValue( ToolArray->Item(sel)->TerminalPersists );	// If so, does this persist or self-close?
		m_sel = sel;																													// Store the selection: if the label gets edited, we'll not be able to find it again
	}
}

void ConfigureNotebook::OnUpdateUI(wxUpdateUIEvent& WXUNUSED(event))
{
bool enable;

switch( GetSelection() )
	{	case UDCommand:	{size_t count = ToolArray->GetCount();
											FindWindow( wxT("UDCEditToolBtn") )->Enable(count && UDCPage!=UDCEdit);	// Disable Edit/Delete buttons if nothing to process, or if they're already 'it'
											FindWindow( wxT("UDCDeleteToolBtn"))->Enable(count && UDCPage!=UDCDelete);
											FindWindow( wxT("UDCAddToolBtn"))->Enable(UDCPage!=UDCNew);
											
											if ( UDCPage == UDCNew )
												{	FindWindow( wxT("UDCTerminalPersists") )->Enable( ((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->IsChecked() );	// Can't persist if no terminal!
													enable = dirty[UDCommand] || ! UDCFilepath->GetValue().IsEmpty();
												}
											  else enable = dirty[UDCommand];
											
											
											if ( UDCPage==UDCEdit )																			// If this is the 'Edit' module:
												{	int sel = m_sel; if ( sel == wxNOT_FOUND ) return;
													if (  UDCCombo->GetValue() != ToolArray->Item( sel )->Label )	// check the Label hasn't been altered
														dirty[UDCommand] = true;
													FindWindow( wxT("UDCTerminalPersists") )->Enable( ((wxCheckBox*)FindWindow( wxT("UDCTerminal") ))->IsChecked() );	// Can't persist if no terminal!
													if ( ! editing )												
														{	FindWindow( wxT("UDCTerminal") )->Disable();						// We don't want the terminal checks altered when we're not editing
															FindWindow( wxT("UDCTerminalPersists") )->Disable();
														}
													enable = dirty[UDCommand] && ! editing;											// We don't want Apply pressed while we're in the middle of editing a command
												}

											FindWindow( wxT("UDCApply") )->Enable( enable ); FindWindow( wxT("UDCCancel") )->Enable( enable );
											return;}

		case UDShortcuts:		enable = MyFrame::mainframe->AccelList->dirty
											#if defined(__WXGTK20__)
																				|| F10check->IsChecked() != USE_F10_KEY
											#endif
																																							;				// This is the ';' from above. It needs 2b outside the ifdef
											((wxButton*)FindWindow( wxT("UDSApply") ))->Enable( enable );
											((wxButton*)FindWindow( wxT("UDSCancel") ))->Enable( enable ); 

		case ConfigDisplay:	ConfDisplay->OnUpdateUI(); return;
		case ConfigTerminals:	ConfTerminals->OnUpdateUI(); return;
		case ConfigNet:			ConfNet->OnUpdateUI(); return;
		case ConfigMisc:		ConfMisc->OnUpdateUI(); return;

		 default: 					return;
	}

}



void ConfigureNotebook::FillUDCMenusCombo( size_t &item, size_t indent )	// Recursively fill the combo, indenting according to sub-ness of the menu
{
if ( item >= FolderArray->GetCount() )   return;

wxString label = FolderArray->Item(item)->Label;
wxString folder( wxT(' '), indent * 5 ); folder += label;														// Prepend an appropriate indent for this (sub)folder
UDCCombo->Append( folder );

size_t Subs = FolderArray->Item(item)->Subs;
for ( size_t c=0;  c < Subs; ++c )  FillUDCMenusCombo( ++item, indent+1 );						// Do children by recursion, indenting by an extra 1
}

bool ConfigureNotebook::Dirty()		// Check there's no unsaved data
{
const wxChar* PageNames[] = { _("Launch Tool Menu"), _("Configure Shortcuts"), _("DevicesEditors"), _("Display"), _("Terminals"),  _("Networks"), _("Misc") };

dirty[UDShortcuts] = MyFrame::mainframe->AccelList->dirty											// AccelList holds this info, as it's shared with the standalone version
#if defined(__WXGTK20__)
											|| F10check->IsChecked() != USE_F10_KEY
#endif
																														;				// This is the ';' from above. It needs 2b outside the ifdef
dirty[ConfigDisplay] = ConfDisplay->dirty;																		// ConfigureDisplay etc hold their own dirt
dirty[ConfigTerminals] = ConfTerminals->dirty;
dirty[ConfigNet] = ConfNet->dirty;
dirty[ConfigMisc] = ConfMisc->dirty;

for ( int n=0; n < NotSelected; ++n )
	if ( dirty[n] )																													// If this page hasn't been saved, check
		{	wxString msg;
			msg.Printf(_("There are unsaved changes on the \"%s\" page.\n                           Really close?"), PageNames[n] );
			wxMessageDialog dialog(this, msg, _("Are you sure?"), wxYES_NO | wxICON_QUESTION  );
			if ( dialog.ShowModal() != wxID_YES )		
				{	SetSelection(n); return true; }																	// If requested, set the appropriate page & signal to redo the config dialog
		}
		
return false;		
}

void ConfigureNotebook::OnPageChanged(wxNotebookEvent& event)
{
enum pagename page = (pagename)event.GetSelection();
if ( currentpage == page ) return;																					// ISQ
switch( page )
	{	case UDCommand:		currentpage = page; HelpContext = HCconfigUDC; break;
		case UDShortcuts:		currentpage = page; HelpContext = HCconfigShortcuts; break;
		case ConfigDevicesEditors: currentpage = page; HelpContext = HCconfigDevices;  break;
		case ConfigDisplay:	currentpage = page; HelpContext = HCconfigDisplay; ConfDisplay->Init( GetPage( page ) ); break;
		case ConfigTerminals:	currentpage = page; HelpContext = HCconfigterminal; ConfTerminals->Init( GetPage( page ) ); break;
		case ConfigNet:			currentpage = page; HelpContext = HCconfigNetworks; ConfNet->Init( GetPage( page ) ); break;
		case ConfigMisc:		currentpage = page; HelpContext = HCconfigMisc; ConfMisc->Init( GetPage( page ) ); break;

		 default: 						currentpage = NotSelected; HelpContext = HCunknown; break;
	}
}

BEGIN_EVENT_TABLE(ConfigureNotebook, wxNotebook)
	EVT_NOTEBOOK_PAGE_CHANGED(-1, ConfigureNotebook::OnPageChanged)
	EVT_COMBOBOX(XRCID("UDCLabelCombo"),ConfigureNotebook::OnUDCLabelComboSelectionChanged)
	EVT_BUTTON(-1, ConfigureNotebook::OnButtonPressed)
	EVT_UPDATE_UI(-1, ConfigureNotebook::OnUpdateUI)
END_EVENT_TABLE()
//-------------------------------------------------------------------------------------------------------------------------------------------

void ConfigureDisplay::InitialiseData()	// Fill the temp vars with the old data. Used initially and also to revert on Cancel
{
tempTREE_INDENT = TREE_INDENT;
tempSHOW_TREE_LINES = SHOW_TREE_LINES;
tempTREE_SPACING = TREE_SPACING;
tempSHOWHIDDEN = SHOWHIDDEN;
tempTREAT_SYMLINKTODIR_AS_DIR = TREAT_SYMLINKTODIR_AS_DIR;
tempUSE_STRIPES = USE_STRIPES;
tempSTRIPE_0 = STRIPE_0;
tempSTRIPE_1 = STRIPE_1;
tempSHOW_TB_TEXTCTRL = SHOW_TB_TEXTCTRL;
tempASK_ON_TRASH = ASK_ON_TRASH;
tempASK_ON_DELETE = ASK_ON_DELETE;
tempTRASHCAN = TRASHCAN;
tempTREE_FONT = TREE_FONT;
tempCHOSEN_TREE_FONT = CHOSEN_TREE_FONT;
tempUSE_DEFAULT_TREE_FONT = USE_DEFAULT_TREE_FONT;
}

void ConfigureDisplay::InsertData()
{
((wxCheckBox*)page->FindWindow(wxT("TREE_LINES")))->SetValue( tempSHOW_TREE_LINES );
((wxSpinCtrl*)page->FindWindow(wxT("TREE_INDENT")))->SetValue( tempTREE_INDENT );
((wxSpinCtrl*)page->FindWindow(wxT("TREE_SPACING")))->SetValue( tempTREE_SPACING );
((wxCheckBox*)page->FindWindow(wxT("SHOWHIDDEN")))->SetValue( tempSHOWHIDDEN );
((wxCheckBox*)page->FindWindow(wxT("TREAT_SYMLINKTODIR_AS_DIR")))->SetValue( tempTREAT_SYMLINKTODIR_AS_DIR );
((wxCheckBox*)page->FindWindow(wxT("USE_STRIPES")))->SetValue( tempUSE_STRIPES );
((wxCheckBox*)page->FindWindow(wxT("SHOW_TB_TEXTCTRL")))->SetValue( tempSHOW_TB_TEXTCTRL );
((wxCheckBox*)page->FindWindow(wxT("ASK_ON_TRASH")))->SetValue( tempASK_ON_TRASH );
((wxCheckBox*)page->FindWindow(wxT("ASK_ON_DELETE")))->SetValue( tempASK_ON_DELETE );
((wxTextCtrl*)page->FindWindow(wxT("TRASHCAN")))->SetValue( tempTRASHCAN );

fonttext = (wxTextCtrl*)page->FindWindow(wxT("CurrentFontText"));
fonttext->SetValue( wxT("Selected tree Font") ); fonttext->SetFont( tempCHOSEN_TREE_FONT );
defaultfonttext = (wxTextCtrl*)page->FindWindow(wxT("DefaultFontText"));
defaultfonttext->SetValue( wxT("Default tree Font") );
}

void ConfigureDisplay::OnSelectColours()
{
ChooseStripesDlg dlg;
wxXmlResource::Get()->LoadDialog(&dlg, page, wxT("ChooseStripesDlg"));
dlg.Init( tempSTRIPE_0, tempSTRIPE_1 );
if ( dlg.ShowModal() == wxID_OK ) 
	{ tempSTRIPE_0 = dlg.Col0; tempSTRIPE_1 = dlg.Col1; }
}

void ConfigureDisplay::OnConfigureTBIcons()	// Add/Edit the 'bookmark' icons in the dirview toolbar
{
ArrayofDirCtrlTBStructs* DirCtrlTBinfoarray = MyFrame::mainframe->configure->DirCtrlTBinfoarray;
ToolbarIconsDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg,MyFrame::mainframe, wxT("ToolbarIconsDialog"));
dlg.Init( DirCtrlTBinfoarray );
dlg.InitDialog();
dlg.ShowModal();

dlg.Save();																			// Save data, whether on not altered. It's really not worth checking

for ( int n = (int)DirCtrlTBinfoarray->GetCount();  n > 0; --n  )	// Now delete the old data, reload
	{ DirCtrlTBStruct* item = DirCtrlTBinfoarray->Item(n-1); delete item; DirCtrlTBinfoarray->RemoveAt(n-1); }
ToolbarIconsDialog::LoadDirCtrlTBinfo( DirCtrlTBinfoarray );
MyFrame::mainframe->RefreshAllTabs( NULL, true );				//  and redisplay
}

void ConfigureDisplay::OnUpdateUI()
{
		// I'm null-testing the first control, as there were occasional segs when the notebook page changed; probably a race condition somehow
if ( page == NULL ) return;
wxSpinCtrl* spin = (wxSpinCtrl*)page->FindWindow(wxT("TREE_INDENT")); if ( spin==NULL ) return;
tempTREE_INDENT = spin->GetValue();
tempSHOW_TREE_LINES = ((wxCheckBox*)page->FindWindow(wxT("TREE_LINES")))->IsChecked();
tempTREE_SPACING = ((wxSpinCtrl*)page->FindWindow(wxT("TREE_SPACING")))->GetValue();
tempSHOWHIDDEN = ((wxCheckBox*)page->FindWindow(wxT("SHOWHIDDEN")))->IsChecked();
tempTREAT_SYMLINKTODIR_AS_DIR = ((wxCheckBox*)page->FindWindow(wxT("TREAT_SYMLINKTODIR_AS_DIR")))->IsChecked();
tempUSE_STRIPES = ((wxCheckBox*)page->FindWindow(wxT("USE_STRIPES")))->IsChecked();
tempSHOW_TB_TEXTCTRL = ((wxCheckBox*)page->FindWindow(wxT("SHOW_TB_TEXTCTRL")))->IsChecked();
tempASK_ON_TRASH = ((wxCheckBox*)page->FindWindow(wxT("ASK_ON_TRASH")))->IsChecked();
tempASK_ON_DELETE = ((wxCheckBox*)page->FindWindow(wxT("ASK_ON_DELETE")))->IsChecked();
tempTRASHCAN = ((wxTextCtrl*)page->FindWindow(wxT("TRASHCAN")))->GetValue();

page->FindWindow(wxT("SelectColoursBtn"))->Enable( tempUSE_STRIPES );

dirty = fontchanged || (SHOW_TREE_LINES!=tempSHOW_TREE_LINES) || (TREE_INDENT!=tempTREE_INDENT) || (tempTREE_SPACING!=TREE_SPACING)
			|| (tempSHOWHIDDEN!=SHOWHIDDEN) || (tempTREAT_SYMLINKTODIR_AS_DIR!=TREAT_SYMLINKTODIR_AS_DIR)
			||(tempUSE_STRIPES!=USE_STRIPES) || (tempSTRIPE_0!=STRIPE_0) || (tempSTRIPE_1!=STRIPE_1) || (tempSHOW_TB_TEXTCTRL!=SHOW_TB_TEXTCTRL)
			|| (tempASK_ON_TRASH!=ASK_ON_TRASH) || (tempASK_ON_DELETE!=ASK_ON_DELETE) || (tempTRASHCAN!=TRASHCAN);

fonttext->Enable( ! tempUSE_DEFAULT_TREE_FONT );
defaultfonttext->Enable( tempUSE_DEFAULT_TREE_FONT ); page->FindWindow(wxT("TreeDefaultFontBtn"))->Enable( ! tempUSE_DEFAULT_TREE_FONT );
page->FindWindow(wxT("Display_OK"))->Enable( dirty );
page->FindWindow(wxT("Display_CANCEL"))->Enable( dirty );
}

void ConfigureDisplay::OnChangeFont()
{
wxFontData data;
data.SetInitialFont(tempCHOSEN_TREE_FONT);
		
wxFontDialog dialog(page, data);
if (dialog.ShowModal() == wxID_OK)
	{	wxFontData retData = dialog.GetFontData();
		tempTREE_FONT = tempCHOSEN_TREE_FONT = retData.GetChosenFont();
		fonttext->SetFont( tempCHOSEN_TREE_FONT );		
		fontchanged = true; tempUSE_DEFAULT_TREE_FONT = false;
	}
}

void ConfigureDisplay::OnDefaultFont()
{
tempTREE_FONT = defaultfonttext->GetFont();
tempUSE_DEFAULT_TREE_FONT = true; fontchanged = (tempUSE_DEFAULT_TREE_FONT!=USE_DEFAULT_TREE_FONT);
}

void ConfigureDisplay::OnCancel()
{
InitialiseData(); InsertData();
fontchanged = false;
}

void ConfigureDisplay::OnOK()
{
SHOW_TREE_LINES = tempSHOW_TREE_LINES;
TREE_INDENT = tempTREE_INDENT;
TREE_SPACING = tempTREE_SPACING;
SHOWHIDDEN = tempSHOWHIDDEN;
TREAT_SYMLINKTODIR_AS_DIR = tempTREAT_SYMLINKTODIR_AS_DIR;
USE_STRIPES = tempUSE_STRIPES;
STRIPE_0 = tempSTRIPE_0; STRIPE_1 = tempSTRIPE_1;
SHOW_TB_TEXTCTRL = tempSHOW_TB_TEXTCTRL;
ASK_ON_TRASH = tempASK_ON_TRASH;
ASK_ON_DELETE = tempASK_ON_DELETE;
TRASHCAN = tempTRASHCAN;

if ( fontchanged ) 
	{	TREE_FONT = tempTREE_FONT; CHOSEN_TREE_FONT = tempCHOSEN_TREE_FONT;
		USE_DEFAULT_TREE_FONT = tempUSE_DEFAULT_TREE_FONT;
		fontchanged = false;
		MyFrame::mainframe->RefreshAllTabs( &TREE_FONT );
	}
 else
	MyFrame::mainframe->RefreshAllTabs( NULL );

Save();
}

//static 
void ConfigureDisplay::LoadDisplayConfig()
{
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)  return; 

config->Read( wxT("/Misc/Display/SHOW_TREE_LINES"), &SHOW_TREE_LINES, 1 );
TREE_INDENT = (int)config->Read( wxT("/Misc/Display/TREE_INDENT"), 10l );
TREE_SPACING = (size_t)config->Read( wxT("/Misc/Display/TREE_SPACING"), 10l );
config->Read( wxT("/Misc/Display/SHOWHIDDEN"), &SHOWHIDDEN, 1 );
config->Read( wxT("/Misc/Display/SYMLINKTODIR"), &TREAT_SYMLINKTODIR_AS_DIR, 1 );
config->Read( wxT("/Misc/Display/USE_STRIPES"), &USE_STRIPES, 0 );
wxColour defaultcol = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW);
long defaultcolour = (int)(defaultcol.Blue() << 16) + (int)(defaultcol.Green() << 8)  + defaultcol.Red();
long stripe0 = config->Read( wxT("/Misc/Display/STRIPE_0"), defaultcolour );
long stripe1 = config->Read( wxT("/Misc/Display/STRIPE_1"), defaultcolour );
STRIPE_0.Set( (unsigned char)stripe0, (unsigned char)(stripe0 >> 8), (unsigned char)(stripe0 >> 16) );
STRIPE_1.Set( (unsigned char)stripe1, (unsigned char)(stripe1 >> 8), (unsigned char)(stripe1 >> 16) );
config->Read( wxT("/Misc/Display/SHOW_TB_TEXTCTRL"), &SHOW_TB_TEXTCTRL, 1 );
config->Read( wxT("/Misc/Display/ASK_ON_TRASH"), &ASK_ON_TRASH, 1 );
config->Read( wxT("/Misc/Display/ASK_ON_DELETE"), &ASK_ON_DELETE, 1 );
config->Read( wxT("/Misc/Display/TRASHCAN"), &TRASHCAN );

config->Read( wxT("/Misc/Display/TREE_FONT/UseSystemDefault"), &USE_DEFAULT_TREE_FONT, true );
CHOSEN_TREE_FONT.SetPointSize( (int)config->Read( wxT("/Misc/Display/TREE_FONT/pointsize"), 14l ) );
CHOSEN_TREE_FONT.SetFamily( (wxFontFamily)config->Read( wxT("/Misc/Display/TREE_FONT/family"),  (long)wxFONTFAMILY_DEFAULT ) );
CHOSEN_TREE_FONT.SetStyle( (int)config->Read( wxT("/Misc/Display/TREE_FONT/style"),  (long)wxFONTSTYLE_NORMAL ) );
CHOSEN_TREE_FONT.SetWeight( (wxFontWeight)config->Read( wxT("/Misc/Display/TREE_FONT/weight"),  (long)wxFONTWEIGHT_NORMAL ) );
bool under; config->Read( wxT("/Misc/Display/TREE_FONT/underlined"), &under, false ); CHOSEN_TREE_FONT.SetUnderlined( under );
wxString face; config->Read( wxT("/Misc/Display/TREE_FONT/facename"), &face ); CHOSEN_TREE_FONT.SetFaceName( face);

if ( USE_DEFAULT_TREE_FONT || ! CHOSEN_TREE_FONT.Ok() ) 
			TREE_FONT = wxSystemSettings::GetFont( wxSYS_DEFAULT_GUI_FONT  );
  else	TREE_FONT = CHOSEN_TREE_FONT;
}

void ConfigureDisplay::Save()
{
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)  return;

config->Write( wxT("/Misc/Display/SHOW_TREE_LINES"), SHOW_TREE_LINES );
config->Write( wxT("/Misc/Display/TREE_INDENT"), (long)TREE_INDENT );
config->Write( wxT("/Misc/Display/TREE_SPACING"), (long)TREE_SPACING );
config->Write( wxT("/Misc/Display/SHOWHIDDEN"), SHOWHIDDEN );
config->Write( wxT("/Misc/Display/SYMLINKTODIR"), TREAT_SYMLINKTODIR_AS_DIR );
config->Write( wxT("/Misc/Display/USE_STRIPES"), USE_STRIPES );
if ( ! STRIPE_0.Ok() ) STRIPE_0 = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW);
if ( ! STRIPE_1.Ok() ) STRIPE_1 = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW);
long colour = (int)(STRIPE_0.Blue() << 16) + (int)(STRIPE_0.Green() << 8)  + STRIPE_0.Red();
config->Write( wxT("/Misc/Display/STRIPE_0"), colour );
colour = (int)(STRIPE_1.Blue() << 16) + (int)(STRIPE_1.Green() << 8)  + STRIPE_1.Red();
config->Write( wxT("/Misc/Display/STRIPE_1"), colour );
config->Write( wxT("/Misc/Display/SHOW_TB_TEXTCTRL"), SHOW_TB_TEXTCTRL );
config->Write( wxT("/Misc/Display/ASK_ON_TRASH"), ASK_ON_TRASH );
config->Write( wxT("/Misc/Display/ASK_ON_DELETE"), ASK_ON_DELETE );
config->Write( wxT("/Misc/Display/TRASHCAN"), TRASHCAN );

config->Write( wxT("/Misc/Display/TREE_FONT/UseSystemDefault"), USE_DEFAULT_TREE_FONT );
if ( ! CHOSEN_TREE_FONT.Ok() ) CHOSEN_TREE_FONT = wxSystemSettings::GetFont( wxSYS_DEFAULT_GUI_FONT  );	// Will be needed on first ever Save
config->Write( wxT("/Misc/Display/TREE_FONT/pointsize"), (long)CHOSEN_TREE_FONT.GetPointSize() );
config->Write( wxT("/Misc/Display/TREE_FONT/family"), (long)CHOSEN_TREE_FONT.GetFamily() );
config->Write( wxT("/Misc/Display/TREE_FONT/style"), (long)CHOSEN_TREE_FONT.GetStyle() );
config->Write( wxT("/Misc/Display/TREE_FONT/weight"), (long)CHOSEN_TREE_FONT.GetWeight() );
config->Write( wxT("/Misc/Display/TREE_FONT/underlined"), CHOSEN_TREE_FONT.GetUnderlined() );
config->Write( wxT("/Misc/Display/TREE_FONT/facename"), CHOSEN_TREE_FONT.GetFaceName() );

config->Flush();
}
//-------------------------------------------------------------------------------------------------------------------------------------------
void ConfigureTerminals::InitialiseData()	// Fill the temp vars with the old data. Used initially and also to revert on Cancel
{
tempTERMINAL = TERMINAL;
tempTERMINAL_COMMAND = TERMINAL_COMMAND.Left( TERMINAL_COMMAND.Len()-1 );	// We don't want the terminal ' ' here
tempTOOLS_TERMINAL_COMMAND = TOOLS_TERMINAL_COMMAND.Left( TOOLS_TERMINAL_COMMAND.Len()-1 );
tempTERMEM_PROMPTFORMAT = TERMEM_PROMPTFORMAT;
tempTERMINAL_FONT = TERMINAL_FONT;
tempCHOSEN_TERMINAL_FONT = CHOSEN_TERMINAL_FONT;
tempUSE_DEFAULT_TERMINALEM_FONT = USE_DEFAULT_TERMINALEM_FONT;
tempPossibleTerminals = PossibleTerminals;
tempPossibleTerminalCommands = PossibleTerminalCommands;
tempPossibleToolsTerminalCommands = PossibleToolsTerminalCommands;
}

void ConfigureTerminals::InsertData()
{
combo1 = (wxComboBox*)page->FindWindow(wxT("TERMINAL")); combo1->Clear();
combo2 = (wxComboBox*)page->FindWindow(wxT("TERMINAL_COMMAND")); combo2->Clear();
combo3 = (wxComboBox*)page->FindWindow(wxT("TOOLS_TERMINAL_COMMAND")); combo3->Clear();

for ( size_t n=0; n < tempPossibleTerminals.GetCount(); ++n ) combo1->Append( tempPossibleTerminals.Item(n) );
combo1->SetStringSelection( tempTERMINAL );

for ( size_t n=0; n < tempPossibleTerminalCommands.GetCount(); ++n ) combo2->Append( tempPossibleTerminalCommands.Item(n) );
combo2->SetStringSelection( tempTERMINAL_COMMAND );

for ( size_t n=0; n < tempPossibleToolsTerminalCommands.GetCount(); ++n ) combo3->Append( tempPossibleToolsTerminalCommands.Item(n) );
combo3->SetStringSelection( tempTOOLS_TERMINAL_COMMAND );

((wxTextCtrl*)page->FindWindow(wxT("TerminalText")))->Clear();
((wxTextCtrl*)page->FindWindow(wxT("TerminalCommandText")))->Clear();
((wxTextCtrl*)page->FindWindow(wxT("ToolsTerminalCommandText")))->Clear();
((wxTextCtrl*)page->FindWindow(wxT("TERMEM_PROMPTFORMAT")))->SetValue( tempTERMEM_PROMPTFORMAT );

fonttext = (wxTextCtrl*)page->FindWindow(wxT("CurrentFontText"));
fonttext->SetValue( wxT("Selected terminal emulator Font") ); fonttext->SetFont( tempCHOSEN_TERMINAL_FONT );
defaultfonttext = (wxTextCtrl*)page->FindWindow(wxT("DefaultFontText"));
defaultfonttext->SetValue( wxT("Default Font") );
}

void ConfigureTerminals::OnUpdateUI()
{
if ( page == NULL || combo1 == NULL ) return;		// Because we've not finished setup yet
if ( combo1->GetCount() ) tempTERMINAL = combo1->GetStringSelection();
if ( combo2->GetCount() ) tempTERMINAL_COMMAND = combo2->GetStringSelection();
if ( combo3->GetCount() ) tempTOOLS_TERMINAL_COMMAND = combo3->GetStringSelection();
tempTERMEM_PROMPTFORMAT = ((wxTextCtrl*)page->FindWindow(wxT("TERMEM_PROMPTFORMAT")))->GetValue();

dirty =  fontchanged || (tempTERMINAL!=TERMINAL) || (tempTERMINAL_COMMAND!=TERMINAL_COMMAND.Left( TERMINAL_COMMAND.Len()-1 ))
			|| (tempTOOLS_TERMINAL_COMMAND!=TOOLS_TERMINAL_COMMAND.Left( TOOLS_TERMINAL_COMMAND.Len()-1 )) || (tempTERMEM_PROMPTFORMAT!=TERMEM_PROMPTFORMAT)
			|| (tempPossibleTerminals!=PossibleTerminals) || (tempPossibleTerminalCommands!=PossibleTerminalCommands) || (tempPossibleToolsTerminalCommands!=PossibleToolsTerminalCommands);

fonttext->Enable( ! tempUSE_DEFAULT_TERMINALEM_FONT );
defaultfonttext->Enable( tempUSE_DEFAULT_TERMINALEM_FONT );
page->FindWindow(wxT("TerminalDefaultFontBtn"))->Enable( ! tempUSE_DEFAULT_TERMINALEM_FONT );
page->FindWindow(wxT("Terminals_OK"))->Enable( dirty );
page->FindWindow(wxT("Terminals_CANCEL"))->Enable( dirty );
}

void ConfigureTerminals::OnAccept( int id )	// When the user has provided a new entry for TERMINAL etc, add it to the combobox
{
wxComboBox* combo; wxTextCtrl* text; wxArrayString* pArray;

if ( id == XRCID("TerminalAccept") )
	{	text = (wxTextCtrl*)page->FindWindow(wxT("TerminalText")); combo = combo1; pArray = &tempPossibleTerminals; }
  else if ( id == XRCID("TerminalCommandAccept") )
	{	text = (wxTextCtrl*)page->FindWindow(wxT("TerminalCommandText")); combo = combo2; pArray = &tempPossibleTerminalCommands; }
  else if ( id == XRCID("ToolsTerminalCommandAccept") )
	{	text = (wxTextCtrl*)page->FindWindow(wxT("ToolsTerminalCommandText")); combo = combo3; pArray = &tempPossibleToolsTerminalCommands; }
  else return;

wxString newstring = text->GetValue(); text->Clear();
newstring.Trim();																							// Though for all but TERMINAL we'll need a terminal ' ', remove any for now for ease of comparison
#if wxVERSION_NUMBER > 2503
	if ( combo->SetStringSelection( newstring ) )  return; 							// Old news
#else
	if ( combo->FindString( newstring ) != wxNOT_FOUND )
		{ combo->SetStringSelection( newstring ); return; }
#endif
pArray->Add( newstring );
combo->Append( newstring );
combo->SetStringSelection( newstring );
}

void ConfigureTerminals::OnChangeFont()
{
wxFontData data;
data.SetInitialFont(tempCHOSEN_TERMINAL_FONT);
		
wxFontDialog dialog(page, data);
if (dialog.ShowModal() == wxID_OK)
	{	wxFontData retData = dialog.GetFontData();
		tempTERMINAL_FONT = tempCHOSEN_TERMINAL_FONT = retData.GetChosenFont();
		fonttext->SetFont( tempCHOSEN_TERMINAL_FONT );		
		fontchanged = true; tempUSE_DEFAULT_TERMINALEM_FONT = false;
	}
}

void ConfigureTerminals::OnDefaultFont()
{
tempTERMINAL_FONT = *wxNORMAL_FONT;//defaultfonttext->GetFont();
tempUSE_DEFAULT_TERMINALEM_FONT = true; fontchanged = (tempUSE_DEFAULT_TERMINALEM_FONT!=USE_DEFAULT_TERMINALEM_FONT);
}

void ConfigureTerminals::OnCancel()
{
InitialiseData(); InsertData();
fontchanged = false;
}

void ConfigureTerminals::OnOK()
{
TERMINAL = tempTERMINAL;
TERMINAL_COMMAND = tempTERMINAL_COMMAND + wxT(" ");	// We've been trimming the terminal ' ' up to now
TOOLS_TERMINAL_COMMAND = tempTOOLS_TERMINAL_COMMAND + wxT(" ");
TERMEM_PROMPTFORMAT = tempTERMEM_PROMPTFORMAT;
if ( fontchanged ) 
	{	TERMINAL_FONT = tempTERMINAL_FONT; CHOSEN_TERMINAL_FONT = tempCHOSEN_TERMINAL_FONT;
		USE_DEFAULT_TERMINALEM_FONT = tempUSE_DEFAULT_TERMINALEM_FONT;
		MyFrame::mainframe->m_tbText->SetFont( TERMINAL_FONT );	// It makes sense to use the same font for the toolbar textctrl too
		MyFrame::mainframe->m_tbText->Refresh();
		fontchanged = false;
	}

PossibleTerminals = tempPossibleTerminals;
PossibleTerminalCommands = tempPossibleTerminalCommands;
PossibleToolsTerminalCommands = tempPossibleToolsTerminalCommands;

Save();
}

//static 
void ConfigureTerminals::LoadTerminalsConfig()
{
wxConfigBase* config = wxConfigBase::Get();	if (config==NULL)  return; 

PossibleTerminals.Clear(); PossibleTerminalCommands.Clear(); PossibleToolsTerminalCommands.Clear();

wxString item = config->Read( wxT("/Misc/Terminal/TERMINAL") ); if ( ! item.IsEmpty() )  TERMINAL = item;
item = config->Read( wxT("/Misc/Terminal/TERMEM_PROMPTFORMAT") ); if ( ! item.IsEmpty() )  TERMEM_PROMPTFORMAT = item;
size_t count = (size_t)config->Read( wxT("/Misc/Terminal/NoOfEntries"), 0l );
for ( size_t n=0; n < count; ++n )
	{	item = config->Read( wxT("/Misc/Terminal/")+CreateSubgroupName(n,count) );  if ( ! item.IsEmpty() )  PossibleTerminals.Add( item ); }

config->Read( wxT("/Misc/Terminal/TERMINAL_FONT/UseSystemDefault"), &USE_DEFAULT_TERMINALEM_FONT, true );
CHOSEN_TERMINAL_FONT.SetPointSize( (int)config->Read( wxT("/Misc/Terminal/TERMINAL_FONT/pointsize"), 14l ) );
CHOSEN_TERMINAL_FONT.SetFamily( (wxFontFamily)config->Read( wxT("/Misc/Terminal/TERMINAL_FONT/family"),  (long)wxFONTFAMILY_DEFAULT ) );
CHOSEN_TERMINAL_FONT.SetStyle( (int)config->Read( wxT("/Misc/Terminal/TERMINAL_FONT/style"),  (long)wxFONTSTYLE_NORMAL ) );
CHOSEN_TERMINAL_FONT.SetWeight( (wxFontWeight)config->Read( wxT("/Misc/Terminal/TERMINAL_FONT/weight"),  (long)wxFONTWEIGHT_NORMAL ) );
bool under; config->Read( wxT("/Misc/Terminal/TERMINAL_FONT/underlined"), &under, false ); CHOSEN_TERMINAL_FONT.SetUnderlined( under );
wxString face; config->Read( wxT("/Misc/Terminal/TERMINAL_FONT/facename"), &face ); CHOSEN_TERMINAL_FONT.SetFaceName( face);

if ( USE_DEFAULT_TERMINALEM_FONT || ! CHOSEN_TERMINAL_FONT.Ok() ) 
			TERMINAL_FONT = wxSystemSettings::GetFont( wxSYS_DEFAULT_GUI_FONT  );
  else	TERMINAL_FONT = CHOSEN_TERMINAL_FONT;

item = config->Read( wxT("/Misc/TerminalCommand/TERMINAL_COMMAND") );
if ( ! item.IsEmpty() )  TERMINAL_COMMAND = item + wxT(" "); // wxConfig doesn't cope well with terminal whitespace, & we need to separate the terminal invocation from the command.
count = (size_t)config->Read( wxT("/Misc/TerminalCommand/NoOfEntries"), 0l );
for ( size_t n=0; n < count; ++n )
	{	item = config->Read( wxT("/Misc/TerminalCommand/")+CreateSubgroupName(n,count) );  if ( ! item.IsEmpty() )  PossibleTerminalCommands.Add( item ); }
	
item = config->Read( wxT("/Misc/ToolsTerminalCommand/TOOLS_TERMINAL_COMMAND") ); if ( ! item.IsEmpty() )  TOOLS_TERMINAL_COMMAND = item + wxT(" ");
count = (size_t)config->Read( wxT("/Misc/ToolsTerminalCommand/NoOfEntries"), 0l );
for ( size_t n=0; n < count; ++n )
	{	item = config->Read( wxT("/Misc/ToolsTerminalCommand/")+CreateSubgroupName(n,count) );  if ( ! item.IsEmpty() )  PossibleToolsTerminalCommands.Add( item ); }
}

//static
void ConfigureTerminals::Save( wxConfigBase* config )
{
if (config==NULL)																							// It'll usually be null, the exception being if coming from ConfigureMisc::OnExportBtn
	 config = wxConfigBase::Get();	if (config==NULL)  return; 

config->Write( wxT("/Misc/Terminal/TERMINAL"), TERMINAL );
config->Write( wxT("/Misc/Terminal/TERMEM_PROMPTFORMAT"), TERMEM_PROMPTFORMAT );
size_t count = PossibleTerminals.GetCount();
config->Write( wxT("/Misc/Terminal/NoOfEntries"), (long)count );
for ( size_t n=0; n < count; ++n ) config->Write( wxT("/Misc/Terminal/")+CreateSubgroupName(n,count), PossibleTerminals.Item(n) );

config->Write( wxT("/Misc/Terminal/TERMINAL_FONT/UseSystemDefault"), USE_DEFAULT_TERMINALEM_FONT );
if ( ! CHOSEN_TERMINAL_FONT.Ok() ) CHOSEN_TERMINAL_FONT = wxSystemSettings::GetFont( wxSYS_DEFAULT_GUI_FONT  );	// Will be needed on first ever Save
config->Write( wxT("/Misc/Terminal/TERMINAL_FONT/pointsize"), (long)CHOSEN_TERMINAL_FONT.GetPointSize() );
config->Write( wxT("/Misc/Terminal/TERMINAL_FONT/family"), (long)CHOSEN_TERMINAL_FONT.GetFamily() );
config->Write( wxT("/Misc/Terminal/TERMINAL_FONT/style"), (long)CHOSEN_TERMINAL_FONT.GetStyle() );
config->Write( wxT("/Misc/Terminal/TERMINAL_FONT/weight"), (long)CHOSEN_TERMINAL_FONT.GetWeight() );
config->Write( wxT("/Misc/Terminal/TERMINAL_FONT/underlined"), CHOSEN_TERMINAL_FONT.GetUnderlined() );
config->Write( wxT("/Misc/Terminal/TERMINAL_FONT/facename"), CHOSEN_TERMINAL_FONT.GetFaceName() );

wxString item = TERMINAL_COMMAND; item.Trim();									// Remove the space after the command, as wxConfig won't cope
config->Write( wxT("/Misc/TerminalCommand/TERMINAL_COMMAND"), item );
count = PossibleTerminalCommands.GetCount();
config->Write( wxT("/Misc/TerminalCommand/NoOfEntries"), (long)count );
for ( size_t n=0; n < count; ++n ) config->Write( wxT("/Misc/TerminalCommand/")+CreateSubgroupName(n,count), PossibleTerminalCommands.Item(n) );

item = TOOLS_TERMINAL_COMMAND; item.Trim();
config->Write( wxT("/Misc/ToolsTerminalCommand/TOOLS_TERMINAL_COMMAND"), item );
count = PossibleToolsTerminalCommands.GetCount();
config->Write( wxT("/Misc/ToolsTerminalCommand/NoOfEntries"), (long)count );
for ( size_t n=0; n < count; ++n ) config->Write( wxT("/Misc/ToolsTerminalCommand/")+CreateSubgroupName(n,count), PossibleToolsTerminalCommands.Item(n) );

config->Flush();
}
//-------------------------------------------------------------------------------------------------------------------------------------------

void ConfigureNet::InitialiseData()	// Fill the temp vars with the old data. Used initially and also to revert on Cancel
{
serverschanged = false;
tempSMBPATH = SMBPATH;
tempSHOWMOUNTFPATH  = SHOWMOUNTFPATH;
tempNFS_SERVERS = NFS_SERVERS;
}

void ConfigureNet::InsertData()
{
ServerList = (wxListBox*)page->FindWindow(wxT("ServerList")); ServerList->Clear(); ServerList->Append( tempNFS_SERVERS );
NewServer = (wxTextCtrl*)page->FindWindow(wxT("NewServer"));
Showmount = (wxTextCtrl*)page->FindWindow(wxT("Showmount")); Showmount->SetValue( tempSHOWMOUNTFPATH );
Samba = (wxTextCtrl*)page->FindWindow(wxT("Samba")); Samba->SetValue( tempSMBPATH );
}

void ConfigureNet::DeleteServer()
{
int sel = ServerList->GetSelection(); if ( sel == wxNOT_FOUND ) return;
wxString server = ServerList->GetString(sel); if ( server.IsEmpty() ) return;

wxMessageDialog ask(MyFrame::mainframe,wxString(_("Delete "))+server+wxT(" ?"), wxString(_("Are you SURE?")), wxYES_NO | wxICON_QUESTION  );
if ( ask.ShowModal() != wxID_YES ) return;

ServerList->Delete(sel); tempNFS_SERVERS.RemoveAt( sel );
serverschanged = true;
}

void ConfigureNet::AddServer()
{
wxString server = NewServer->GetValue(); if ( server.IsEmpty() ) return;
if ( ! IsPlausibleIPaddress( server ) )
	{ wxMessageBox( wxT("That doesn't seem to be a valid ip address") ); NewServer->SetValue( wxEmptyString ); return; }

if ( tempNFS_SERVERS.Index( server ) != wxNOT_FOUND )
	{ wxMessageBox( wxT("That server is already on the list") ); NewServer->SetValue( wxEmptyString ); return; }

ServerList->Append( server ); tempNFS_SERVERS.Add( server );
NewServer->SetValue( wxEmptyString ); serverschanged = true;
}

void ConfigureNet::OnUpdateUI()
{
		// I'm null-testing the first control, as there were occasional segs when the notebook page changed; probably a race condition somehow
if ( page == NULL || Showmount == NULL) return;
tempSHOWMOUNTFPATH = Showmount->GetValue();
tempSMBPATH = Samba->GetValue();

dirty = serverschanged || (SHOWMOUNTFPATH != tempSHOWMOUNTFPATH) || (SMBPATH != tempSMBPATH);

page->FindWindow(wxT("Net_OK"))->Enable( dirty );
page->FindWindow(wxT("Net_CANCEL"))->Enable( dirty );
}

void ConfigureNet::OnOK()
{
SHOWMOUNTFPATH = tempSHOWMOUNTFPATH;
SMBPATH = tempSMBPATH;
NFS_SERVERS = tempNFS_SERVERS;
serverschanged = false;

Save();
}

//static 
void ConfigureNet::LoadNetworkConfig()
{
wxConfigBase* config = wxConfigBase::Get();	if (config==NULL)  return; 

wxString defaultvalue( SHOWMOUNTFPATH ); if ( defaultvalue.IsEmpty() ) defaultvalue = wxT("/usr/sbin/showmount");
config->Read( wxT("/Network/ShowmountFilepath"), &SHOWMOUNTFPATH, defaultvalue );
defaultvalue = wxT("/usr/bin/"); config->Read( wxT("/Network/SMBPATH"), &SMBPATH, defaultvalue );

wxString ip, key( wxT("/Network/IPAddresses/") );
long count = config->Read( key+wxT("Count"), 0l );
for ( long n=0; n < count; ++n )
	{	config->Read( key + CreateSubgroupName( n, count ), &ip );
		if ( ip.IsEmpty() ) continue;
		if ( NFS_SERVERS.Index( ip ) == wxNOT_FOUND )	
				NFS_SERVERS.Add(ip);														// Check for duplication.  If it's OK, add it to the server array
	}
}

void ConfigureNet::Save()
{
wxConfigBase* config = wxConfigBase::Get();	if (config==NULL)  return; 
config->DeleteGroup( wxT("/Network") );

if ( ! SHOWMOUNTFPATH.IsEmpty() )												// Make sure SHOWMOUNTFPATH is sane
	{	if  ( SHOWMOUNTFPATH.Last() == wxFILE_SEP_PATH ) SHOWMOUNTFPATH = SHOWMOUNTFPATH.BeforeLast(wxFILE_SEP_PATH);
		if ( ! SHOWMOUNTFPATH.Contains( wxT("showmount") ) ) SHOWMOUNTFPATH += wxT("/showmount");
		config->Write( wxT("/Network/ShowmountFilepath"), SHOWMOUNTFPATH );
	}

if ( ! SMBPATH.IsEmpty() )																	// Similarly SMBPATH
	{	if  ( SMBPATH.Last() != wxFILE_SEP_PATH )   SMBPATH += wxFILE_SEP_PATH;
		config->Write( wxT("/Network/SMBPATH"), SMBPATH );
	}

wxString key( wxT("/Network/IPAddresses/") );
size_t count = NFS_SERVERS.GetCount();
config->Write( key+wxT("Count"), (long)count );
for ( size_t n=0; n < count; ++n )
	config->Write( key + CreateSubgroupName( n, count ), NFS_SERVERS.Item(n) );

config->Flush();
}

//-------------------------------------------------------------------------------------------------------------------------------------------

void ConfigureMisc::InitialiseData()	// Fill the temp vars with the old data. Used initially and also to revert on Cancel
{
tempMAX_NUMBER_OF_UNDOS = MAX_NUMBER_OF_UNDOS;
tempMAX_DROPDOWN_DISPLAY = MAX_DROPDOWN_DISPLAY;
tempAUTOCLOSE_TIMEOUT = AUTOCLOSE_TIMEOUT;
tempAUTOCLOSE_FAILTIMEOUT = AUTOCLOSE_FAILTIMEOUT;
tempDEFAULT_BRIEFLOGSTATUS_TIME = DEFAULT_BRIEFLOGSTATUS_TIME;
tempWHICH_SU = WHICH_SU;
tempMETAKEYS_MOVE = METAKEYS_MOVE; tempMETAKEYS_COPY = METAKEYS_COPY;
tempMETAKEYS_HARDLINK = METAKEYS_HARDLINK; tempMETAKEYS_SOFTLINK = METAKEYS_SOFTLINK;
tempLINES_PER_MOUSEWHEEL_SCROLL = LINES_PER_MOUSEWHEEL_SCROLL;
tempDnDXTRIGGERDISTANCE = DnDXTRIGGERDISTANCE; tempDnDYTRIGGERDISTANCE = DnDYTRIGGERDISTANCE;

LoadStatusbarWidths( widths ); LoadStatusbarWidths( tempwidths );	// There isn't a global for widths, so load both here
}

void ConfigureMisc::InsertData()
{
((wxSpinCtrl*)page->FindWindow(wxT("MAX_NUMBER_OF_UNDOS")))->SetValue( tempMAX_NUMBER_OF_UNDOS );
((wxSpinCtrl*)page->FindWindow(wxT("MAX_DROPDOWN_DISPLAY")))->SetValue( tempMAX_DROPDOWN_DISPLAY );
((wxSpinCtrl*)page->FindWindow(wxT("AUTOCLOSE_TIMEOUT")))->SetValue( tempAUTOCLOSE_TIMEOUT / 1000 );
((wxSpinCtrl*)page->FindWindow(wxT("AUTOCLOSE_FAILTIMEOUT")))->SetValue( tempAUTOCLOSE_FAILTIMEOUT / 1000 );
((wxSpinCtrl*)page->FindWindow(wxT("DEFAULT_BRIEFLOGSTATUS_TIME")))->SetValue( tempDEFAULT_BRIEFLOGSTATUS_TIME );
Ksu = (wxRadioButton*)page->FindWindow(wxT("Radiok"));
Gksu = (wxRadioButton*)page->FindWindow(wxT("Radiogk"));
Gnsu = (wxRadioButton*)page->FindWindow(wxT("Radiognome"));
	// Check again for available su-ers: it was done in Configure::FindPrograms(), but new ones might have been added since
KDESU_PRESENT = Configure::TestExistence( wxT("kdesu") ); XSU_PRESENT = Configure::TestExistence( wxT("gnomesu") ); GKSU_PRESENT = Configure::TestExistence( wxT("gksu") );
Ksu->Enable( KDESU_PRESENT ); Gksu->Enable( GKSU_PRESENT ); Gnsu->Enable( XSU_PRESENT );
if ( ! (KDESU_PRESENT || GKSU_PRESENT || XSU_PRESENT) ) return;	// If none are present, all  the buttons will be disabled; so don't try to set one
switch( tempWHICH_SU )
	{	case gksu: 			Gksu->SetValue( true ); break;
		case gnomesu: 	Gnsu->SetValue( true ); break;
			default:			Ksu->SetValue( true );
	}
}

void ConfigureMisc::OnConfigureDnD()
{
int MOVE=0,COPY=0,HARDLINK=0,SOFTLINK=0; bool dup;
const wxChar* checks[] = { wxT("MS"), wxT("MC"), wxT("MA"), wxT("CS"), wxT("CC"), wxT("CA"), wxT("HS"), wxT("HC"), wxT("HA"), wxT("SS"), wxT("SC"), wxT("SA"),  };
int metakeys[4] = { tempMETAKEYS_MOVE, tempMETAKEYS_COPY, tempMETAKEYS_HARDLINK, tempMETAKEYS_SOFTLINK };
int mask[3] = { wxACCEL_SHIFT, wxACCEL_CTRL, wxACCEL_ALT };

wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, page, wxT("ConfigDnDDlg"));

for ( size_t n=0; n<12; ++n )
		((wxCheckBox*)dlg.FindWindow(checks[n]))->SetValue( metakeys[n/3] & mask[n%3] );

((wxSpinCtrl*)dlg.FindWindow(wxT("Horiz")))->SetValue( tempDnDXTRIGGERDISTANCE );
((wxSpinCtrl*)dlg.FindWindow(wxT("Vertical")))->SetValue( tempDnDYTRIGGERDISTANCE );
((wxSpinCtrl*)dlg.FindWindow(wxT("LINES_PER_MOUSEWHEEL_SCROLL")))->SetValue( tempLINES_PER_MOUSEWHEEL_SCROLL );

HelpContext = HCdnd;
do
 	{	if ( dlg.ShowModal() != wxID_OK ) return;
		for ( size_t n=0; n<3; ++n )
			if ( ((wxCheckBox*)dlg.FindWindow(checks[n]))->GetValue() ) MOVE |= mask[n];
		for ( size_t n=3; n<6; ++n )
			if ( ((wxCheckBox*)dlg.FindWindow(checks[n]))->GetValue() ) COPY |= mask[n%3];
		for ( size_t n=6; n<9; ++n )
			if ( ((wxCheckBox*)dlg.FindWindow(checks[n]))->GetValue() ) HARDLINK |= mask[n%3];
		for ( size_t n=9; n<12; ++n )
			if ( ((wxCheckBox*)dlg.FindWindow(checks[n]))->GetValue() ) SOFTLINK |= mask[n%3];
		dup = (MOVE==COPY) || (MOVE==HARDLINK) || (MOVE==SOFTLINK) || (COPY==HARDLINK) || (COPY==SOFTLINK) || (HARDLINK==SOFTLINK);
		if ( dup && wxMessageBox(_("Each metakey pattern must be unique. Try again?"), _("Oops"), wxYES_NO ) != wxYES )  return;
	}
  while (dup);

tempMETAKEYS_MOVE = MOVE; tempMETAKEYS_COPY = COPY; tempMETAKEYS_HARDLINK = HARDLINK; tempMETAKEYS_SOFTLINK = SOFTLINK;
tempDnDXTRIGGERDISTANCE = ((wxSpinCtrl*)dlg.FindWindow(wxT("Horiz")))->GetValue();
tempDnDYTRIGGERDISTANCE = ((wxSpinCtrl*)dlg.FindWindow(wxT("Vertical")))->GetValue();
tempLINES_PER_MOUSEWHEEL_SCROLL = ((wxSpinCtrl*)dlg.FindWindow(wxT("LINES_PER_MOUSEWHEEL_SCROLL")))->GetValue();
}

void ConfigureMisc::OnConfigureStatusbar()
{
const wxChar* spinnames[] = { wxT("Menu"), wxT("Success"), wxT("Filepath"), wxT("Filters") };
HelpContext = HCconfigMiscStatbar;
wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, page, wxT("ConfigStatusDlg"));

for ( size_t n=0; n < 4; ++n )
		((wxSpinCtrl*)dlg.FindWindow(spinnames[n]))->SetValue( -tempwidths[n] );		// View as +ve numbers, though they're used as negatives

if ( dlg.ShowModal() == wxID_OK )
	for ( size_t n=0; n < 4; ++n )
		tempwidths[n] = -((wxSpinCtrl*)dlg.FindWindow(spinnames[n]))->GetValue();	// See above

HelpContext = HCconfigMisc;
}

void ConfigureMisc::OnExportBtn( ConfigureNotebook* note )	// Intended for packagers, to export distro-specific bits of the config
{
HelpContext = HCconfigMiscExport;
wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, page, wxT("ConfigExportDlg"));
int ans = dlg.ShowModal();
HelpContext = HCconfigMisc;
if ( ans != wxID_OK ) return;

bool misc = ((wxCheckBox*)dlg.FindWindow( wxT("CheckMisc") ))->IsChecked(),
		edit = ((wxCheckBox*)dlg.FindWindow( wxT("CheckEditors") ))->IsChecked(),
		files = ((wxCheckBox*)dlg.FindWindow( wxT("CheckFileAss") ))->IsChecked(),
		term = ((wxCheckBox*)dlg.FindWindow( wxT("CheckTerminals") ))->IsChecked(),
		tool = ((wxCheckBox*)dlg.FindWindow( wxT("CheckTools") ))->IsChecked();

if ( ! (misc||edit||files||term||tool) )
	{ wxMessageBox(_("You chose not to export any data type! Aborting."), wxT(":/")); return; }

wxString conf( wxGetHomeDir() + wxT("/4Pane.conf") );
wxConfigBase* Export = new wxConfig( wxEmptyString, wxEmptyString, conf );

if ( misc )
	{	DeviceManager::SaveMiscData( Export );				// ConfigureDevices stuff
		DeviceBitmapButton::SaveDefaults( Export );
		AdvancedDialog::SaveConfig( Export );
	}

if ( edit )
	{	EditorsDialog edlg;													// The easiest way to do Editors is to load the configure dialog, but not ShowModal()
		wxXmlResource::Get()->LoadDialog(&edlg, page, wxT("ConfigEditorsEtcDlg")); edlg.Init( (DeviceManager*)page );
		edlg.Save( Export );
	}

if ( files )
	{	wxString home = wxGetHomeDir();							// just because we need a file to pass to...
		FiletypeManager manager( home );						// File Associations, Extensions
		manager.LoadFiletypes(); manager.LoadExtdata();
		manager.SaveFiletypes( Export ); manager.SaveExtdata( Export );
	}

if ( term )
	ConfigureTerminals::Save( Export );							// Terminals

if ( tool )
	note->LMT->Export( Export );										// UserDefined Tools

Export->Flush();
delete Export;
}
void ConfigureMisc::OnUpdateUI()
{
		// I'm null-testing the first control, as there were occasional segs when the notebook page changed; probably a race condition somehow
if ( page == NULL ) return;
wxSpinCtrl* spin = (wxSpinCtrl*)page->FindWindow(wxT("MAX_NUMBER_OF_UNDOS")); if ( spin==NULL ) return;
tempMAX_NUMBER_OF_UNDOS = spin->GetValue();
tempMAX_DROPDOWN_DISPLAY = ((wxSpinCtrl*)page->FindWindow(wxT("MAX_DROPDOWN_DISPLAY")))->GetValue();
tempAUTOCLOSE_TIMEOUT = ((wxSpinCtrl*)page->FindWindow(wxT("AUTOCLOSE_TIMEOUT")))->GetValue() * 1000;
tempAUTOCLOSE_FAILTIMEOUT = ((wxSpinCtrl*)page->FindWindow(wxT("AUTOCLOSE_FAILTIMEOUT")))->GetValue() * 1000;
tempDEFAULT_BRIEFLOGSTATUS_TIME = ((wxSpinCtrl*)page->FindWindow(wxT("DEFAULT_BRIEFLOGSTATUS_TIME")))->GetValue();

tempWHICH_SU = (sutype)((kdesu * (Ksu->IsEnabled() && Ksu->GetValue())) + 
							(gksu * (Gksu->IsEnabled() && Gksu->GetValue())) + (gnomesu * (Gnsu->IsEnabled() && Gnsu->GetValue())));

dirty = (MAX_NUMBER_OF_UNDOS!=tempMAX_NUMBER_OF_UNDOS) || (tempMAX_DROPDOWN_DISPLAY!=MAX_DROPDOWN_DISPLAY)
			|| (tempAUTOCLOSE_TIMEOUT!=AUTOCLOSE_TIMEOUT) || (tempAUTOCLOSE_FAILTIMEOUT!=AUTOCLOSE_FAILTIMEOUT) || (tempWHICH_SU!=WHICH_SU)
			|| (tempDEFAULT_BRIEFLOGSTATUS_TIME!=DEFAULT_BRIEFLOGSTATUS_TIME) || (tempLINES_PER_MOUSEWHEEL_SCROLL!=LINES_PER_MOUSEWHEEL_SCROLL)
			|| tempMETAKEYS_MOVE!=METAKEYS_MOVE || tempMETAKEYS_COPY!=METAKEYS_COPY || tempMETAKEYS_HARDLINK!=METAKEYS_HARDLINK
			|| tempMETAKEYS_SOFTLINK!=METAKEYS_SOFTLINK || tempDnDXTRIGGERDISTANCE!=DnDXTRIGGERDISTANCE || tempDnDYTRIGGERDISTANCE!=DnDYTRIGGERDISTANCE
			|| tempwidths[0]!=widths[0] || tempwidths[1]!=widths[1] || tempwidths[2]!=widths[2] || tempwidths[3]!=widths[3];

page->FindWindow(wxT("Misc_OK"))->Enable( dirty );
page->FindWindow(wxT("Misc_CANCEL"))->Enable( dirty );
}

void ConfigureMisc::OnOK()
{
MAX_NUMBER_OF_UNDOS = tempMAX_NUMBER_OF_UNDOS;
MAX_DROPDOWN_DISPLAY = tempMAX_DROPDOWN_DISPLAY;
AUTOCLOSE_TIMEOUT = tempAUTOCLOSE_TIMEOUT;
AUTOCLOSE_FAILTIMEOUT = tempAUTOCLOSE_FAILTIMEOUT;
DEFAULT_BRIEFLOGSTATUS_TIME = tempDEFAULT_BRIEFLOGSTATUS_TIME;
LINES_PER_MOUSEWHEEL_SCROLL = tempLINES_PER_MOUSEWHEEL_SCROLL;
WHICH_SU = tempWHICH_SU;
METAKEYS_MOVE = tempMETAKEYS_MOVE; METAKEYS_COPY = tempMETAKEYS_COPY;
METAKEYS_HARDLINK = tempMETAKEYS_HARDLINK; METAKEYS_SOFTLINK = tempMETAKEYS_SOFTLINK;
DnDXTRIGGERDISTANCE = tempDnDXTRIGGERDISTANCE; DnDYTRIGGERDISTANCE = tempDnDYTRIGGERDISTANCE;
widths[0] = tempwidths[0]; widths[1] = tempwidths[1]; widths[2] = tempwidths[2]; widths[3] = tempwidths[3];

Save();
}

//static 
void ConfigureMisc::LoadMiscConfig()
{
wxConfigBase* config = wxConfigBase::Get();	if (config==NULL)  return; 

MAX_NUMBER_OF_UNDOS = (size_t)config->Read( wxT("/Misc/Misc/MAX_NUMBER_OF_UNDOS"), 10000l );
MAX_DROPDOWN_DISPLAY = (size_t)config->Read( wxT("/Misc/Misc/MAX_DROPDOWN_DISPLAY"), 15l );
AUTOCLOSE_TIMEOUT = config->Read( wxT("/Misc/Misc/AUTOCLOSE_TIMEOUT"), 2000l );
AUTOCLOSE_FAILTIMEOUT = config->Read( wxT("/Misc/Misc/AUTOCLOSE_FAILTIMEOUT"), 5000l );
DEFAULT_BRIEFLOGSTATUS_TIME = config->Read( wxT("/Misc/Misc/DEFAULT_BRIEFLOGSTATUS_TIME"), 10l );
LINES_PER_MOUSEWHEEL_SCROLL = config->Read( wxT("/Misc/Misc/LINES_PER_MOUSEWHEEL_SCROLL"), 5l );
WHICH_SU = (sutype)config->Read( wxT("/Misc/Which_Su"), 0l );
METAKEYS_MOVE = (int)config->Read( wxT("/Misc/Misc/METAKEYS_MOVE"), 0l );
METAKEYS_COPY = (int)config->Read( wxT("/Misc/Misc/METAKEYS_COPY"), 2l );
METAKEYS_HARDLINK = (int)config->Read( wxT("/Misc/Misc/METAKEYS_HARDLINK"), 6l );
METAKEYS_SOFTLINK = (int)config->Read( wxT("/Misc/Misc/METAKEYS_SOFTLINK"), 5l );
DnDXTRIGGERDISTANCE = (int)config->Read( wxT("/Misc/Misc/DnDXTRIGGERDISTANCE"), 10l );
DnDYTRIGGERDISTANCE = (int)config->Read( wxT("/Misc/Misc/DnDYTRIGGERDISTANCE"), 15l );
}

//static 
void ConfigureMisc::LoadStatusbarWidths( int* widths )
{		// Load numbers and negate them: they're stored as positive numbers, but used as negative proportions
widths[0] = - (int)wxConfigBase::Get()->Read( wxT("/Misc/Misc/StatusbarWidth0"), (long)DEFAULT_STATUSBAR_WIDTHS[0] );
widths[1] = - (int)wxConfigBase::Get()->Read( wxT("/Misc/Misc/StatusbarWidth1"), (long)DEFAULT_STATUSBAR_WIDTHS[1] );
widths[2] = - (int)wxConfigBase::Get()->Read( wxT("/Misc/Misc/StatusbarWidth2"), (long)DEFAULT_STATUSBAR_WIDTHS[2] );
widths[3] = - (int)wxConfigBase::Get()->Read( wxT("/Misc/Misc/StatusbarWidth3"), (long)DEFAULT_STATUSBAR_WIDTHS[3] );
}

void ConfigureMisc::Save()
{
wxConfigBase* config = wxConfigBase::Get();	if (config==NULL)  return; 

config->Write( wxT("/Misc/Misc/MAX_NUMBER_OF_UNDOS"), (long)MAX_NUMBER_OF_UNDOS );
config->Write( wxT("/Misc/Misc/MAX_DROPDOWN_DISPLAY"), (long)MAX_DROPDOWN_DISPLAY );
config->Write( wxT("/Misc/Misc/AUTOCLOSE_TIMEOUT"), (long)AUTOCLOSE_TIMEOUT );
config->Write( wxT("/Misc/Misc/AUTOCLOSE_FAILTIMEOUT"), (long)AUTOCLOSE_FAILTIMEOUT );
config->Write( wxT("/Misc/Misc/DEFAULT_BRIEFLOGSTATUS_TIME"), (long)DEFAULT_BRIEFLOGSTATUS_TIME );
config->Write( wxT("/Misc/Misc/LINES_PER_MOUSEWHEEL_SCROLL"), (long)LINES_PER_MOUSEWHEEL_SCROLL );
config->Write( wxT("/Misc/Which_Su"), (long)WHICH_SU );

config->Write( wxT("/Misc/Misc/METAKEYS_MOVE"), (long)METAKEYS_MOVE );
config->Write( wxT("/Misc/Misc/METAKEYS_COPY"), (long)METAKEYS_COPY );
config->Write( wxT("/Misc/Misc/METAKEYS_HARDLINK"), (long)METAKEYS_HARDLINK );
config->Write( wxT("/Misc/Misc/METAKEYS_SOFTLINK"), (long)METAKEYS_SOFTLINK );
config->Write( wxT("/Misc/Misc/DnDXTRIGGERDISTANCE"), (long)DnDXTRIGGERDISTANCE );
config->Write( wxT("/Misc/Misc/DnDYTRIGGERDISTANCE"), (long)DnDYTRIGGERDISTANCE );

config->Write( wxT("/Misc/Misc/StatusbarWidth0"), (long)-widths[0] );	// Save as positive numbers: widths are used as negative proportions, but easier to store as +ves
config->Write( wxT("/Misc/Misc/StatusbarWidth1"), (long)-widths[1] );
config->Write( wxT("/Misc/Misc/StatusbarWidth2"), (long)-widths[2] );
config->Write( wxT("/Misc/Misc/StatusbarWidth3"), (long)-widths[3] );

config->Flush();
}
//-------------------------------------------------------------------------------------------------------------------------------------------

void ChooseStripesDlg::DisplayCols()
{
FindWindow(wxT("Col0Text"))->SetBackgroundColour( Col0 );
FindWindow(wxT("Col1Text"))->SetBackgroundColour( Col1 );
}

void ChooseStripesDlg::OnChangeButton(wxCommandEvent& event)
{
wxColourData data;
for (int i = 0; i < 16; i++)
  { wxColour colour(i*16, i*16, i*16); data.SetCustomColour(i, colour); }
if ( Col0 != wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW) )	// If there are already custom choices, provide them in the dialog
	data.SetCustomColour(0, Col0);
if ( Col1 != wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW) )
	data.SetCustomColour(1, Col1);

wxColourDialog dialog(this, &data);
if ( dialog.ShowModal() == wxID_OK )
	{	wxColourData retData = dialog.GetColourData();
    	wxColour col = retData.GetColour();
		if ( event.GetId() == XRCID("Colour0Btn") )
			{	if ( Col0 != col )
					{ Col0 = col; dirty = true; }
			}
		  else
			{	if ( Col1 != col )
					{ Col1 = col; dirty = true; }
			}
		if ( dirty ) DisplayCols();
	}
}

BEGIN_EVENT_TABLE(ChooseStripesDlg, wxDialog)
	EVT_BUTTON(XRCID("Colour0Btn"), ChooseStripesDlg::OnChangeButton)
	EVT_BUTTON(XRCID("Colour1Btn"), ChooseStripesDlg::OnChangeButton)
END_EVENT_TABLE()

//-------------------------------------------------------------------------------------------------------------------------------------------

void ToolbarIconsDialog::Init( ArrayofDirCtrlTBStructs* DCTBarray )
{
DirCtrlTBinfoarray = DCTBarray;

wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)  return;
long NoOfKnownIcons = config->Read( wxT("/Misc/DirCtrlTB/Icons/NoOfKnownIcons"), 0l );
for ( long n=0; n < NoOfKnownIcons; ++n )
	{	wxString Icon;
		config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icon"), &Icon ); Icons.Add( Icon );
		long Icontype = config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icontype"), -1l ); Icontypes.Add( Icontype );
	}

list = (wxListBox*)FindWindow( wxT("List") );
}

void ToolbarIconsDialog::OnInitDialog(wxInitDialogEvent& WXUNUSED(event))
{
FillListbox();
}
//static
void ToolbarIconsDialog::LoadDirCtrlTBinfo( ArrayofDirCtrlTBStructs* DirCtrlTBinfoarray )	// Load info on the desired extra buttons to load to each Dirview mini-toolbar
{
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)   return;

config->SetPath( wxT("/Misc/DirCtrlTB/Buttons/") );
size_t count = config->GetNumberOfGroups();					// Count the subgroups
long NoOfKnownIcons = config->Read( wxT("/Misc/DirCtrlTB/Icons/NoOfKnownIcons"), 0l );

DirCtrlTBinfoarray->Clear();

wxString key, histprefix(wxT("button_"));	
for ( size_t n=0; n < count; ++n )										// Create the key for every item, from button_a to button_('a'+count-1)
	{	wxString bitmaplocation, tooltip, destination; int iconNo;
		key.Printf(wxT("%c"), wxT('a') + (wxChar)n );				// Make key hold a, b, c etc
		config->Read( histprefix+key+wxT("/destination"), &destination );	if ( destination.IsEmpty() )   continue;
		iconNo = (int)config->Read( histprefix+key+wxT("/iconNo"), -1l ); if ( iconNo == -1 || iconNo >=NoOfKnownIcons ) continue;
		config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(iconNo,NoOfKnownIcons)+wxT("/Icon"), &bitmaplocation );
		long Icontype = config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(iconNo,NoOfKnownIcons)+wxT("/Icontype"), -1l );
		config->Read( histprefix+key+wxT("/tooltip"), &tooltip );
				
		DirCtrlTBStruct* tbs = new DirCtrlTBStruct;				// Make a new struct, & store the data in it
		wxString path; if ( bitmaplocation.Left(1) != wxFILE_SEP_PATH ) path = BITMAPSDIR;
		tbs->bitmaplocation = path + bitmaplocation;
		tbs->tooltip = tooltip;
		tbs->iconNo = iconNo;
		tbs->icontype = (wxBitmapType)Icontype;
		tbs->destination = destination;
		DirCtrlTBinfoarray->Add( tbs );									// Store it in the struct array
	}

config->SetPath(wxT("/"));	
}

void ToolbarIconsDialog::Save()
{
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)  return;

config->DeleteGroup( wxT("/Misc/DirCtrlTB/Buttons") );	// Delete current info
config->SetPath( wxT("/Misc/DirCtrlTB/Buttons/") );

size_t count = DirCtrlTBinfoarray->GetCount();				// Count the entries to be saved

wxString key, histprefix(wxT("button_"));	
for ( size_t n=0; n < count; ++n )									// Create the key for every item, from button_a to button_('a'+count-1)
	{	key.Printf(wxT("%c"), wxT('a') + (wxChar)n );			// Make key hold a, b, c etc
		config->Write( histprefix+key+wxT("/IconNo"), DirCtrlTBinfoarray->Item(n)->iconNo );
		config->Write( histprefix+key+wxT("/tooltip"), DirCtrlTBinfoarray->Item(n)->tooltip );
		config->Write( histprefix+key+wxT("/destination"), DirCtrlTBinfoarray->Item(n)->destination );
	}


config->Flush();
config->SetPath(wxT("/"));
}

void ToolbarIconsDialog::SaveIcons()
{
size_t NoOfKnownIcons = Icons.GetCount(); if ( ! NoOfKnownIcons ) return;
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)  return;

config->DeleteGroup( wxT("/Misc/DirCtrlTB/Icons") );
config->Write( wxT("/Misc/DirCtrlTB/Icons/NoOfKnownIcons"), (long)NoOfKnownIcons );
for ( size_t n=0; n < NoOfKnownIcons; ++n )
	{	config->Write( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icon"), Icons[n] );
		config->Write( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icontype"), Icontypes[n] );
	}

config->Flush();
}

void ToolbarIconsDialog::AddIcon( wxString& newfilepath, size_t type )
{
if ( newfilepath.IsEmpty() ) return;
if ( newfilepath.find( BITMAPSDIR ) == 0 ) newfilepath = newfilepath.AfterLast(wxFILE_SEP_PATH);	// If stored in BITMAPSDIR, use only the filename

Icons.Add( newfilepath ); Icontypes.Add( type );
SaveIcons();
}

void ToolbarIconsDialog::FillListbox()
{
list->Clear();																		// In case this is a refill

for ( size_t n=0; n < DirCtrlTBinfoarray->GetCount(); ++n )  
	{	wxString item = DirCtrlTBinfoarray->Item(n)->destination;
		list->Append( TruncatePathtoFitBox( item, list ) );
	}

if ( DirCtrlTBinfoarray->GetCount() ) list->SetSelection(0);
}

void ToolbarIconsDialog::OnAddButton(wxCommandEvent& WXUNUSED(event))
{
AddToobarIconDlg dlg;
wxXmlResource::Get()->LoadDialog(&dlg, this, wxT("AddToobarIconDlg"));
size_t nexticon = DirCtrlTBinfoarray->GetCount();				// See if we can guess the icon to use: the next available one
size_t NoOfKnownIcons = (size_t)wxConfigBase::Get()->Read( wxT("/Misc/DirCtrlTB/Icons/NoOfKnownIcons"), 0l );
size_t icon = nexticon < NoOfKnownIcons ? nexticon : 0;

dlg.Init( this, MyFrame::mainframe->GetActivePane()->GetPath(), wxEmptyString, icon );
if ( dlg.ShowModal() != wxID_OK ) return;

wxString Filepath = dlg.Filepath->GetValue(); if ( Filepath.IsEmpty() ) return;

DirCtrlTBStruct* tbs = new DirCtrlTBStruct;						// Make a new struct, & store the data in it
tbs->destination = Filepath; tbs->tooltip = dlg.Tooltip->GetValue(); tbs->iconNo = dlg.iconno;
DirCtrlTBinfoarray->Add( tbs );

Save();																				// Rather than faffing around with a 'dirty' flag, just save each time
FillListbox();																		// Redo the listbox to reflect the new entry
}

void ToolbarIconsDialog::OnEditButton(wxCommandEvent& WXUNUSED(event))
{
int index = list->GetSelection(); if ( index == wxNOT_FOUND ) return;

AddToobarIconDlg dlg;
wxXmlResource::Get()->LoadDialog(&dlg, this, wxT("AddToobarIconDlg"));

DirCtrlTBStruct* tbs = DirCtrlTBinfoarray->Item(index);		// Readability
dlg.Init( this, tbs->destination, tbs->tooltip, tbs->iconNo );
if ( dlg.ShowModal() != wxID_OK ) return;

wxString Filepath = dlg.Filepath->GetValue(); if ( Filepath.IsEmpty() ) return;	// It's difficult to imagine what would be more appropriate
tbs->destination = Filepath; tbs->tooltip = dlg.Tooltip->GetValue(); tbs->iconNo = dlg.iconno;

Save();																				// Rather than faffing around with a 'dirty' flag, just save each time
FillListbox();																		// Redo the listbox to reflect the alteration
}

void ToolbarIconsDialog::OnDeleteButton(wxCommandEvent& WXUNUSED(event))
{
int index = list->GetSelection(); if ( index == wxNOT_FOUND ) return;

wxMessageDialog ask(this,_("Delete this toolbar button?"),_("Are you sure?"), wxYES_NO | wxICON_QUESTION  );
if ( ask.ShowModal() == wxID_YES )
	{	DirCtrlTBinfoarray->RemoveAt(index);
		Save(); FillListbox();
	}
}

int ToolbarIconsDialog::SetBitmap( wxBitmapButton* btn, wxWindow* pa, size_t iconno )
{
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)  return false;

long NoOfKnownIcons = config->Read( wxT("/Misc/DirCtrlTB/Icons/NoOfKnownIcons"), 0l );

wxString Icon; config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(iconno,NoOfKnownIcons)+wxT("/Icon"), &Icon );
long Icontype = config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(iconno,NoOfKnownIcons)+wxT("/Icontype"), -1l );	// Can be wxBITMAP_TYPE_PNG, wxBITMAP_TYPE_XPM or wxBITMAP_TYPE_BMP (other handlers aren't loaded)
if ( Icontype==-1 ) return false;

wxString path; if ( Icon.Left(1) != wxFILE_SEP_PATH ) path = BITMAPSDIR;		// If icon is just a filename, we'll prepend with BITMAPSDIR
wxBitmap bitmap(path+Icon, (wxBitmapType)Icontype );	// gtk-2.4.2 was happy without the cast, but not 2.6.2
int id = NextID++;
if ( ! btn->Create( pa, id, bitmap, wxDefaultPosition, wxDefaultSize, wxNO_BORDER ) ) return false;
return id;
}

void ToolbarIconsDialog::DoUpdateUI(wxUpdateUIEvent& event)
{
event.Enable( list->GetCount() > 0 ); 
}

BEGIN_EVENT_TABLE(ToolbarIconsDialog,wxDialog)
	EVT_BUTTON(XRCID("AddBtn"), ToolbarIconsDialog::OnAddButton)
	EVT_BUTTON(XRCID("EditBtn"), ToolbarIconsDialog::OnEditButton)
	EVT_BUTTON(XRCID("DeleteBtn"), ToolbarIconsDialog::OnDeleteButton)
	EVT_INIT_DIALOG(ToolbarIconsDialog::OnInitDialog)
	EVT_UPDATE_UI( XRCID("EditBtn"), ToolbarIconsDialog::DoUpdateUI)
	EVT_UPDATE_UI( XRCID("DeleteBtn"), ToolbarIconsDialog::DoUpdateUI)
END_EVENT_TABLE()

void AddToobarIconDlg::Init( ToolbarIconsDialog* granddad, wxString defaultfilepath, wxString defaulttooltip, long icon )
{
grandparent = granddad; iconno = icon; btn = NULL;

Filepath = (wxTextCtrl*)(FindWindow( wxT("Filepath") ));
if ( ! defaultfilepath.IsEmpty() )  Filepath->SetValue( defaultfilepath );
Tooltip = (wxTextCtrl*)(FindWindow( wxT("Tooltip") ));
if ( ! defaulttooltip.IsEmpty() )  Tooltip->SetValue( defaulttooltip );


int id = CreateButton();																			// Create the current bitmapbutton
if ( id ) Connect( id, wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction)&AddToobarIconDlg::OnIconClicked );	// & connect it to Select Icon dialog

Connect( XRCID("Browse"), wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction)&AddToobarIconDlg::OnBrowseBtn ); // Connect the browse button too
}

int AddToobarIconDlg::CreateButton()
{
wxStaticText* st = (wxStaticText*)(FindWindow( wxT("StaticText") )); 	// Find the sizer to which to add the bitmapbutton
wxSizer* sizer = st->GetContainingSizer();

if ( btn != NULL) btn->Destroy();															// Kill any old button, in case we're changing a bitmap
btn = new wxBitmapButton; 

int id = grandparent->SetBitmap( btn, this, iconno ); if ( id ) sizer->Add( btn, 0,  wxALIGN_CENTRE );	// Create the current icon and add to sizer
sizer->Layout();
return id;
}

void AddToobarIconDlg::OnIconClicked(wxCommandEvent& WXUNUSED(event))
{
TBIconSelectDlg dlg;
wxXmlResource::Get()->LoadDialog(&dlg, this, wxT("IconSelectDlg"));
dlg.Init( grandparent, iconno );
if ( dlg.ShowModal() != wxID_OK ) return;

iconno = dlg.iconno;																				// Get any altered icon choice
int id = CreateButton();																			// Recreate the current bitmapbutton
if ( id ) Connect( id, wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction)&AddToobarIconDlg::OnIconClicked );	// & connect the new version to Select Icon dialog
}

void AddToobarIconDlg::OnBrowseBtn(wxCommandEvent& WXUNUSED(event))
{
FileDirDlg fdlg( this, FD_SHOWHIDDEN, HOME, _("Browse for the Filepath to Add") );
if ( fdlg.ShowModal() != wxID_OK ) return;

Filepath->SetValue( fdlg.GetFilepath() );
}

#include <math.h>							// For sqrt
void TBIconSelectDlg::DisplayButtonChoices()
{
wxConfigBase* config = wxConfigBase::Get(); if (config==NULL)  return;
long NoOfKnownIcons = config->Read( wxT("/Misc/DirCtrlTB/Icons/NoOfKnownIcons"), 0l );

if ( panel != NULL ) panel->Destroy();														// In case this is a rerun
panel = new wxPanel( this, wxID_ANY );
size_t cols = (size_t)sqrt( NoOfKnownIcons ); cols = wxMax( cols, 1 ); size_t colsremainder = (((size_t)NoOfKnownIcons) - (cols*cols)) > 0;
size_t rows = ((size_t)NoOfKnownIcons)/(cols+1) + ( ((size_t)NoOfKnownIcons)%(cols+1) > 0);
gridsizer = new wxGridSizer(cols + colsremainder, 3, 3);

int AverageIconHt=0;	// We'll need this to work out how big to make the panel, as a wrong guess resulted in invisible OK/Cancel buttons. Width doesn't matter so much.
StartId = NextID;																							// Store the id that'll be given to the 1st button
for (int n=0; n < (int)NoOfKnownIcons; ++n)
	{	wxString Icon; config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icon"), &Icon );
		long Icontype = config->Read( wxT("/Misc/DirCtrlTB/Icons/")+CreateSubgroupName(n,NoOfKnownIcons)+wxT("/Icontype"), -1l );	// Can be wxBITMAP_TYPE_PNG, wxBITMAP_TYPE_XPM or wxBITMAP_TYPE_BMP (other handlers aren't loaded)
		if ( Icontype == -1 ) continue;
		
		wxString path; if ( Icon.Left(1) != wxFILE_SEP_PATH ) path = BITMAPSDIR;	// If icon is just a filename, we'll prepend with BITMAPSDIR
		wxBitmap bitmap(path+Icon, (wxBitmapType)Icontype );					// gtk-2.4.2 was happy without the cast, but not 2.6.2
		if ( &bitmap == (wxBitmap*)NULL ) continue;
		wxBitmapButton* btn = new wxBitmapButton( panel, NextID++, bitmap, wxDefaultPosition, wxDefaultSize, wxNO_BORDER );
		if ( btn == NULL) { --NextID; continue; }
		
		int iconht, iconwd; btn->GetSize(&iconwd,&iconht); AverageIconHt += iconht;
		 gridsizer->Add( btn );
	}	

AverageIconHt /= NoOfKnownIcons;
panel->SetSize((cols+1)*30, rows*AverageIconHt );
panel->SetSizer( gridsizer );
bsizer->Add( panel, 1, wxALIGN_CENTRE );

SetSize( width+(cols+1)*30, height+(rows+1)*AverageIconHt );				// Resize the whole dialog to take the panel into account. I couldn't find a pure sizer way to do this :(
GetSizer()->Layout();

EndId = NextID - 1;																						// Store the id that was given to the last button
Connect( StartId, EndId, wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction)&TBIconSelectDlg::OnNewSelection );
}


//-------------------------------------------------------------------------------------------------------------------------------------------














