/////////////////////////////////////////////////////////////////////////////
// Name:        Misc.cpp
// Purpose:    Misc stuff
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////
#include "wx/wxprec.h" 

#ifndef WX_PRECOMP
	#include "wx/artprov.h"
	#include "wx/dirctrl.h"
#endif

#include "Misc.h"

const int DEFAULT_BRIEFMESSAGEBOX_TIME = 2;

BriefMessageBox::BriefMessageBox( wxString Message, double secondsdisplayed /*= -1*/, wxString Caption /*= wxEmptyString*/ )
														:	wxDialog( NULL, -1, Caption )
{
wxStaticBox* staticbox = new wxStaticBox( this, -1, wxT("") );	// Start by doing sizer things, so that there's somewhere to put the message
wxStaticBoxSizer *textsizer = new wxStaticBoxSizer( staticbox, wxHORIZONTAL );	
textsizer->Add( CreateTextSizer( Message ), 1, wxCENTRE | wxALL, 10 );

wxBoxSizer *mainsizer = new wxBoxSizer( wxHORIZONTAL );
mainsizer->Add( textsizer, 1, wxCENTER | wxALL, 20 );

SetSizer( mainsizer ); mainsizer->Fit( this );
Centre( wxBOTH | wxCENTER_FRAME);

BriefTimer.Setup(this);																				// Let the timer know who's boss

double length;
if ( secondsdisplayed <= 0 ) 																		// Minus number means default, and zero would mean infinite which we don't want here
	length = DEFAULT_BRIEFMESSAGEBOX_TIME;
 else length = secondsdisplayed;
 
BriefTimer.Start( (int)(length * 1000), wxTIMER_ONE_SHOT );				// Start timer as one-shot.  *1000 gives seconds

ShowModal();																								// Display the message
}

wxString ParseSize( unsigned long bytes )				// Returns a string filled with the size, but in bytes, KB or MB according to magnitude
{
wxString text;

if ( bytes < 1024 )	text.Printf( wxT("%uB"), (uint)bytes );
 else if  ( bytes < 1048576 )	text.Printf( wxT("%.1fK"), ((float)bytes) / 1024 );
   else text.Printf( wxT("%.1fM"), ((float)bytes) / 1048576 );

return text;
}

wxString DoEscape( wxString& filepath, const wxString& target )	// Escapes e.g. spaces or quotes within e.g. filepaths
{
wxString replacement =( wxT('\\') ); replacement << target;
size_t index, nStart = 0;		
while (1)
	{	index = filepath.find( target, nStart );													// This std::string find starts at nStart, so we can reset this after each Find
		if ( index == wxString::npos ) break;													// npos means no more found
		if ( filepath.GetChar( index-1 ) != wxT('\\') ) 										// If it's not already escaped, do it, and change nStart accordingly.  +2, one for the \ one for the space
			{	filepath.replace( index, 1, replacement ); nStart = index+2; }
		  else nStart = index+1;																		// Already escaped so just inc over the it
	}
return filepath;
}

wxString EscapeQuote( wxString& filepath )	// Often, we use single quotes around a filepath to avoid problems with spaces etc.
{																		// However this CAUSES a problem with apostrophes!  This function goes thru filepath, Escaping any we find
return DoEscape( filepath, wxT("\'") );
}

wxString EscapeQuoteStr( wxString filepath )	// A wxString (cf wxString&) version of the above
{
wxString fp( filepath );
return EscapeQuote( filepath );
}

wxString EscapeSpace( wxString& filepath )	// Escapes spaces within e.g. filepaths
{
return DoEscape( filepath, wxT(" ") );
}

wxString TruncatePathtoFitBox( wxString filepath, wxWindow* box )	// Returns a string small enough to be displayed in box, with ../ prepend if needed
{
if ( filepath.IsEmpty() || box==NULL ) return wxEmptyString;

int x, y, clx, cly;
box->GetClientSize( &clx, &cly ); box->GetTextExtent( filepath, &x, &y );
if ( x > clx )																									// If the textextent is greater than the clientsize, 
	{	wxString dots( wxT(".../") ); int dx, dy; box->GetTextExtent( dots, &dx, &dy );	// make a string with .../ in it, and get *its* textextent too
		filepath = filepath.AfterFirst(wxFILE_SEP_PATH);								// We know it'll begin with /
		do
			{	filepath = filepath.AfterFirst(wxFILE_SEP_PATH);						// Now go thru the path, progressively chopping off its head until it plus dots will fit
				box->GetTextExtent( filepath, &x, &y );
			}
		 while  ( (x+dx) > clx );
		return ( dots + filepath );																	// Return the truncated filepath, preceeded by ../
	}
return filepath;																							// Return the untruncated version, as it fitted anyway
}

void QueryWrapWithShell( wxString& command )		// Wraps command with sh -c if appropriate
{
if ( command.Contains( wxT("sh -c") ) ) return;							// We don't need two of them

const wxChar* relevantchars = wxT("~|><*?");
if ( command.find_first_of( relevantchars ) != wxString::npos )
	{	EscapeQuote( command );													// We're going to wrap with sh -c '  '   to avoid clashes with double-quotes, so escape any single-quotes first
		command = wxT("sh -c \'") + command; command += wxT('\''); 
	}
}

void QuoteExcludingWildcards( wxString& command )	// Wraps command with " " if appropriate, excluding any terminal globs. Needed for grep
{
if ( command.IsEmpty() ) return;
if ( command.find_first_of( wxT('\"') ) != wxString::npos ) return;				// If there's a double-quote in the command, leave well alone: surrounding by " " will be guaranteed to fail

const wxChar* relevantchars = wxT("' ");
if ( command.find_first_of( relevantchars ) == wxString::npos ) return;	// If there isn't a single-quote or space, nothing needs be done

wxString path( command.BeforeLast( wxFILE_SEP_PATH ) ); if ( path.IsEmpty() ) return;	// If there isn't a '/', we're buggered anyway so ignore
wxString name( command.AfterLast( wxFILE_SEP_PATH ) ); path << wxFILE_SEP_PATH;

const wxChar* globs = wxT("?*");																// Now see if there's a wildcard in *name* (if it were in path (?allowed), quoting would tame its wildness)
if ( name.find_first_of( globs ) != wxString::npos )										// If one is found, double-quote path, but not name
	{	command = wxT("\\\"") + path; command << wxT("\\\"") << name; }
 else
	{	command = wxT("\\\"") + command; command << wxT("\\\""); }			// Otherwise escape-quote the lot
}


#include <sys/utsname.h>
#include "wx/tokenzr.h"

bool GetFloat( wxString& datastr, double* fl, size_t skip )		// Parses datastr to find a float. Skip means skip over the first n of them
{
wxString str;

wxStringTokenizer tknr( datastr ); size_t count = tknr.CountTokens();
for ( size_t c=0; c < count; ++c )	
	{	wxString token = tknr.GetNextToken();
		size_t index = token.find_first_of(wxT("0123456789")); 			// We need to cope with "foo bar 1.2.3"
		if ( index == wxString::npos ) continue;
		if ( skip ) { --skip; continue; }													// Skip allows us to cope with  "Foo i386 version 2.3"
		
		str = token.Mid( index ); break;
		}
	
if ( str.IsEmpty() ) return false;															// We're digitless

wxStringTokenizer tkn( str, wxT(".") );													// We need to cope with "1.2.3"
wxString token = tkn.GetNextToken();
if ( token.ToDouble( fl ) && tkn.HasMoreTokens() )							// Because of the above, the first token really ought to be a digit or 3
	{	token += wxT('.') + tkn.GetNextToken();									// We've got the 1, so add '.' plus the next digit, and ignore any others
		token.ToDouble( fl );
	}

return true;
}

bool KernelLaterThan( wxString minimum )	// Tests whether the active kernel is >= the parameter
{
struct utsname krnl;
if ( uname( &krnl ) == -1 )	return false;														// If -1 the function call failed

wxString kernel(krnl.release, wxConvLocal);	
wxStringTokenizer knl( kernel, wxT(".") ); wxStringTokenizer mnl( minimum, wxT(".") );		// Tokenise both
int knlcount = knl.CountTokens(), minimalcount = mnl.CountTokens();								// Count tokens;  we may need later to check for equality
while ( knl.HasMoreTokens() && mnl.HasMoreTokens() )
	{	wxString kerneltkn = knl.GetNextToken(), minimaltkn = mnl.GetNextToken();			// Get first token of each & compare
		long kernellong, minimallong; 
		if ( ! kerneltkn.ToLong( &kernellong ) || ! minimaltkn.ToLong( &minimallong ) )  return false;	// If either conversion to numeral failed, abort
		if ( kernellong == minimallong ) continue;																// If this number==, check the next pair of tokens
		
		return  kernellong > minimallong;																			// We have an inequality.  Return 0 or 1, depending on which is larger
	}
				// If we're here, 1 or both strings has run out.  We may be in a 2.4.10 versus 2.4.10 situation, or a 2.4.10 versus 2.4 situation
return knlcount >= minimalcount;																					// If the kernel has >= numbers, it passes
}


IMPLEMENT_DYNAMIC_CLASS(FileDirDlg, wxDialog)

FileDirDlg::FileDirDlg( wxWindow* parent, int style/*FD_MULTIPLE | FD_SHOWHIDDEN*/, wxString defaultpath/*=wxEmptyString*/, wxString label/*=_( "File and Dir Select Dialog" )*/ )
										: wxDialog( parent, wxNewId(), label )
{
SHCUT_HOME = wxNewId();

wxBoxSizer* mainsizer = new wxBoxSizer( wxVERTICAL );

wxBoxSizer* sizer1 = new wxBoxSizer( wxHORIZONTAL );
wxBitmap bmp = wxArtProvider::GetBitmap( wxART_GO_HOME );
wxBitmapButton* Home = new wxBitmapButton( this, SHCUT_HOME, bmp );
sizer1->Add( 0,0, 4 ); sizer1->Add( Home, 0, wxRIGHT, 30 ); sizer1->Add( 0,0, 1 );
mainsizer->Add( sizer1, 0, wxTOP|wxBOTTOM|wxEXPAND, 10 );

wxStaticBoxSizer* sizer2 = new wxStaticBoxSizer( new wxStaticBox(this, -1, wxEmptyString), wxVERTICAL );
if ( defaultpath.IsEmpty() ) defaultpath = wxGetCwd();

GDC = new wxGenericDirCtrl( this, wxID_ANY, defaultpath );
int treestyle = GDC->GetTreeCtrl()->GetWindowStyle();
if ( style & FD_MULTIPLE ) treestyle |= (wxTR_MULTIPLE | wxTR_EXTENDED);
GDC->GetTreeCtrl()->SetWindowStyle( treestyle );
GDC->ShowHidden( style & FD_SHOWHIDDEN );
#if wxVERSION_NUMBER > 2402 && wxVERSION_NUMBER < 2600
	#ifdef __UNIX__
        GDC->SetFilter( wxT("*") );
	#else
	    GDC->SetFilter( wxT("*.*") );
	#endif
#endif
sizer2->Add( GDC, 1, wxEXPAND );
mainsizer->Add( sizer2, 1, wxLEFT|wxRIGHT|wxEXPAND, 15 );

wxBoxSizer* sizer3 = new wxBoxSizer( wxHORIZONTAL );
wxBoxSizer* sizer4 = new wxBoxSizer( wxVERTICAL );
Text = new wxTextCtrl( this, wxID_ANY );
wxCheckBox* ShowHidden =  new wxCheckBox( this, wxID_ANY, wxT("Show Hidden") );
ShowHidden->SetValue( style & FD_SHOWHIDDEN );
sizer4->Add( Text, 0, wxTOP|wxBOTTOM|wxEXPAND, 3 );
sizer4->Add( ShowHidden, 0, wxALIGN_CENTRE );

wxBoxSizer* sizer5 = new wxBoxSizer( wxVERTICAL );
wxButton* OK = new wxButton( this, wxID_OK,  _("OK") );
wxButton* CANCEL = new wxButton( this, wxID_CANCEL,  _("CANCEL") );
sizer5->Add( OK );
sizer5->Add( CANCEL, 0, wxTOP, 5 );
sizer3->Add( sizer4, 1);
sizer3->Add( sizer5, 0, wxLEFT, 15 );
mainsizer->Add( sizer3, 0, wxALL|wxEXPAND, 15 );

SetSizer( mainsizer );
mainsizer->Layout();

wxSize size = wxGetDisplaySize();
SetSize( size.x / 2, size.y / 2 );

filesonly =  style & FD_RETURN_FILESONLY;

	// Use Connect() here instead of an event table entry, as otherwise an event is fired before the dialog is created and GDC and Text initialised
Connect( wxID_ANY, wxEVT_COMMAND_TREE_SEL_CHANGED, (wxObjectEventFunction)&FileDirDlg::OnSelChanged );
Connect( wxID_ANY, wxEVT_COMMAND_CHECKBOX_CLICKED,  (wxObjectEventFunction)&FileDirDlg::OnHidden );
Connect( SHCUT_HOME, wxEVT_COMMAND_BUTTON_CLICKED,  (wxObjectEventFunction)&FileDirDlg::OnHidden );
GDC->GetTreeCtrl()->Connect( wxID_ANY,wxEVT_ERASE_BACKGROUND,  (wxObjectEventFunction)&FileDirDlg::OnEraseBackground );
}

void FileDirDlg::OnHome(wxCommandEvent& WXUNUSED(event))
{
wxString home = wxGetHomeDir(); GDC->SetPath( home );
}

void FileDirDlg::OnHidden(wxCommandEvent& event)
{
GDC->ShowHidden( event.IsChecked() );
}

void FileDirDlg::OnSelChanged(wxTreeEvent& WXUNUSED(event))
{
if ( filesonly ) filepath = GDC->GetFilePath();		// If we don't want dirs, use GetFilePath() which ignores them
  else filepath = GDC->GetPath();						// Otherwise GetPath() which returns anything
Text->SetValue( filepath );
}

void FileDirDlg::GetPaths( wxArrayString& filepaths )	// Get multiple filepaths
{
size_t count;
wxArrayTreeItemIds selectedIds;

count = GDC->GetTreeCtrl()->GetSelections( selectedIds );

for ( size_t c=0; c < count; ++c )
	{	wxDirItemData* data = (wxDirItemData*) GDC->GetTreeCtrl()->GetItemData( selectedIds[c] );
		if ( filesonly && wxDirExists( data->m_path ) ) continue;		// If we only want files, skip dirs
		filepaths.Add( data->m_path );
	}
}

void FileDirDlg::OnEraseBackground( wxEraseEvent& event )	// This is only needed in gtk2 under kde and brain-dead theme-manager, but it doesn't seem to do any harm elsewhere
{
wxClientDC* clientDC = NULL;  // We may or may not get a valid dc from the event, so be prepared
if ( !event.GetDC() ) clientDC = new wxClientDC(  GDC->GetTreeCtrl() );
wxDC* dc = clientDC ? clientDC : event.GetDC();

wxColour colBg = GetBackgroundColour();  // Use the chosen background if there is one
if ( ! colBg.Ok() ) colBg = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW);

dc->SetBackground(wxBrush(colBg, wxSOLID));
dc->Clear();

if ( clientDC ) delete clientDC;
}




