/////////////////////////////////////////////////////////////////////////////
// Name:        MyFiles.cpp
// Purpose:    File-view
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3. Part copied from wxWidgets: wxWindows licence
/////////////////////////////////////////////////////////////////////////////
// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifndef WX_PRECOMP
    #include "wx/log.h"

    #include "wx/app.h"
    #include "wx/frame.h" 
	#include "wx/tokenzr.h"	
    #include "wx/scrolwin.h"
    #include "wx/menu.h"
	#include <wx/dirctrl.h>
#endif

#include "wx/splitter.h"
#include <wx/file.h>
#include "wx/dragimag.h"
#include "wx/config.h"

#include "MyTreeCtrl.h"
#include "MyDirs.h"
#include "MyGenericDirCtrl.h"
#include "Accelerators.h"
#include "MyFrame.h"
#include "Redo.h"
#include "Dup.h"
#include "Archive.h"
#include "MyFiles.h"
#include "Filetypes.h"
#include "Misc.h"							// For FileDirDilg needed by OnProperties

//---------------------------------------------------------------------------------------------------------------------------
#include <pwd.h>
#include <grp.h>

void GetAllUsers( wxArrayString& names )		// Global function, to acquire list of all usernames
{
setpwent();																							// This call initialises what we're doing
do
	{	struct passwd* p = getpwent();													// For every entry, get the user struct
		if ( p == NULL ) 
			{ endpwent(); 																			// If p==NULL we've run out.  Call endpwent() to sign off
				names.Sort(); return; 														// Sort the array & return
			}
		wxString str(p->pw_name, wxConvUTF8);									// Extract this username & add it to the ArrayString
		names.Add( str );
	}
 while (true);
}


void GetProcessesGroupNames( wxArrayString& names )		// Global function, to acquire list of user's groups
{
if ( getuid()==0 )							// If we're root, we want a list of ALL groups, not just ones root happens to be a member of
	{	setgrent();																						// This call initialises what we're doing
		do
			{	struct group* g = getgrent();													// For every entry, get the group struct
				if ( g == NULL ) 
					{ endgrent(); 																		// If g==NULL we've run out.  Call endgrent() to sign off
						names.Sort(); return; 													// Sort the array & return
					}
				wxString str(g->gr_name, wxConvUTF8);								// Extract the group's name & add it to the ArrayString
				names.Add( str );
			}
		 while (true);
	}
												// Otherwise do it the standard way, looking only at the user's groups
int ngroups = getgroups(0, NULL);														// This first call to getgroups() returns their number
if ( !ngroups ) return;

gid_t *idarray	= new gid_t[ngroups * sizeof (gid_t)];							// Now we know the no of groups, make an int array to store them

int result = getgroups(ngroups, idarray);												// This 2nd call to getgroups() fills this array
if ( result != -1 )
	for ( int n=0; n < ngroups; ++n )
		{	struct group* g = getgrgid( idarray [n] );									// For every entry, get the group struct
			if ( g != NULL ) 
				{	wxString str(g->gr_name, wxConvUTF8);							// Extract the group's name & add it to the ArrayString
					names.Add( str );
				}
		}

names.Sort();																							// Sort the array
delete[] idarray;
}

//---------------------------------------------------------------------------------------------------------------------------

static int wxCMPFUNC_CONV FilenameCompareFunction(DataBase **first, DataBase **second)
{
return (*first)->GetFilename().CmpNoCase((*second)->GetFilename());
}

static int wxCMPFUNC_CONV FilenameReverseCompareFunction(DataBase **second, DataBase **first)	// // Same as above but reverse the sort
{
return (*first)->GetFilename().CmpNoCase((*second)->GetFilename());
}

static int wxCMPFUNC_CONV CompareFunction(DataBase **first, DataBase **second)	// // Sort based on the string inserted into sortstring
{
int ans;
if ( (ans = (*first)->sortstring.CmpNoCase((*second)->sortstring)) == 0 )	
	return (*first)->GetFilename().CmpNoCase((*second)->GetFilename());				// If they match, re-sort by filename
	
return ans;
}

static int wxCMPFUNC_CONV ReverseCompareFunction(DataBase **second, DataBase **first)
{
int ans;
if ( (ans = (*first)->sortstring.CmpNoCase((*second)->sortstring)) == 0 )	
	return (*first)->GetFilename().CmpNoCase((*second)->GetFilename());				// If they match, re-sort by filename
	
return ans;
}

static int wxCMPFUNC_CONV SortintCompareFunction(DataBase **first, DataBase **second)	// // Sort based on contents of sortint
{
if ( (*first)->sortint == (*second)->sortint )	
	return (*first)->GetFilename().CmpNoCase((*second)->GetFilename());				// If they match, re-sort by filename
	
return (*first)->sortint > (*second)->sortint;
}

static int wxCMPFUNC_CONV SortintReverseCompareFunction(DataBase **second, DataBase **first)
{
if ( (*first)->sortint == (*second)->sortint )	
	return (*first)->GetFilename().CmpNoCase((*second)->GetFilename());				// If they match, re-sort by filename
	
return (*first)->sortint > (*second)->sortint;
}




const int HEADER_HEIGHT = 12;							// // The ht of the header column.  Originally 23

FileGenericDirCtrl::FileGenericDirCtrl( wxWindow* parent, const wxWindowID id, const wxString& START_DIR , const wxPoint& pos, const wxSize& size,
																																																 long style, bool full_tree, const wxString& name )
							:   MyGenericDirCtrl( parent, (MyGenericDirCtrl*)this, id, START_DIR ,  pos,  size ,  style , wxEmptyString, 0, name, ISRIGHT, full_tree )
{
SizeIsValid = false; SelectedCumSize = 0;

headerwindow = new TreeListHeaderWindow( this, NextID++, (MyTreeCtrl*)GetTreeCtrl() );
((MyTreeCtrl*)GetTreeCtrl())->headerwindow = headerwindow;

CreateAcceleratorTable();
}

void FileGenericDirCtrl::CreateAcceleratorTable()
{
int AccelEntries[] = {	SHCUT_CUT, SHCUT_COPY, SHCUT_PASTE, SHCUT_HARDLINK, SHCUT_SOFTLINK, SHCUT_TRASH,
										SHCUT_DELETE, SHCUT_RENAME, SHCUT_NEW, SHCUT_UNDO, SHCUT_REDO,
										SHCUT_REFRESH, SHCUT_FILTER, SHCUT_TOGGLEHIDDEN, SHCUT_PROPERTIES
									};
const size_t shortcutNo = sizeof(AccelEntries)/sizeof(int);
MyFrame::mainframe->AccelList->CreateAcceleratorTable( this, AccelEntries,  shortcutNo );
}

void FileGenericDirCtrl::RecreateAcceleratorTable()
{
				// It would be nice to be able to empty the old accelerator table before replacing it, but trying caused segfaulting!
CreateAcceleratorTable();																						// Make a new one, with up-to-date data
((MyTreeCtrl*)GetTreeCtrl())->RecreateAcceleratorTable();									// Tell the tree to do likewise
}

void FileGenericDirCtrl::DoSize()
{
int w, h; GetClientSize(&w, &h);

GetTreeCtrl()->SetSize(0, HEADER_HEIGHT + 1, w, h - HEADER_HEIGHT - 1);	// Set the treectrl 'body' size
headerwindow->SetSize(0, 0, w, HEADER_HEIGHT);											// Set the header-window size

headerwindow->SizeColumns();																				//   and adjust the columns intelligently. Also does the scrollbars
}

void FileGenericDirCtrl::ShowContextMenu(wxContextMenuEvent& event)
{
if ( ((MyTreeCtrl*)GetTreeCtrl())->QueryIgnoreRtUp() ) return;				// If we've been Rt-dragging & now stopped, we don't want a context menu

int flags;													// First, make sure there is a selection, unless the user Rt-clicked in mid-air
if ( GetPath().IsEmpty() )																				// If there is no selection, select one if that's what the user probably wants
	{	wxTreeItemId item = GetTreeCtrl()->HitTest( GetTreeCtrl()->ScreenToClient( event.GetPosition() ), flags );
		if ( item.IsOk() )																					// Find out if we're on the level of a tree item
			GetTreeCtrl()->SelectItem( item );													// If so, select it
	}
 else
	if ( GetPath() == startdir )					// If the selection is startdir, yet the click is directly on an item, select it.  That's probably what the user wants
		{ wxTreeItemId item = GetTreeCtrl()->HitTest( GetTreeCtrl()->ScreenToClient( event.GetPosition() ), flags );
			if ( item.IsOk()  && ! (flags & wxTREE_HITTEST_ONITEMRIGHT) ) 	// Find out if we're on a tree item (& not in the space to the right of it)
				GetTreeCtrl()->SelectItem( item );
		}

bool show_Open = false;
bool show_OpenWith = false;
bool show_OpenWithKdesu = false;
bool ItsaDir;
bool WithinArchive = arcman->IsArchive();												// Are we rummaging around inside a virtual archive?

		// See if the user has cleverly filtered out *all* of the pane's content. If so, use startdir (otherwise he can't get a contextmenu to reset the filter!)
wxString filepath( GetPath() );
NoVisibleItems = filepath.IsEmpty();
if ( NoVisibleItems )  filepath = startdir;														// The NoVisibleItems flag is used by updateUI to disable menuitems like Del, Cut

FileData stat(filepath );
if ( WithinArchive ) ItsaDir = arcman->GetArc()->Getffs()->GetRootDir()->FindSubdirByName( filepath ) != NULL;	// We can't test for dir-ness except by finding one with this name
 else 
 	{	ItsaDir = stat.IsDir();																			// We need to know if this file is in fact a Dir.  If so, we'll call it such in the menu
		if ( stat.IsSymlink() ) filepath = stat.GetUltimateDestination();				// If it's a symlink, make filepath the (ultimate) destination.  This makes FiletypeManager behave more sensibly
	}
	
FiletypeManager manager;
manager.Init(filepath);																					// Use specialist class to get file info. NB this won't work inside archives; we'll do those in a minute
char buf[3];																								// A char array will hold the answers
buf[0] = buf[1] = buf[2] = 0;
if ( manager.QueryCanOpen( buf ) )															// See if we can open the file by double-clicking, ie executable or ext with default command
	{	show_OpenWith = buf[0];																	// No, but the file exists
		show_Open = buf[1];																			// Yes, we can open
		show_OpenWithKdesu = buf[2];															// Yes, but we need root privilege
	}

wxMenu menu(wxT(""));

int Firstsection[] = { SHCUT_CUT, SHCUT_COPY, SHCUT_PASTE, SHCUT_SOFTLINK, SHCUT_HARDLINK, wxID_SEPARATOR };
if ( WithinArchive )
	{	for ( size_t n=0; n < 3; ++n )																// We start with a section where it's easier to insert with a loop. Just the 1st 3 for archives
			MyFrame::mainframe->AccelList->AddToMenu( menu, Firstsection[n] );
		if ( ! ItsaDir )
			{ show_Open = manager.QueryCanOpenArchiveFile( filepath ); show_OpenWith = true; }
		
		MyFrame::mainframe->AccelList->AddToMenu( menu, wxID_SEPARATOR );
		MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_ARCHIVE_EXTRACT, _("Extract from archive") );
		MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_DELETE, _("Delete from archive") );
		if ( ItsaDir )	
			{	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_RENAME,  _("Rename Dir within archive") );
				MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_DUP,  _("Duplicate Dir within archive") );
			}
		 else
			{	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_RENAME,  _("Rename File within archive") );
				MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_DUP,  _("Duplicate File within archive") );
			}
	}
  else
	{	for ( size_t n=0; n < 6; ++n )																		// We start with a section where it's easier to insert with a loop
			MyFrame::mainframe->AccelList->AddToMenu( menu, Firstsection[n] );
		
		wxString delmsg, trashmsg;
		if ( ItsaDir )						{ delmsg =  _("Delete Dir"); trashmsg = _("Send Dir to Trashcan"); }
		 else	if ( stat.IsSymlink() )	{ delmsg =  _("Delete Symlink"); trashmsg = _("Send Symlink to Trashcan"); }
					else						{ delmsg =  _("Delete File"); trashmsg = _("Send File to Trashcan"); }
		
		MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_TRASH, trashmsg );	// A GTK1.2 bug prevents the display of Del on menus. Still works though, & OK in GTK2
		MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_DELETE, delmsg );
	}

menu.AppendSeparator();

if ( ! WithinArchive && stat.IsSymlink() )														// If we're dealing with a symlink, add the relevant GOTO commands
	{	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_GOTO_SYMLINK_TARGET );
		if ( stat.IsSymlinktargetASymlink() ) MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_GOTO_ULTIMATE_SYMLINK_TARGET );
		menu.AppendSeparator();
	}

if ( show_Open )	 menu.Append(SHCUT_OPEN, _("Open"));
if ( show_OpenWithKdesu )
	{	wxArrayString Commands;																	// Invent an array to pass in OfferChoiceOfApplics.  Using CommandArray directly segfaulted
		if ( manager.OfferChoiceOfApplics(Commands) && !show_OpenWith )	// If there are known applics to offer for this filetype, & we're not about to do this below
			{	wxMenu* submenu = new wxMenu;											//  make a submenu & populate it with the applics
				for ( size_t n=0; n < manager.Array.GetCount(); ++n )
					submenu->Append( SHCUT_RANGESTART + n, manager.Array[n] );
				submenu->AppendSeparator();													// Finish the submenu with a separator followed by an 'Other' option -> OpenWith
				submenu->Append(SHCUT_OPENWITH, _("Other . . ."));
				menu.Append(NextID++, _("Open using root privileges with . . . . "),submenu);
				CommandArray = Commands;													// Store any commands
			}
		  else	menu.Append(SHCUT_OPENWITH_KDESU, _("Open using root privileges"));  // Give kdesu option if appropriate
	}

if ( show_OpenWith )	 
	{	wxArrayString Commands;																	// Invent an array to pass in OfferChoiceOfApplics.  Using CommandArray directly segfaulted
		if ( manager.OfferChoiceOfApplics(Commands) )								// If there are known applics to offer for this filetype
			{	wxMenu* submenu = new wxMenu;											//  make a submenu & populate it with the applics
				for ( size_t n=0; n < manager.Array.GetCount(); ++n )
					submenu->Append( SHCUT_RANGESTART + n, manager.Array[n] );
				submenu->AppendSeparator();													// Finish the submenu with a separator followed by an 'Other' option -> OpenWith
				submenu->Append(SHCUT_OPENWITH, _("Other . . ."));
				menu.Append( NextID++, _("Open with . . . . "), submenu );
				CommandArray = Commands;													// Store any commands
			}
		  else	menu.Append(SHCUT_OPENWITH, _("Open with . . . . "));
	}
if ( show_Open || show_OpenWith || show_OpenWithKdesu )					// If there's an OpenSomething, add another separator
 	menu.AppendSeparator();

MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_REFRESH);
if ( ! WithinArchive ) 
	{	if ( ItsaDir )
			{	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_RENAME,  _("Rename Dir") );
				MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_DUP, _("Duplicate Dir") );
			}
		 else
			{	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_RENAME,  _("Rename File") );
				MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_DUP, _("Duplicate File") );
			}
		MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_NEW);
		menu.AppendSeparator();

									// Now for Archive/Compression entries.  Don't add any if there's no selection, otherwise choose depending on selected filetypes
		wxArrayString paths; bool arc=false, comp=false, ordinaryfile=false;
		size_t count = GetMultiplePaths( paths );
		for ( size_t n=0; n < count; ++n )																// Look thru every selection, seeing if there're any archives, compressed files, ordinary ones
			switch ( Archive::Categorise( paths[n] ) )
				{	case NotArchive:					ordinaryfile = true; break;
					case taronly: 							arc = true; break;
					case gzip: case bzip:				comp = true; break;
						default:								comp = true; arc = true;				// Default catches targz, tarbz, zip
				}
		if ( arc || comp )					MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_ARCHIVE_EXTRACT );
		if ( ordinaryfile || comp )	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_ARCHIVE_CREATE );
		if ( arc )									MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_ARCHIVE_APPEND );
		if ( arc || comp )					MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_ARCHIVE_TEST );
		if ( ordinaryfile )					MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_ARCHIVE_COMPRESS );
		if ( count )								menu.AppendSeparator();
	}

MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_FILTER );
if ( GetShowHidden() )
	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_TOGGLEHIDDEN, _("Hide hidden dirs and files") );
 else
 	MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_TOGGLEHIDDEN, _("Show hidden dirs and files") );
menu.AppendSeparator();

wxMenu* colmenu = new wxMenu;
int columns[] = { SHCUT_SHOW_COL_EXT, SHCUT_SHOW_COL_SIZE, SHCUT_SHOW_COL_TIME, SHCUT_SHOW_COL_PERMS, SHCUT_SHOW_COL_OWNER,
								SHCUT_SHOW_COL_GROUP, SHCUT_SHOW_COL_LINK, wxID_SEPARATOR, SHCUT_SHOW_COL_ALL, SHCUT_SHOW_COL_CLEAR };
for ( size_t n=0; n < 10; ++n )																	// This submenu is easy to insert with a loop
	{	MyFrame::mainframe->AccelList->AddToMenu( *colmenu, columns[n], wxEmptyString,  (n < 7 ? wxITEM_CHECK : wxITEM_NORMAL) );
	#ifdef __WXX11__
		if ( n < 7 ) colmenu->Check( columns[n], ! GetHeaderWindow()->IsHidden(n+1) );	// The usual updateui way of doing this doesn't work in X11
	#endif
	}
menu.Append(NextID++, _("Columns to Display"), colmenu);

int Secondsection[] = {  wxID_SEPARATOR, SHCUT_SPLITPANE_VERTICAL, SHCUT_SPLITPANE_HORIZONTAL, SHCUT_SPLITPANE_UNSPLIT, wxID_SEPARATOR,
											SHCUT_REPLICATE, SHCUT_SWAPPANES, wxID_SEPARATOR };
size_t ptrsize = sizeof(Secondsection)/sizeof(int);
for ( size_t n=0; n < ptrsize; ++n )																		// And another section where it's easier to insert with a loop
	MyFrame::mainframe->AccelList->AddToMenu( menu, Secondsection[n] );

MyFrame::mainframe->Layout->m_notebook->LaunchFromMenu->LoadMenu( &menu );	// Now add the User-defined Tools submenu

if ( ! WithinArchive ) { menu.AppendSeparator(); MyFrame::mainframe->AccelList->AddToMenu( menu, SHCUT_PROPERTIES ); }

wxPoint pt = event.GetPosition();
ScreenToClient(&pt.x, &pt.y);

PopupMenu(&menu, pt.x, pt.y);
}

void FileGenericDirCtrl::OpenWithSubmenu(wxCommandEvent& event)		// Events from the submenu of the context menu end up here 
{
if ( ! CommandArray.GetCount() ) return;
if ( event.GetId() >= SHCUT_RANGEFINISH )  return;									// Check for rubbish

wxString open = CommandArray[event.GetId() - SHCUT_RANGESTART];
if ( open.IsEmpty() ) return;

FileData stat( GetPath() );																				// We need to stat again to check if read-access: all the old info was lost with Context menu
bool usingsu = ! arcman->IsArchive() && ! stat.CanTHISUserRead();			// stat will be invalid if this is inside an archive: I don't believe we can fail to have permissions though	
FiletypeManager::Openfunction( open, usingsu );											// Do the rest in FiletypeManager submethod, shared by 'Open with' etc
}

void FileGenericDirCtrl::NewFile(wxCommandEvent& event)						// Create a new File or Dir
{
wxString path, name;
int ItsADir, flag = 0;
																		// Where to put the new item
path = GetPath();																							// Get path of selected item, in case it's a subdir
if ( path.IsEmpty() )  path = startdir;																// If no selection, use "root"
	else
		{	FileData fd( path );
			if ( ! fd.IsDir() )  path = path.BeforeLast( wxFILE_SEP_PATH );		// If it's not a dir, it's a filename; remove this to leave the path
		}
if ( path.Last() != wxFILE_SEP_PATH )  path += wxFILE_SEP_PATH; 		// Whichever, we almost certainly need to add a terminal '/'				

FileData DestinationFD( path );
if ( ! DestinationFD.CanTHISUserWriteExec() )												// Make sure we have permission to write to the dir
	{	wxString msg;
		if  ( arcman != NULL && arcman->IsArchive() )  msg = _("I'm afraid you can't Create inside an archive.\nHowever you can create the new element outside, then Move it in.");
			else msg = _("I'm afraid you don't have permission to Create in this directory");
		wxMessageDialog dialog(this, msg, _("No Entry!"), wxOK | wxICON_ERROR );
		dialog.ShowModal(); return;
    }

wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("NewFileOrDir"));

do																	// Do the rest in a loop, to give a 2nd chance if there's a name clash etc
	{	
		if ( dlg.ShowModal() != wxID_OK )		return;									// If user cancelled, abort
		
		name = ((wxTextCtrl*)dlg.FindWindow(  wxT("FileDir_text")))->GetValue();		// Get the desired name
		if ( name.IsEmpty() ) return;
		
		ItsADir = ((wxRadioBox*)dlg.FindWindow(  wxT("FileDir_radiobox")) )->GetSelection();	// Dir or File requested. File is 0, Dir 1
				
		flag = MyGenericDirCtrl::OnNewItem( name, path, (ItsADir==1) );		// Do the rest in MyGenericDirCtrl, as shared with MyDirs version
	}
 while ( flag==wxID_YES );							// Loop if requested: means there was an error & we want another go
 
if ( !flag ) return;
 
wxArrayInt IDs; IDs.Add( GetId() );																// Update the panes
MyFrame::mainframe->OnUpdateTrees( path, IDs );
UnRedoManager::StartCluster();																// Start a new UnRedoManager cluster
UnRedoNewDirFile *UnRedoptr = new UnRedoNewDirFile( name, path, (ItsADir==1), IDs );	// Make a new UnRedoNewDirFile
UnRedoManager::AddEntry( UnRedoptr );												 						// and store it for Undoing 
UnRedoManager::EndCluster();

BriefLogStatus bls( ItsADir==1 ? _("New directory created") : _("New file created") );
}

void FileGenericDirCtrl::OnArchiveAppend(wxCommandEvent& event)
{
MyFrame::mainframe->OnArchiveAppend( event );
}

void FileGenericDirCtrl::OnArchiveTest(wxCommandEvent& event)
{
MyFrame::mainframe->OnArchiveTest( event );
}

void MyGenericDirCtrl::OnDup(wxCommandEvent& WXUNUSED(event))
{
wxArrayString paths; size_t count = GetMultiplePaths( paths );
if ( ! count ) return;
if ( count > 1 ) DoMultipleRename( paths, true );
 else DoRename( true );																								// Because this method can also rename, "true" flags that this is duplication
}

void MyGenericDirCtrl::OnRename(wxCommandEvent& WXUNUSED(event))				// Rename a File or Dir
{
wxArrayString paths; size_t count = GetMultiplePaths( paths );
if ( ! count ) return;
if ( count > 1 ) DoMultipleRename( paths, false );
 else DoRename( false );																								// Because this method can also duplicate, "false" flags that this is a rename
}

void MyGenericDirCtrl::DoRename( bool duplicate)			// Renames (or duplicates) a File or Dir
{
wxString path, origname, newname, oldname, displayname;
int ItsADir, flag = 0;
bool renamingstartdir = false;

origname = GetPath();																								// Get name of selected item
if ( origname.IsEmpty() )  return;																				// If no selection, disappear

while ( origname.Len() > 1 && origname.Right(1) == wxFILE_SEP_PATH ) origname.RemoveLast();		// Avoid '/' problems

if ( origname == wxT("/") )
	{	wxString msg;		
		if ( duplicate )	msg = _("I don't think you really want to duplicate root, do you");
		  else					msg = _("I don't think you really want to rename root, do you");
		wxMessageDialog dialog(this, msg, _("Oops!"), wxICON_ERROR ); dialog.ShowModal(); return;
    }

FileData fd( origname ); ItsADir = fd.IsDir();																// Store whether it's a dir; needed later

path = origname.BeforeLast( wxFILE_SEP_PATH );													// Get the path for this dir or file:  SHOULD work for both
path += wxFILE_SEP_PATH; 																					// Add back the terminal '/'				

							// We need tocheck whether we're renaming within an archive: if so things are different
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane();	// Yes, I know we're inside MyGenericDirCtrl, but we want the correct pane
bool WithinArc = (active != NULL && active->arcman != NULL && active->arcman->IsArchive());

FileData DestinationFD( path );
if ( ! WithinArc && ! DestinationFD.CanTHISUserWriteExec() )									// Make sure we have permission to write to the dir
	{	wxString msg;		
		if ( duplicate )	msg = _("I'm afraid you don't have permission to Duplicate in this directory");
		  else				msg = _("I'm afraid you don't have permission to Rename in this directory");
		wxMessageDialog dialog(this, msg, _("No Entry!"), wxOK | wxICON_ERROR );
		dialog.ShowModal(); return;
    }
		
wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe,  wxT("RenameDlg"));

								// Use the filename in the dialog title, prefixed by .../
oldname = origname.AfterLast( wxFILE_SEP_PATH );
displayname = wxFILE_SEP_PATH + oldname;
if ( displayname.Len() < origname.Len() )  displayname =  wxT("...") + displayname; // Don't want a prefix for eg /home

if ( duplicate ) 	dlg.SetTitle( _("Duplicate   ") + displayname );									// Use the title as a way to confirm we're trying to dup the right item
 else 				dlg.SetTitle(  _("Rename   ") + displayname );									//   or rename it
 
do							// Do the rest in a loop, to give a 2nd chance if there's a name clash etc
	{	wxTextCtrl* text = (wxTextCtrl*)dlg.FindWindow( wxT("newfilename") );
		text->Clear(); 																						// In case this isn't the first iteration of the loop
		text->WriteText(oldname); text->SetSelection(-1, -1); text->SetFocus(); text->SetInsertionPoint(0);	// Insert the original name & highlight it. Then move to start of text, so that END will  unselect: this being a likely thing to want to do
		
		if ( dlg.ShowModal() != wxID_OK )		return;													// If user cancelled, abort
		
		newname = text->GetValue();																				// Get the desired name
		if ( newname.IsEmpty() ) return;
		if ( newname == oldname )
			{	if ( duplicate ) 	wxMessageBox(_("No, the idea is that you supply a NEW name"));
				  else 					wxMessageBox(_("No, the idea is that you CHANGE the name")); 
				flag = wxID_YES; continue; 
			}
			
		if ( ItsADir )
			{	oldname += wxFILE_SEP_PATH; newname += wxFILE_SEP_PATH; }		// Make sure they're recognised as dirs 
		
		if ( ! WithinArc )		flag = Rename( path, oldname, newname, ItsADir, duplicate );	// Do the actual rename/dup in a submethod
#if wxVERSION_NUMBER >= 2504
		 else															// Things are very different within an archive
		 	{	wxArrayString filepaths, newpaths; 	// Make arrays so as to cope with renaming a dir: all its descendants need to be listed too
				filepaths.Add( path+oldname ); newpaths.Add( path+newname ); 
				if ( active->arcman->GetArc()->Alter( filepaths, duplicate ? arc_dup : arc_rename, newpaths ) )	// and do the rename
					{	UnRedoManager::StartCluster();
						wxArrayInt IDs; IDs.Add( GetId() );

						UnRedoArchiveRen* UnRedoptr = new UnRedoArchiveRen( filepaths, newpaths, IDs, active, duplicate );
						UnRedoManager::AddEntry( UnRedoptr );
						UnRedoManager::EndCluster();
						MyFrame::mainframe->OnUpdateTrees( path, IDs );
						DoBriefLogStatus( 1, wxEmptyString, duplicate ?_("duplicated") :  _("renamed") );
						return;
					}
				if ( wxMessageBox(_("Sorry, that didn't work. Try again?"), _("Oops"), wxYES_NO ) != wxYES )  return;
				flag = wxID_YES;																						// This flags to retry	
			}
#endif // #if wxVERSION_NUMBER >= 2504
	}
 while ( flag==wxID_YES );					// Loop if requested: means there was an error & we want another go
 
if ( ! flag ) return;

if ( ItsADir && !duplicate && origname == wxGetCwd() )								// If we've just renamed the cwd, change the cwd.  Otherwise Ctrl-T will fail
            wxSetWorkingDirectory(path + newname);

wxArrayInt IDs; IDs.Add( GetId() );																	// Update the panes
if ( origname == startdir )																				// If we've just renamed startdir, need to do things differently
	{	wxString newstartdir = path + newname.BeforeLast(wxFILE_SEP_PATH);	// We know that newname ends in /, we just put it there!
		MyFrame::mainframe->OnUpdateTrees( origname, IDs, newstartdir ); 	// Newstartdir will be substituted for the old one
		renamingstartdir = true;
	}
 else MyFrame::mainframe->OnUpdateTrees( path, IDs );
 
UnRedoManager::StartCluster();																	// Start a new UnRedoManager cluster
if ( duplicate )
	{	UnRedoDup* UnRedoptr = new UnRedoDup( path, path, oldname, newname, ItsADir, IDs );	
		UnRedoManager::AddEntry( UnRedoptr );
		DoBriefLogStatus( 1, wxEmptyString, _("duplicated") ); 
	}
 else
	{	UnRedoRen* UnRedoptr;
		if ( ItsADir )											// To get the UnRedoRen right, if it's a dir, we have to merge path & names
			{	newname = path + newname; 
				if ( renamingstartdir )
					UnRedoptr = new UnRedoRen( origname, newname, myNullString, newname, ItsADir, IDs );	// If renaming startdir, use 4th string as notification
				 else
					UnRedoptr = new UnRedoRen( origname, newname, myNullString, myNullString, ItsADir, IDs );// Otherwise make a new UnRedoRen the standard way
		
				wxString filepath = newname;		// While we're in a ItsADir area, we need to select the new dir, else nothing is selected & the fileview is empty
				while ( filepath.Right(1) == wxFILE_SEP_PATH && filepath.Len() > 1 )  filepath.RemoveLast();	// We don't want a terminal '/' in FindIdForPath
				wxTreeItemId item = FindIdForPath( filepath );									// Find the new dir's id, & use this to select it
				if ( item.IsOk() )   GetTreeCtrl()->SelectItem( item );
			}
		  else UnRedoptr = new UnRedoRen( path, path, oldname, newname, ItsADir, IDs );

		DoBriefLogStatus( 1, wxEmptyString, _("renamed") );
		UnRedoManager::AddEntry( UnRedoptr );
	}
UnRedoManager::EndCluster();
}

void MyGenericDirCtrl::DoMultipleRename( wxArrayString& OriginalFilepaths, bool duplicate )		// Renames (or duplicates) multiple files/dirs
{
if ( ! OriginalFilepaths.GetCount() ) return;

							// We need tocheck whether we're renaming within an archive: if so things are different
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane();	// Yes, I know we're inside MyGenericDirCtrl, but it doesn't get IsArchive() right otherwise
bool WithinArc = (active != NULL && active->arcman != NULL && active->arcman->IsArchive());

FileData DestinationFD( OriginalFilepaths[0].BeforeLast( wxFILE_SEP_PATH ) );
if ( ! WithinArc && ! DestinationFD.CanTHISUserWriteExec() )						// Make sure we have permission to write to the dir
	{	wxString msg;		
		if ( duplicate )	msg = _("I'm afraid you don't have permission to Duplicate in this directory");
		  else					msg = _("I'm afraid you don't have permission to Rename in this directory");
		wxMessageDialog dialog(this, msg, _("No Entry!"), wxOK | wxICON_ERROR );
		dialog.ShowModal(); return;
    }

for ( size_t count = OriginalFilepaths.GetCount(); count > 0; --count )
	{	wxString name = OriginalFilepaths[count-1];
		while ( name.Len() > 1 && name.Right(1) == wxFILE_SEP_PATH ) name.RemoveLast();	// Avoid '/' problems
		
		if ( name == wxT("/") )
			{	wxString msg;		
				if ( duplicate )	msg = _("I don't think you really want to duplicate root, do you");
					else				msg = _("I don't think you really want to rename root, do you");
				wxMessageDialog dialog(this, msg, _("Oops!"), wxICON_ERROR ); dialog.ShowModal();
				OriginalFilepaths.RemoveAt( count-1 );
				if ( OriginalFilepaths.GetCount()==1 )	// If as a result of removing /, we no longer have a multiple situation, revert to the single method
					{ SetPath( OriginalFilepaths[0] ); return DoRename( duplicate); }
			}
	}

MultipleRename Mult( MyFrame::mainframe, OriginalFilepaths, duplicate );
wxArrayString NewFilepaths = Mult.DoRename();											//This fills NewFilepaths with the altered filenames
if ( NewFilepaths.IsEmpty() )	return;
#if wxVERSION_NUMBER >= 2504
if ( WithinArc )																									// If in archive, do things elsewhere
	{	if ( active->arcman->GetArc()->Alter( OriginalFilepaths, duplicate ? arc_dup : arc_rename, NewFilepaths ) )
			{	UnRedoManager::StartCluster();
				wxArrayInt IDs; IDs.Add( GetId() );
				wxString path = OriginalFilepaths[0].BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH;
				
				UnRedoArchiveRen* UnRedoptr = new UnRedoArchiveRen( OriginalFilepaths, NewFilepaths, IDs, active, duplicate );
				UnRedoManager::AddEntry( UnRedoptr );
				UnRedoManager::EndCluster();
				MyFrame::mainframe->OnUpdateTrees( path, IDs );
				DoBriefLogStatus( OriginalFilepaths.GetCount(), wxEmptyString, duplicate ? _("duplicated") :_("renamed") );
			}
		return; 
	} 
#endif // #if wxVERSION_NUMBER >= 2504

int successes = 0;
UnRedoManager::StartCluster();																	// Start a new UnRedoManager cluster
for ( size_t n = NewFilepaths.GetCount(); n > 0 ; --n )									// Loop backwards, in case we're renaming both a dir & its children: the dir should be before children in the array
	{	bool renamingstartdir = false;
		wxString origfilepath = OriginalFilepaths[n-1], newfilepath = NewFilepaths[n-1];
		if ( newfilepath.Right(1) == wxFILE_SEP_PATH ) newfilepath.RemoveLast();	// Make sure we can do AfterLast in a minute to get the filename: can't use FileData before the file exists
		FileData Oldfd( origfilepath ); bool ItsADir = Oldfd.IsDir();	
		wxString oldname = Oldfd.GetFilename(), newname = newfilepath.AfterLast(wxFILE_SEP_PATH), path = Oldfd.GetPath();
		if ( path.Right(1) != wxFILE_SEP_PATH ) path += wxFILE_SEP_PATH;
		if ( ItsADir )
			{	oldname += wxFILE_SEP_PATH; newname += wxFILE_SEP_PATH; }	// Make sure they'll be recognised as dirs 

		Rename( path, oldname, newname, ItsADir, duplicate );						// Do the actual rename/dup in a submethod
		++successes;

		if ( ItsADir && !duplicate && origfilepath == wxGetCwd() )						// If we've just renamed the cwd, change the cwd.  Otherwise Ctrl-T will fail
            wxSetWorkingDirectory( newfilepath );

		wxArrayInt IDs; IDs.Add( GetId() );															// Update the panes
		if ( origfilepath == startdir )																		// If we've just renamed startdir, need to do things differently
			{	wxString newstartdir = path + newname.BeforeLast(wxFILE_SEP_PATH);	// We know that newname ends in /, we just put it there!
				MyFrame::mainframe->OnUpdateTrees( origfilepath, IDs, newstartdir ); 		// Newstartdir will be substituted for the old one
				renamingstartdir = true;
			}
		 else MyFrame::mainframe->OnUpdateTrees( path, IDs );

		if ( duplicate )
			{	UnRedoDup* UnRedoptr = new UnRedoDup( path, path, oldname, newname, ItsADir, IDs );	
				UnRedoManager::AddEntry( UnRedoptr );
			}
		 else
			{	UnRedoRen* UnRedoptr;
				if ( ItsADir )											// To get the UnRedoRen right, if it's a dir, we have to merge path & names
					{	newname = path + newname; 												// Reconstitute the full filepath. Could have used newfilepath, but this way we know there's a terminal /
						if ( renamingstartdir )
							UnRedoptr = new UnRedoRen( origfilepath, newname, myNullString, newname, ItsADir, IDs );	// If renaming startdir, use 4th string as notification
						 else
							UnRedoptr = new UnRedoRen( origfilepath, newname, myNullString, myNullString, ItsADir, IDs );// Otherwise make a new UnRedoRen the standard way
					}
				 else UnRedoptr = new UnRedoRen( path, path, oldname, newname, ItsADir, IDs );
					
				UnRedoManager::AddEntry( UnRedoptr );
			}			
	}

if ( fileview==ISLEFT )
	{	SetPath( startdir );										// We need to select the (possibly new) startdir, else nothing is selected & the fileview is empty
		ReCreateTreeFromSelection();
	}

UnRedoManager::EndCluster();
DoBriefLogStatus( successes, wxEmptyString, duplicate ? _("duplicated") :_("renamed") );
}

int MyGenericDirCtrl::OnNewItem( wxString& newname, wxString& path, bool ItsADir )
{
if ( (newname == wxT(".")) || (newname == wxT("..")) ||  (newname == wxT("/")) )
	{	wxMessageDialog dialog(this, _("Sorry, this name is Illegal\n        Try again?"), _("Oops!"), wxYES_NO |wxICON_ERROR );
		int ans = dialog.ShowModal();
		if ( ans==wxID_YES )        return wxID_YES;
		return false;
    }

wxString filepath = path + newname;										// Now we can combine path & name
		
wxLogNull log;

FileData fd(filepath);
if ( fd.IsValid() )
	{	wxMessageDialog dialog(this, _("Sorry, One of these already exists\n            Try again?"), _("Oops!"), wxYES_NO |wxICON_ERROR );
		int ans = dialog.ShowModal();
		if ( ans==wxID_YES )        return wxID_YES;
		return false;
    }

bool success;
if ( ItsADir )	
		success = wxFileName::Mkdir( filepath, 0777, wxPATH_MKDIR_FULL );	// If a dir is requested. wxPATH_MKDIR_FULL means intermediate dirs are created too
 else
 	{	wxFile file; success = file.Create( filepath ); }											// else ditto a file

if ( !success )
	{	wxMessageDialog dialog(this, _("Sorry, for some reason this operation failed\nTry again?"),  _("Oops!"), wxYES_NO |wxICON_ERROR );
		int ans = dialog.ShowModal();
		if ( ans==wxID_YES )        return wxID_YES;
		return false;
    }

return true;
}

int MyGenericDirCtrl::Rename( wxString& path, wxString& origname, wxString& newname, bool ItsADir, bool dup /*=false*/ )
{
int answer;

if ( (newname == wxT(".")) || (newname == wxT("..")) ||  (newname == wxT("/")) )
	{	wxMessageDialog dialog(this, _("Sorry, this name is Illegal\n        Try again?"), _("Oops!"), wxYES_NO |wxICON_ERROR );
		int ans = dialog.ShowModal();
		if ( ans==wxID_YES )        return wxID_YES;
		return false;
    }

wxLogNull log;

wxString filepath = path + newname;																// Combine path & name, & check for pre-existence

FileData newfilepath( filepath );
if ( newfilepath.IsValid() )
	{	wxMessageDialog dialog(this, _("Sorry, that name is already taken.\n            Try again?"), _("Oops!"), wxYES_NO | wxICON_ERROR );
		int ans = dialog.ShowModal();
		if ( ans==wxID_YES )        return wxID_YES;
		return false;
    }

wxString OriginalFilepath = path + origname;
CheckDupRen CheckIt( this, OriginalFilepath, myNullString );	// Make an instance of the class that tests for Permission problems etc
CheckIt.WhattodoifCantRead = Unknown;
FileData stat( OriginalFilepath );
if ( stat.IsDir() )																		// If OriginalFilepath is a dir, see if it has offspring
	{	wxDir d( OriginalFilepath );
		CheckIt.IsMultiple = ( d.HasSubDirs() || d.HasFiles() );		// If so, there are >1 items to test so use different dialogs in case of failure
	}

if ( ! CheckIt.CheckForAccess( dup ) )  return false;					// This checks for Read permission failure.  Pass dup, as if false (renaming) we don't want to check dir contents, only the dir itself

										// OK, we can go ahead with the Rename or Dup, subcontracting as CheckDupRen is already set up to do so
if ( ItsADir )
	{	wxString origpath = path + origname, newpath = path + newname;
		answer = CheckDupRen::RenameDir( origpath, newpath, dup );
	}
else
	answer = CheckDupRen::RenameFile( path, origname, newname, dup );

if ( !answer )
	{	wxMessageDialog dialog(this, _("Sorry, for some reason this operation failed\nTry again?"),  _("Oops!"), wxYES_NO |wxICON_ERROR );
		int ans = dialog.ShowModal();
		if ( ans==wxID_YES )        return wxID_YES;
		return false;
    }

return answer;
}

void FileGenericDirCtrl::OnToggleHidden(wxCommandEvent& event)		// Ctrl-H
{
bool hidden = ! GetShowHidden();		// Do the toggle
ShowHidden( hidden );							//  & make it happen.  This time, I'm not telling partner, as it's slightly more likely that we'll want hidden dirs but not files

wxString path = GetPath();
UpdateStatusbarInfo( path );					// Make it appear in the statusbar
}

void FileGenericDirCtrl::OpenWithKdesu( wxCommandEvent& event)	// Execute, or Launch appropriate application for this file-type
{
wxString filepath = GetPath();														// Get the selected item.
if ( filepath.IsEmpty() )	return;													// Check there IS a valid selection 

FileData stat( filepath );																// See if it's a symlink.  If so, try to Open the target, as opening a symlink is unlikely to be sensible
if ( stat.IsSymlink() )  filepath = stat.GetUltimateDestination();

if ( !wxDirExists( filepath ) )															// If it's not a dir, try to process it
	{	wxString WorkingDir = wxGetCwd();									// Change the working dir to the file's path.  Some badly-written programs malfunction otherwise
		wxString thisdir = filepath.BeforeLast( wxFILE_SEP_PATH ); 
		wxSetWorkingDirectory( thisdir );										
		
		FiletypeManager manager( filepath );	
		manager.SetKdesu();
		manager.Open();
		
		wxSetWorkingDirectory(WorkingDir);									// Revert to the original working directory
	}
}

void FileGenericDirCtrl::OnOpen( wxCommandEvent& WXUNUSED(event) )		// From Context menu, passes on to Open
{
wxString filepath = GetPath();														// Get the selected item
DoOpen( filepath );
}

void FileGenericDirCtrl::DoOpen( wxString& filepath )	// From Context menu or DClick in pane or TerminalEm
{
FiletypeManager manager;

if ( filepath.IsEmpty() )	return;														// Check there IS a valid selection 

DataBase* stat;
bool IsArchive = false;
if ( arcman && arcman->IsArchive() ) IsArchive = true;
if ( IsArchive ) stat = new FakeFiledata(filepath );
  else stat = new FileData( filepath );

if ( !IsArchive && stat->IsSymlink() )
	{  filepath = ((FileData*)stat)->GetUltimateDestination();	// See if it's a symlink.  If so, try to Open the target, as opening a symlink is unlikely to be sensible
		if ( filepath.IsEmpty() ) { delete stat; return; }				// unless of course it's a broken symlink!
	}

bool IsDir =  stat->IsDir();
delete stat; if ( IsDir ) return;
																	// If it's not a dir, try to open directly with FiletypeManager::Open()
wxString WorkingDir = wxGetCwd();												// Change the working dir to the file's path.  Some badly-written programs malfunction otherwise
wxString thisdir = filepath.BeforeLast( wxFILE_SEP_PATH ); 
wxSetWorkingDirectory( thisdir );										

manager.Init( filepath );
bool result = manager.Open();

wxSetWorkingDirectory(WorkingDir);												// Revert to the original working directory
if ( !result ) return;																			// If returns false, we're done

								// If we're here, Open returned true, signifying there's applic-choice data in Array/CommandArray
								//	So do a pop-up menu with them
if ( ! CommandArray.GetCount() ) return;										// after a quick check
wxMenu menu(wxT("                                  "));								// Make a menu & populate it with the known applics
for ( size_t n=0; n < manager.Array.GetCount(); ++n )
	menu.Append( SHCUT_RANGESTART + n, _("Open with ") + manager.Array[n] );
menu.AppendSeparator();																// Finish with a separator followed by an 'Other' option -> OpenWith
menu.Append(SHCUT_OPENWITH, _("Other . . ."));

wxPoint pt = ScreenToClient( wxGetMousePosition() );
PopupMenu(&menu, pt.x, pt.y);														// Now do the pop-up.  This relays to OpenWithSubmenu() or OpenWith
}


void FileGenericDirCtrl::OnOpenWith(wxCommandEvent& event)
{
wxString filename = GetPath();															// Get name of selected item
if ( filename.IsEmpty() )  return;															// If no selection, disappear

FiletypeManager manager( filename );													// Use dedicated class to do the work
manager.OpenWith();

}

void FileGenericDirCtrl::OnIdle(wxIdleEvent&WXUNUSED(event))		// Calls UpdateStatusbarInfo()
{
if ( SizeIsValid ) return;

wxString selected = GetPath(); 
if ( selected.IsEmpty() ) return;	//We must check for emptiness, as this method is called before creation is absolutely finished, whereupon arcman->IsArchive() segs

if ( arcman && ! arcman->IsArchive() ) 
	{	if ( SHOW_RECURSIVE_FILEVIEW_SIZE ) SelectedCumSize = GetDirSizeRecursively( selected );	// Do we use the slow & complete recursive method
			else SelectedCumSize = GetDirSize( selected );																				//   or the quick and incomplete one-shot?
	}
	
SizeIsValid = true;
UpdateStatusbarInfo( selected );
}

void FileGenericDirCtrl::UpdateStatusbarInfo(wxString& selected)		// Writes filter info to statusbar(3), &  selection's name & size (or whatever) to statusbar(2)
{
														// First the filter info into the 4th statusbar pane
wxString filterinfo, start, filter = GetFilterString(); if ( filter.IsEmpty() )  filter = wxT("*");	// Translate "" into "*"
start = (DisplayFilesOnly ? wxT(" F") : wxT(" D F"));										// Create a message string, D for dirs if applicable, F for files, R for recursive subdirs
if ( SHOW_RECURSIVE_FILEVIEW_SIZE ) start += (wxT(" R") );					// R for recursive subdirs
start += (GetShowHidden() ? wxT(" H     ") : wxT("     "));								// Add whether we're displaying Hidden ones
filterinfo.Printf( wxT("%s%s"),start.c_str(), filter.c_str() );								// Merge into message string, adding the current filter
MyFrame::mainframe->SetStatusText( filterinfo, 3 ); 

if ( selected.IsEmpty() ) 
	{ MyFrame::mainframe->SetStatusText( wxT(""), 2 ); return; }					// Nothing to display if the first-and-only focus click was on a blank line

wxString text;
DataBase* stat; bool IsArchive = arcman && arcman->IsArchive();
if ( IsArchive )																									// Archives have to be treated differently
	{	FakeFiledata* fd = arcman->GetArc()->Getffs()->GetRootDir()->FindFileByName( selected );
		if ( fd != NULL )
			{	stat = new FakeFiledata( *fd );														// We can't just make a new FakeFiledata: the size etc would be unknown; so copy the real one
				SelectedCumSize = stat->Size();
			}
		 else		
			{	FakeDir* fd = arcman->GetArc()->Getffs()->GetRootDir()->FindSubdirByName( selected );	// If it wasn't a file, try for a dir
				if ( fd == NULL ) return;
				stat = new FakeDir( *fd );
				if ( SHOW_RECURSIVE_FILEVIEW_SIZE ) 	SelectedCumSize = fd->GetTotalSize();
				  else SelectedCumSize = fd->GetSize();
			}
	}
 else  stat = new FileData(selected);
 
if ( ! stat->IsSymlink() )
	{	text.Printf( wxT("%s:  %s"), stat->TypeString().c_str(), selected.AfterLast( wxFILE_SEP_PATH ).c_str());
		if ( SizeIsValid ) 
			{	text += wxString::Format( wxT("   %s"), ParseSize( SelectedCumSize ).c_str() );
				if ( stat->IsDir() ) text += wxString::Format( wxT("%s"), SHOW_RECURSIVE_FILEVIEW_SIZE ? _("  of files and subdirectories") : _(" of files") );
			}
	}
 else
	{	text.Printf( wxT("%s:  %s"), stat->TypeString().c_str(), selected.AfterLast( wxFILE_SEP_PATH ).c_str() ); // If possible, add the symlink target
		if ( stat->IsFake )
			{ if ( ! stat->GetSymlinkDestination().IsEmpty() ) 
					{ wxString target; target.Printf( wxT("  --> %s"), stat->GetSymlinkDestination( true ).c_str() ); text << target; }
			}
		  else
			 if ( stat->GetSymlinkData() )
				{	wxString target( wxT("  --> ") ); size_t len = target.Len();			// Store length without target
					if ( stat->GetSymlinkData()->IsValid() ) 
						target << stat->GetSymlinkDestination( true );						// 'true' says to display relative symlinks as relative
					  else  target << ((FileData*)stat)->GetSymlinkData()->BrokenlinkName;	// If ! IsValid  store any known broken-link name
					if ( target.Len() > len )																	// See if we've successfully appended a target i.e. added something to "  --> "
						text << target;																			// If so, use it
				}
	}

MyFrame::mainframe->SetStatusText( text, 2 ); 
delete stat;
}

void FileGenericDirCtrl::ReloadTree( wxString path, wxArrayInt& IDs )	// Either F5 or after eg a Cut/Paste/Del/UnRedo, to keep treectrl congruent with reality
{
	// First decide whether it's important for this pane to be refreshed ie if it is/was the active pane, or otherwise if a file/dir that's just been acted upon is visible here
wxWindowID pane = GetId();																	// This is us
for ( unsigned int c=0; c < IDs.GetCount(); c++ )								// For every ID passed here,
	if ( IDs[c] == pane )	{ ReCreateTreeFromSelection(); return; }		//  see if it's us. If so, recreate the tree unconditionally

		// There are 2 possibilities.  Either the file/subdir that's just been acted upon is visible, or its not.  Remember, invisible means no TreeItemId exists
wxTreeItemId item = FindIdForPath( path );											// Translate path into wxTreeItemId
if ( ! item.IsOk() )																						// If it's not valid,
	{	if ( path  != wxFILE_SEP_PATH )													//    & it's not root!
			{	while ( path.Last() == wxFILE_SEP_PATH && path.Len() > 1 )	//	 and if it's got a terminal '/' or two 
					path.RemoveLast();															// 		    remove it and try again without.
				item = FindIdForPath( path );													// This is cos the last dir of a branch doesn't have them, yet path DOES, so would never match
			}
	}

if ( item.IsOk() )																						// If it's valid, the item is visible, so definitely refresh it
		ReCreateTreeFromSelection();														// Since this is a fileview, there's no percentage in trying to be clever.  Just redo the tree
}


enum columntype FileGenericDirCtrl::GetSelectedColumn()		// Get the 'selected' col ie that has the up/down icon to control sort-orientation
{
return headerwindow->GetSelectedColumn(); 
}

void FileGenericDirCtrl::SetSelectedColumn(enum columntype col)	// Set the 'selected' col ie that has the up/down icon to control sort-orientation
{
if ( col==headerwindow->GetSelectedColumn() )						 							// See if the clicked column is was already selected
	{ headerwindow->ReverseSelectedColumn(); return; }										// If so, reverse its sort order
 		
headerwindow->SelectColumn( col );																	// Otherwise this must be a new selection
}

void FileGenericDirCtrl::CreateColumns( bool left )		// Sets the column parameters.  The bool distinguishes a LeftTop from a RtBottom pane
{
int* pArray = (left ? ParentTab->tabdata->Lwidthcol : ParentTab->tabdata->Rwidthcol);	// Tabdata widths were loaded here
bool* pBool = (left ? ParentTab->tabdata->LColShown : ParentTab->tabdata->RColShown);	// Tabdata Do-we-show-this-column?

for ( size_t n=0; n < TOTAL_NO_OF_FILEVIEWCOLUMNS; ++n )
	{	wxTreeListColumnInfo col( wxT("") );
		col.SetAlignment(wxTL_ALIGN_LEFT);
		int width = pArray[n];
		if ( width == 0 )																	// If there's no user-defined width for this column, use the default
				width = GetTreeCtrl()->GetCharWidth() * DEFAULT_COLUMN_WIDTHS[n];
		col.SetDesiredWidth( width );												// This is the width to use when/if the column is visible

		width = pBool[n]==true ? width : 0;									// If it's not supposed to be visible atm, zero it
		col.SetWidth( width );
		headerwindow->AddColumn(col);
	}
headerwindow->SizeColumns();

enum columntype selectedcol = (columntype)(left ? ParentTab->tabdata->Lselectedcol : ParentTab->tabdata->Lselectedcol);	// The tabdata's selected col
headerwindow->SelectColumn( selectedcol );

bool reversesort = (left ? ParentTab->tabdata->Lreversesort : ParentTab->tabdata->Rreversesort);	// Was the sort-order reversed
if ( reversesort )    headerwindow->ReverseSelectedColumn();

ReCreateTree();																			// Now we have to redo the tree, or the col selection/reverse doesn't happen
}

void FileGenericDirCtrl::OnColumnSelectMenu(wxCommandEvent& event)
{
headerwindow->OnColumnSelectMenu(event);
}

void FileGenericDirCtrl::HeaderWindowClicked(wxListEvent& event)		// A L click occurred on a header window column to select or reverse-sort
{
int col = event.GetColumn();
if ( col >= 0 && col < (int)headerwindow->GetColumnCount() ) // If the column is valid, change the master column (or reverse-sort if it's the same one) 
	SetSelectedColumn( (enum columntype)col );						// If invalid it'll be from a column "hide & change selected" situation, so we've already done this
ReCreateTree();																			// This makes the change visible
#if ! defined(__WXGTK20__)
headerwindow->Refresh();															// A gtk1.2 bug for a change: without this, the header doesn't scroll back to 0, yet the fileview does
#endif
}


void FileGenericDirCtrl::SortStats( FileDataObjArray& array )		 // Sorts the FileData array according to column selection
{
bool reverse = headerwindow->GetSortOrder();
switch( headerwindow->GetSelectedColumn() )												// Since the col selection defines the sort method required
	{	case filename:			if ( reverse )															// Here we're just sorting on the filepath, so use it direct, reverse or otherwise
														array.Sort( FilenameReverseCompareFunction );
											 else	array.Sort( FilenameCompareFunction ); break;
											 
		case ext:					for ( size_t n = 0; n < array.GetCount(); ++n )			// For each array entry,
												array[n].sortstring = (array[n].GetFilename().Mid(1)).AfterFirst(wxT('.'));	// We're sorting on the filename ext, so put it in sortstring
											if ( reverse )																								// Note the Mid(1) to avoid hidden files being called ext.s!
														array.Sort( ReverseCompareFunction );
											 else	array.Sort( CompareFunction ); break;

		case filesize:				for ( size_t n = 0; n < array.GetCount(); ++n )			// For each array entry,
												array[n].sortint = array[n].Size();					// We're sorting on the size, so put it in sortint
											if ( reverse )
														array.Sort( SortintReverseCompareFunction );
											 else	array.Sort( SortintCompareFunction ); break;
	
		case modtime:			for ( size_t n = 0; n < array.GetCount(); ++n )			// For each array entry,
												array[n].sortint = array[n].ModificationTime();	// We're sorting on an 'int' variable, so put it in sortint
											if ( reverse )
														array.Sort( SortintReverseCompareFunction );
											 else	array.Sort( SortintCompareFunction ); break;
	
		case permissions:		for ( size_t n = 0; n < array.GetCount(); ++n )
												array[n].sortstring = array[n].PermissionsToText();
											if ( reverse )
														array.Sort( ReverseCompareFunction );
											 else	array.Sort( CompareFunction ); break;
											 
		case owner:				for ( size_t n = 0; n < array.GetCount(); ++n )
												{ array[n].sortstring = array[n].GetOwner(); 
													if ( array[n].sortstring.IsEmpty() )  array[n].sortstring = wxT("zzz");
												}
											if ( reverse )
														array.Sort( ReverseCompareFunction );
											 else	array.Sort( CompareFunction ); break;
											 
		case group:				for ( size_t n = 0; n < array.GetCount(); ++n )
												{ array[n].sortstring = array[n].GetGroup(); 
													if ( array[n].sortstring.IsEmpty() )  array[n].sortstring = wxT("zzz");
												}
											if ( reverse )
														array.Sort( ReverseCompareFunction );
											 else	array.Sort( CompareFunction ); break;
											 
		case linkage:				for ( size_t n = 0; n < array.GetCount(); ++n )
												if ( array[n].IsSymlink() )
													array[n].sortstring = array[n].GetSymlinkDestination(); else array[n].sortstring = wxT("zzz");
											if ( reverse )
														array.Sort( ReverseCompareFunction );
											 else	array.Sort( CompareFunction );
	}
	
}

void MyGenericDirCtrl::OnFilter(wxCommandEvent& WXUNUSED(event))
{
wxDialog dlg;
wxXmlResource::Get()->LoadDialog(&dlg, this, wxT("FilterDlg"));

wxCheckBox* FilesChk = (wxCheckBox*)dlg.FindWindow( wxT("FilesOnly") );
if ( fileview==ISLEFT )
	{	wxSizer* sizer = FilesChk->GetContainingSizer();															// If we're a dirview, remove this checkbox, as we can't display files anyway
		FilesChk->Hide();																											// For some reason we seem to have to hide it first
#if wxVERSION_NUMBER <2800
		if ( sizer->Remove( FilesChk ) ) FilesChk->Destroy();
#else
		if ( sizer->Detach( FilesChk ) ) FilesChk->Destroy();
#endif
	}
 else FilesChk->SetValue( DisplayFilesOnly );																		// If it's a fileview, Set the checkbox according to the pane's current status
 	
if ( ! FilterHistory.Count() ) FilterHistory.Add( wxT("*") );														// If there's no history, provide '*'

wxComboBox* combo = (wxComboBox*)dlg.FindWindow( wxT("Combo") );
for (size_t n=0; n < FilterHistory.Count(); ++n ) combo->Append( FilterHistory[n] );				// Add history to combobox
if ( FilterHistory.Count() ) combo->SetValue( FilterHistory[0] );												// Select the last-used value

HelpContext = HCfilter;
if ( dlg.ShowModal() != wxID_OK )	  { HelpContext = HCunknown; return; }
HelpContext = HCunknown;

bool FilesOnly = false;
if ( fileview==ISRIGHT ) FilesOnly = FilesChk->GetValue();													// If we're a fileview, see if we should be displaying files only

wxString selection = combo->GetValue();
if ( ((wxCheckBox*)dlg.FindWindow( wxT("Reset") ))->GetValue() == true )							// If the Reset checkbox is ticked, kill any selection from the combobox
	selection.Empty();
	
int index = FilterHistory.Index( selection, false );																	// See if this command already exists in the history array
if ( index != wxNOT_FOUND )
	FilterHistory.RemoveAt( index );																							// If so, remove it.  We'll add it back into position zero	
FilterHistory.Insert( selection, 0 );																							// Either way, insert into position zero

if ( selection ==  wxT("*") ) selection.Empty();																		// Since emptystring is equivalent

SetFilterArray( selection, FilesOnly );																						// Set this pane's filter
wxArrayInt IDs;	IDs.Add( GetId() ); MyFrame::mainframe->OnUpdateTrees( startdir, IDs );	// & refresh

if ( ((wxCheckBox*)dlg.FindWindow( wxT("AllPanes") ))->GetValue() == true )					// If the AllPanes checkbox is ticked,
	{	wxArrayInt IDs;	IDs.Add( partner->GetId() );
		if ( partner != NULL )																										// Do our twin
			{	partner->SetFilterArray( selection, FilesOnly );
				wxArrayInt IDs;	IDs.Add( partner->GetId() );  MyFrame::mainframe->OnUpdateTrees( partner->startdir, IDs );
			}
			
		if ( ParentTab->splitstatus != unsplit )																				// If the tab is split, do the other side too
			{ 	MyGenericDirCtrl* oppositepane = GetOppositeTwinPane();									// This returns a ptr to the opposite pane's dirview
				if ( oppositepane ==  NULL ) return;
				
				oppositepane->SetFilterArray( selection, FilesOnly );
				wxArrayInt IDs;	IDs.Add( oppositepane->GetId() ); MyFrame::mainframe->OnUpdateTrees( oppositepane->startdir, IDs );
				if ( oppositepane->partner != NULL )	
					{	oppositepane->partner->SetFilterArray( selection, FilesOnly ); 
						wxArrayInt IDs;	IDs.Add( oppositepane->partner->GetId() );  MyFrame::mainframe->OnUpdateTrees( oppositepane->partner->startdir, IDs );
					}
			}
	}
}

void MyGenericDirCtrl::SetFilterArray( const wxString& filterstring, bool FilesOnly /*=false*/ )	// Parse the string into individual filters & store in array
{
GetFilterArray().Clear();

if ( ! filterstring.IsEmpty() )
	{	wxStringTokenizer tkz( filterstring, wxT(",|"), wxTOKEN_STRTOK );	// The filters may be separated by ',' or '|' (but not spaces. What if "ext_with_a_space.ex t")
		while ( tkz.HasMoreTokens() )						
			{	wxString fil	=  tkz.GetNextToken(); fil.Trim(true); fil.Trim(false); GetFilterArray().Add( fil ); }
	}
 else GetFilterArray().Add( wxEmptyString );
 
DisplayFilesOnly = FilesOnly;
}

wxString MyGenericDirCtrl::GetFilterString()		// Turns the filterarray back into "*.jpg | *.png"
{
wxString CombinedFilters;

for ( size_t n=0; n < GetFilterArray().GetCount(); ++n )							// Go thru the list of filters, appending to the CombinedFilters string
	{	wxString fil = GetFilterArray().Item(n);
		if ( ! fil.IsEmpty() )
			{	if ( !CombinedFilters.IsEmpty() )  CombinedFilters += wxT(" | ");	// If there's already data, add a separator
				CombinedFilters += fil;
			}
	}

return CombinedFilters;
}

void MyGenericDirCtrl::OnProperties(wxCommandEvent& WXUNUSED(event))
{
static const wxChar* MiscPermissionStrings[] = { wxT("OtherExec"), wxT("OtherWrite"), wxT("OtherRead"), wxT("GroupExec"), wxT("GroupWrite"), wxT("GroupRead"),
																		wxT("UserExec"), wxT("UserWrite"), wxT("UserRead"), wxT("Sticky"), wxT("Sguid"), wxT("Suid"),
							wxT("PermStatic1"), wxT("PermStatic2"), wxT("PermStatic3"), wxT("PermStatic4"), wxT("PermStatic5"), wxT("PermStatic6"), wxT("PermStatic7") };

wxArrayString filepaths; GetMultiplePaths( filepaths ); if ( filepaths.IsEmpty() ) return;
		// Though there might be multiple selections, focus on the 1st for convenience.  Multiple makes sense only chmod & chown.
FileData* stat = new FileData( filepaths[0] );  if ( ! stat->IsValid() ) { delete stat; return; }


wxDialog dlg;
if ( ! stat->IsSymlink() ) wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("PropertiesDlg"));
 else 
	{	if ( stat->IsSymlinktargetASymlink() )
 			{	wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("UltSymlinkPropertiesDlg"));	// The multiple-symlink version has 2 extra sections
				wxTextCtrl* UltTarget = (wxTextCtrl*)dlg.FindWindow( wxT("UltLinkTargetText"));
				UltTarget->SetValue( TruncatePathtoFitBox( stat->GetUltimateDestination(), UltTarget ) );	// Put the ultimate target filepath into its textctrl
				dlg.Connect( XRCID("UltLinkTargetBtn"), wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction)&MyGenericDirCtrl::OnLinkTargetButton );
			}
		 else  	wxXmlResource::Get()->LoadDialog(&dlg, MyFrame::mainframe, wxT("SymlinkPropertiesDlg"));	// The Symlink version has 1 extra section
	}

PropertyDlg = &dlg;																											// Store the address of dlg, in case OnLinkTargetButton needs it

((wxNotebook*)dlg.FindWindow( wxT("PropertiesNB")))->SetSelection( MyFrame::mainframe->PropertiesPage ); // Set the notebk page to that last used

wxTextCtrl* Name = (wxTextCtrl*)dlg.FindWindow( wxT("Name")); Name->SetValue( stat->GetFilename() );

wxStaticText* Path = (wxStaticText*)dlg.FindWindow( wxT("Path"));
wxString path = filepaths[0].BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH;
Path->SetLabel( TruncatePathtoFitBox( path, Path ) );			// Most paths are too big for the Path static, so check first, truncate if needed

wxTextCtrl* UserText = (wxTextCtrl*)dlg.FindWindow( wxT("UserText"));
wxComboBox* UserCombo = (wxComboBox*)dlg.FindWindow( wxT("UserCombo"));

bool isroot = (getuid()==0);																						// Are we root?
if ( isroot )																													// Only root can change ownership
	{	wxSizer* sizer = UserText->GetContainingSizer();			// There is both a wxTextCtrl & a wxComboBox in this sizer.  If we're root, remove the textctrl
		UserText->Hide();															// For some reason we seem to have to hide it first,
#if wxVERSION_NUMBER <2800
		if ( sizer->Remove( UserText ) ) UserText->Destroy();	//  & then remove it!  And no, Layout() didn't work
#else
		if ( sizer->Detach( UserText ) ) UserText->Destroy();
#endif
		wxArrayString usernames; GetAllUsers( usernames );											//  Use this global function to get list of usernames
		for (size_t n=0; n < usernames.Count(); ++n ) UserCombo->Append( usernames[n] );	// Add them to combobox
		wxString owner = stat->GetOwner(); if ( owner.IsEmpty() ) owner.Printf( wxT("%u"), stat->OwnerID() );
		UserCombo->SetValue( owner );									// Select the current owner;  his name if available, else his id
	}
 else
	{	wxSizer* sizer = UserCombo->GetContainingSizer();		// Similarly if we're not root:  remove the combobox
		UserCombo->Hide();
#if wxVERSION_NUMBER <2800
		if ( sizer->Remove( UserCombo ) ) UserCombo->Destroy();
#else
		if ( sizer->Detach( UserCombo ) ) UserCombo->Destroy();
#endif
		wxString owner = stat->GetOwner(); if ( owner.IsEmpty() ) owner.Printf( wxT("%u"), stat->OwnerID() );
		UserText->SetValue( owner );
	}

wxComboBox* Group = (wxComboBox*)dlg.FindWindow( wxT("Group"));
wxArrayString groupnames; GetProcessesGroupNames( groupnames );						// Use this global function to get list of alternative group-names
for (size_t n=0; n < groupnames.Count(); ++n ) Group->Append( groupnames[n] );	// Add them to combobox
Group->SetValue( stat->GetGroup() );																			// Set combobox to current group


((wxStaticText*)dlg.FindWindow( wxT("Type")))->SetLabel( stat->TypeString() );

((wxStaticText*)dlg.FindWindow( wxT("SizeStatic")))->SetLabel( stat->GetParsedSize() );

wxDateTime time( stat->ModificationTime() ); 
((wxStaticText*)dlg.FindWindow( wxT("ModStatic")))->SetLabel( time.Format( wxT("%d/%m/%y %H:%M")) );

time.Set( stat->AccessTime() ); 
((wxStaticText*)dlg.FindWindow( wxT("AccessStatic")))->SetLabel( time.Format( wxT("%d/%m/%y %H:%M")) );

time.Set( stat->ChangedTime() ); 
((wxStaticText*)dlg.FindWindow( wxT("ChangedStatic")))->SetLabel( time.Format( wxT("%d/%m/%y %H:%M")) );

size_t perms = stat->GetPermissions();																		// Get current permissions in numerical form
for ( int n=0; n < 12; ++n )																							//  & set each checkbox appropriately.  Note the clever-clever use of << to select each bit
	((wxCheckBox*)dlg.FindWindow( MiscPermissionStrings[n] ))->SetValue( perms & (1 << n) );

if ( stat->IsSymlink() )
	{	if ( stat->TypeString() == wxT("Broken Symbolic Link") )									// If a broken symlink, disable the GoToTarget button
			{ wxWindow* win = dlg.FindWindow(wxT("LinkTargetBtn")); if (win) win->Disable(); }
		((wxTextCtrl*)dlg.FindWindow( wxT("LinkTargetText")))->SetValue( stat->GetSymlinkDestination( true ) );	// 'true' says to display relative symlinks as relative
		dlg.Connect( XRCID("LinkTargetBtn"), wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction)&MyGenericDirCtrl::OnLinkTargetButton );
		dlg.Connect( XRCID("Browse"), wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction)&MyGenericDirCtrl::OnBrowseButton );
	}
																		// Now do the Permissions/Ownership page
bool writable = stat->CanTHISUserRename();
if ( ! writable )
	{	Name->Enable(false); Path->Enable(false); }

bool chmodable = stat->CanTHISUserChmod();
if ( ! chmodable )
	{	for ( int n=0; n < 19; ++n )
			((wxCheckBox*)dlg.FindWindow( MiscPermissionStrings[n] ))->Enable(false);
		
		Group->Enable(false);																						// The requirements for changing group are the same as for chmod
	}
			
if ( ! isroot )		UserText->Enable(false);																	// Only root can change ownership

wxCheckBox* Recursebox = (wxCheckBox*)dlg.FindWindow( wxT("RecurseCheck"));
bool queryrecurse=false;
for ( size_t sel=0; sel < filepaths.GetCount(); ++sel )
	{	FileData selstat( filepaths[sel] ); if ( selstat.IsDir() ) queryrecurse = true; }
Recursebox->Show(queryrecurse);																			// Don't show the Apply to Descendants checkbox unless >0 selection's a dir!

																		// Now do the Esoterica page
((wxStaticText*)dlg.FindWindow( wxT("DeviceID")))->SetLabel( wxString::Format( wxT("%d"), (int)stat->GetDeviceID() ) );
((wxStaticText*)dlg.FindWindow( wxT("Inode")))->SetLabel( wxString::Format( wxT("%lu"), stat->GetInodeNo() ) );
((wxStaticText*)dlg.FindWindow( wxT("Hardlinks")))->SetLabel( wxString::Format(wxT("%u"), (uint)stat->GetHardLinkNo() ) );
((wxStaticText*)dlg.FindWindow( wxT("BlockNo")))->SetLabel( wxString::Format(wxT("%lu"), stat->GetBlockNo() ) );
((wxStaticText*)dlg.FindWindow( wxT("Blocksize")))->SetLabel( wxString::Format(wxT("%lu"), stat->GetBlocksize() ) );
((wxStaticText*)dlg.FindWindow( wxT("OwnerID")))->SetLabel( wxString::Format(wxT("%u"), stat->OwnerID() ) );
((wxStaticText*)dlg.FindWindow( wxT("GroupID")))->SetLabel( wxString::Format(wxT("%u"), stat->GroupID() ) );

UnRedoManager::StartCluster();					// Start a new UnRedoManager cluster. Doing it here catches both rename & changesymlinktarget
do
	{	HelpContext = HCproperties;
		if ( dlg.ShowModal() != wxID_OK )		
			{ 	MyFrame::mainframe->PropertiesPage = ((wxNotebook*)dlg.FindWindow( wxT("PropertiesNB") ))->GetSelection();	// Save the notebook pageno
				HelpContext = HCunknown;
				delete stat; UnRedoManager::EndCluster(); return;								// If user cancelled, abort
			}
		HelpContext = HCunknown;
		if ( writable )																										// If user was allowed to change the filename
			{	if ( Name->GetValue() != stat->GetFilename() )											//   & has,
					{	bool ItsADir = stat->IsDir();
						wxString path = Path->GetLabel() + wxFILE_SEP_PATH, oldname = stat->GetFilename(), newname = Name->GetValue();
						int ans = Rename( path, oldname, newname, ItsADir );
						if ( ans == wxID_YES ) continue;														// If there's a filename problem, redo the dialog 
				
						if ( ans )
							{	wxString newstartdir = path + newname;
								bool renamingstartdir = false;
								wxArrayInt IDs; IDs.Add( GetId() );												// Update the panes
								if ( path.BeforeLast(wxFILE_SEP_PATH) == startdir )				// If we've just renamed startdir, need to do things differently
									{	MyFrame::mainframe->OnUpdateTrees( path, IDs, newstartdir ); 	// Newstartdir will be substituted for the old one
										renamingstartdir = true;
									}
								 else MyFrame::mainframe->OnUpdateTrees( path, IDs );
																					// Make a new UnRedoRen
								UnRedoRen* UnRedoptr;
								if ( ItsADir )
									{	path += oldname;
										if ( renamingstartdir )
											UnRedoptr = new UnRedoRen( path, newstartdir, myNullString, newstartdir, ItsADir, IDs );	// If renamingstartdir, use 4th string as notification
										 else
											UnRedoptr = new UnRedoRen( path, newstartdir, myNullString, myNullString, ItsADir, IDs );	// Otherwise make an UnRedoRen the standard way
									}
								 else UnRedoptr = new UnRedoRen( path, path, oldname, newname, ItsADir, IDs );	
								UnRedoManager::AddEntry( UnRedoptr );
																					// Since we've just changed the name, we need to recreate stat to match. If not, there'd be failures below
								delete stat; stat = new FileData( newstartdir );  if ( ! stat->IsValid() ) { delete stat; UnRedoManager::EndCluster(); return; }
							}
					}
			}

		if ( stat->IsSymlink() )
			{	wxString dest = ((wxTextCtrl*)dlg.FindWindow( wxT("LinkTargetText")))->GetValue();
				if ( ! dest.IsEmpty() && dest != stat->GetSymlinkDestination( true ) )
					{	bool absOK = true; wxFileName MakeAbs( dest );
						if ( MakeAbs.IsRelative() )				// FileDatas can't cope with relative names so, temporarily MakeAbsolute
							{ wxString currentfilepath = stat->GetFilepath();  absOK = MakeAbs.MakeAbsolute( currentfilepath ); }
						
						FileData deststat( MakeAbs.GetPath() );
						if ( !absOK || !deststat.IsValid() )															// If the desired new target doesn't exist
							{	wxMessageDialog dialog(this, _("The new target for the link doesn't seem to exist.\nTry again?"), wxT(""),wxYES_NO | wxICON_QUESTION );
								if ( dialog.ShowModal() == wxID_YES ) continue;						// Try again
							}
						 else
							{	wxString currentfilepath = stat->GetFilepath(), oldtarget = stat->GetSymlinkDestination( true ); 
								bool result = CheckDupRen::ChangeLinkTarget( currentfilepath, dest );
								if ( !result ) wxMessageBox(_("For some reason, changing the symlink's target seems to have failed.  Sorry!"));
									else
										{	wxArrayInt IDs; IDs.Add( GetId() );									// Update the panes
											MyFrame::mainframe->OnUpdateTrees( currentfilepath, IDs );

											UnRedoChangeLinkTarget *UnRedoptr = new UnRedoChangeLinkTarget( currentfilepath, IDs, dest, oldtarget );
											UnRedoManager::AddEntry( UnRedoptr );
											
																					// Since we've just deleted/recreated the link, we need to recreate stat to match. If not, there'd be failures below											
											delete stat; stat = new FileData( currentfilepath );  if ( ! stat->IsValid() ) { delete stat; UnRedoManager::EndCluster(); return; }
										}
							}
					}
	}
		break;
	}
		while(true);

bool Recurse = Recursebox->GetValue();																	// (If >0 selection is a dir) are we supposed to implement changes recursively
size_t totalpsuccesses=0, totalpfailures=0, totalosuccesses=0, totalofailures=0, totalgsuccesses=0, totalgfailures=0;
wxArrayInt IDs; IDs.Add( GetId() );

for ( size_t sel=0; sel < filepaths.GetCount(); ++sel )
	{	size_t psuccesses=0, pfailures=0, osuccesses=0, ofailures=0, gsuccesses=0, gfailures=0;
		FileData* stat = new FileData( filepaths[sel] );  if ( ! stat->IsValid() ) { delete stat; continue; }
		
		if ( isroot )																// Only root can change owner
			{	long uid; wxString newuser = UserCombo->GetValue();							// This will usually be a name, but...
				if ( newuser.ToLong( &uid ) )																		//   check if a number was entered e.g. 999 in a *buntu livecd
						RecursivelyChangeAttributes( stat->GetFilepath(), IDs, uid, osuccesses, ofailures, ChangeOwner, Recurse );	// If so, use it
				  else
					{	struct passwd* p = getpwnam( newuser.mb_str(wxConvUTF8) );		// Otherwise find the pw-struct for the selected user-name
						if ( p != NULL )										//  Try to change to this user. Nothing will happen if no change was made
						RecursivelyChangeAttributes( stat->GetFilepath(), IDs, p->pw_uid, osuccesses, ofailures, ChangeOwner, Recurse );
					}
			}

		if ( chmodable )														// Now group & permissions, which have the same Access requirement
			{	long gid; wxString newgroup = Group->GetValue();
				if ( newgroup.ToLong( &gid ) )
						RecursivelyChangeAttributes( stat->GetFilepath(), IDs, gid, gsuccesses, gfailures, ChangeGroup, Recurse );
				  else
					{	struct group* g = getgrnam( newgroup.mb_str(wxConvUTF8) );
						if ( g != NULL )
							RecursivelyChangeAttributes( stat->GetFilepath(), IDs, g->gr_gid, gsuccesses, gfailures, ChangeGroup, Recurse );
					}

				size_t newperms = 0;
				for ( int n=0; n < 12; ++n )																					// Create a new permissions int by <<ing each checkbox appropriately
					newperms |= ( ((wxCheckBox*)dlg.FindWindow( MiscPermissionStrings[n] ))->GetValue() << n );
				RecursivelyChangeAttributes( stat->GetFilepath(), IDs, newperms, psuccesses, pfailures, ChangePermissions, Recurse );	// Try to change permissions, recursively if requested. Nothing will happen if no change was made				
			}
			
		delete stat;
		totalpsuccesses += psuccesses; totalosuccesses += osuccesses; totalgsuccesses += gsuccesses;
		totalpfailures += pfailures; totalofailures += ofailures; totalpfailures += gfailures; 
	}
																		// Sum all possible successes/failures, in case eg both owner & group were changed
size_t successes = totalpsuccesses+totalosuccesses+totalgsuccesses; size_t failures = totalpfailures+totalofailures+totalgfailures;
if ( successes + failures )																							// Only display if something was actually changed (or attempts failed)
	{	wxString msg; msg.Printf( wxT("%u%s%u%s"), (uint)successes, _(" alterations successfully made, "), (uint)failures, _(" failures") );
		BriefLogStatus bls( msg  );
		
		if ( successes )	MyFrame::mainframe->OnUpdateTrees( stat->GetFilepath(), IDs ); // Update the panes if appropriate
	}

UnRedoManager::EndCluster();

int page = ((wxNotebook*)dlg.FindWindow( wxT("PropertiesNB") ))->GetSelection();
MyFrame::mainframe->PropertiesPage = page<0 ? 0 : page;											// Save the notebook pageno
delete stat;
}

void MyGenericDirCtrl::OnLinkTargetButton(wxCommandEvent& event)	// // If symlink-target button clicked in Properties dialog
{
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane();								// This method behaves as though it were static:  we can't use 'this' as the active pane
if ( active != NULL ) active->LinkTargetButton(event);													// So jump to where we can!
}

void MyGenericDirCtrl::LinkTargetButton(wxCommandEvent& event)		// // Recreate tree to goto selected link's target (or ultimate target)
{
if ( PropertyDlg != NULL )  PropertyDlg->EndModal( wxID_OK );									// First shutdown the originating dialog

GoToSymlinkTarget( event.GetId() == XRCID("UltLinkTargetBtn") );								// Use another method to do the GoTo, as it's also used by a context menu entry
}

void MyGenericDirCtrl::GoToSymlinkTarget( bool ultimate )		// // Goto selected link's target (or ultimate target)
{
FileData stat( GetPath() ); if ( ! ( stat.IsValid() && stat.GetSymlinkData()->IsValid() ) ) return;  // Locate the selected link's target (or ultimate target)

wxString SymlinkTarget, storedSymlinkTarget;
if ( ultimate )	 SymlinkTarget = stat.GetUltimateDestination();
 else SymlinkTarget = stat.MakeAbsolute( stat.GetSymlinkDestination(), stat.GetPath() );
 
storedSymlinkTarget = SymlinkTarget; 																		// OnOutsideSelect() alters the string
OnOutsideSelect( SymlinkTarget ); 
SetPath(storedSymlinkTarget);																					// OnOutsideSelect() only does the dir, not the file selection
}

void MyGenericDirCtrl::OnBrowseButton(wxCommandEvent& WXUNUSED(event))		// // If symlink-target Browse button clicked in Properties dialog
{
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane();		// This method behaves as though it were static:  we can't use 'this' as the active pane
if ( active != NULL ) active->BrowseButton();											// So jump to where we can!
}

void MyGenericDirCtrl::BrowseButton()		// //  The Browse button was clicked in Properties dialog
{
if ( PropertyDlg == NULL ) return;

wxString currenttargetpath, currenttargetname;

FileData stat( GetPath() );											// We need to decide with which filepath to load the dialog
if ( stat.GetSymlinkData()->IsValid() )																			// Providing it's valid, use the symlink's current target
	{	if ( stat.GetSymlinkData()->IsDir() )
			{	currenttargetpath = stat.GetSymlinkDestination(); currenttargetname = wxEmptyString; }
		 else		
  			{	currenttargetpath = stat.GetSymlinkDestination().BeforeLast(wxFILE_SEP_PATH);
				currenttargetname = stat.GetSymlinkDestination().AfterLast(wxFILE_SEP_PATH);
			}
	}
 else																															// Otherwise use the symlink itself
 	{	if ( stat.IsDir() )
			{	currenttargetpath = stat.GetFilepath(); currenttargetname = wxEmptyString; }
		 else		
  			{	currenttargetpath = stat.GetFilepath().BeforeLast(wxFILE_SEP_PATH);
				currenttargetname = stat.GetFilepath().AfterLast(wxFILE_SEP_PATH);
			}
	}

FileDirDlg fdlg( PropertyDlg, FD_SHOWHIDDEN, currenttargetpath, _("Choose a different file or dir to be the new target") );
if ( fdlg.ShowModal() != wxID_OK ) return;

wxString newtarget = fdlg.GetFilepath();

((wxTextCtrl*)PropertyDlg->FindWindow( wxT("LinkTargetText")))->SetValue( newtarget );	// Paste into the textctrl
}


BEGIN_EVENT_TABLE(FileGenericDirCtrl,MyGenericDirCtrl)
	EVT_SIZE(FileGenericDirCtrl::OnSize)
	EVT_MENU(SHCUT_NEW, FileGenericDirCtrl::NewFile)
	
	EVT_MENU(SHCUT_ARCHIVE_APPEND, FileGenericDirCtrl::OnArchiveAppend)
	EVT_MENU(SHCUT_ARCHIVE_TEST, FileGenericDirCtrl::OnArchiveTest)
	
	EVT_MENU(SHCUT_TOGGLEHIDDEN, FileGenericDirCtrl::OnToggleHidden)
	EVT_MENU(SHCUT_OPENWITH, FileGenericDirCtrl::OnOpenWith)
	EVT_MENU(SHCUT_OPEN, FileGenericDirCtrl::OnOpen)
	EVT_MENU(SHCUT_OPENWITH_KDESU, FileGenericDirCtrl::OpenWithKdesu)
	EVT_MENU_RANGE(SHCUT_RANGESTART,SHCUT_RANGEFINISH, FileGenericDirCtrl::OpenWithSubmenu)
	EVT_MENU_RANGE(SHCUT_SHOW_COL_EXT, SHCUT_SHOW_COL_CLEAR, FileGenericDirCtrl::OnColumnSelectMenu)
	
	EVT_LIST_COL_CLICK(-1, FileGenericDirCtrl::HeaderWindowClicked)
	EVT_IDLE(FileGenericDirCtrl::OnIdle)	

 	EVT_CONTEXT_MENU(FileGenericDirCtrl::ShowContextMenu)
END_EVENT_TABLE()

#if wxVERSION_NUMBER < 2502
// ----------------------------------------------------------------------------
// check current state of a key  Copied from 2.5.2 source:  it's absent in earlier versions
// ----------------------------------------------------------------------------

#include <wx/app.h>
#include "wx/unix/utilsx11.h"
//#include "wx/image.h"
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <X11/keysym.h>

KeySym wxCharCodeWXToX(int id)
{
    KeySym keySym;

    switch (id)
    {
        case WXK_CANCEL:        keySym = XK_Cancel; break;
        case WXK_BACK:          keySym = XK_BackSpace; break;
        case WXK_TAB:           keySym = XK_Tab; break;
        case WXK_CLEAR:         keySym = XK_Clear; break;
        case WXK_RETURN:        keySym = XK_Return; break;
        case WXK_SHIFT:         keySym = XK_Shift_L; break;
        case WXK_CONTROL:       keySym = XK_Control_L; break;
        case WXK_ALT:           keySym = XK_Meta_L; break;
        case WXK_CAPITAL:          keySym = XK_Caps_Lock; break;
        case WXK_MENU :         keySym = XK_Menu; break;
        case WXK_PAUSE:         keySym = XK_Pause; break;
        case WXK_ESCAPE:        keySym = XK_Escape; break;
        case WXK_SPACE:         keySym = ' '; break;
        case WXK_PAGEUP:        keySym = XK_Prior; break;
        case WXK_PAGEDOWN:      keySym = XK_Next; break;
        case WXK_END:           keySym = XK_End; break;
        case WXK_HOME :         keySym = XK_Home; break;
        case WXK_LEFT :         keySym = XK_Left; break;
        case WXK_UP:            keySym = XK_Up; break;
        case WXK_RIGHT:         keySym = XK_Right; break;
        case WXK_DOWN :         keySym = XK_Down; break;
        case WXK_SELECT:        keySym = XK_Select; break;
        case WXK_PRINT:         keySym = XK_Print; break;
        case WXK_EXECUTE:       keySym = XK_Execute; break;
        case WXK_INSERT:        keySym = XK_Insert; break;
        case WXK_DELETE:        keySym = XK_Delete; break;
        case WXK_HELP :         keySym = XK_Help; break;
        case WXK_NUMPAD0:       keySym = XK_KP_0; break; case WXK_NUMPAD_INSERT:     keySym = XK_KP_Insert; break;
        case WXK_NUMPAD1:       keySym = XK_KP_1; break; case WXK_NUMPAD_END:           keySym = XK_KP_End; break;
        case WXK_NUMPAD2:       keySym = XK_KP_2; break; case WXK_NUMPAD_DOWN:      keySym = XK_KP_Down; break;
        case WXK_NUMPAD3:       keySym = XK_KP_3; break; case WXK_NUMPAD_PAGEDOWN:  keySym = XK_KP_Page_Down; break;
        case WXK_NUMPAD4:       keySym = XK_KP_4; break; case WXK_NUMPAD_LEFT:         keySym = XK_KP_Left; break;
        case WXK_NUMPAD5:       keySym = XK_KP_5; break;
        case WXK_NUMPAD6:       keySym = XK_KP_6; break; case WXK_NUMPAD_RIGHT:       keySym = XK_KP_Right; break;
        case WXK_NUMPAD7:       keySym = XK_KP_7; break; case WXK_NUMPAD_HOME:       keySym = XK_KP_Home; break;
        case WXK_NUMPAD8:       keySym = XK_KP_8; break; case WXK_NUMPAD_UP:             keySym = XK_KP_Up; break;
        case WXK_NUMPAD9:       keySym = XK_KP_9; break; case WXK_NUMPAD_PAGEUP:   keySym = XK_KP_Page_Up; break;
        case WXK_NUMPAD_DECIMAL:    keySym = XK_KP_Decimal; break; case WXK_NUMPAD_DELETE:   keySym = XK_KP_Delete; break;
        case WXK_NUMPAD_MULTIPLY:   keySym = XK_KP_Multiply; break;
        case WXK_NUMPAD_ADD:             keySym = XK_KP_Add; break;
        case WXK_NUMPAD_SUBTRACT: keySym = XK_KP_Subtract; break;
        case WXK_NUMPAD_DIVIDE:        keySym = XK_KP_Divide; break;
        case WXK_NUMPAD_ENTER:        keySym = XK_KP_Enter; break;
        case WXK_NUMPAD_SEPARATOR:  keySym = XK_KP_Separator; break;
        case WXK_F1:            keySym = XK_F1; break;
        case WXK_F2:            keySym = XK_F2; break;
        case WXK_F3:            keySym = XK_F3; break;
        case WXK_F4:            keySym = XK_F4; break;
        case WXK_F5:            keySym = XK_F5; break;
        case WXK_F6:            keySym = XK_F6; break;
        case WXK_F7:            keySym = XK_F7; break;
        case WXK_F8:            keySym = XK_F8; break;
        case WXK_F9:            keySym = XK_F9; break;
        case WXK_F10:           keySym = XK_F10; break;
        case WXK_F11:           keySym = XK_F11; break;
        case WXK_F12:           keySym = XK_F12; break;
        case WXK_F13:           keySym = XK_F13; break;
        case WXK_F14:           keySym = XK_F14; break;
        case WXK_F15:           keySym = XK_F15; break;
        case WXK_F16:           keySym = XK_F16; break;
        case WXK_F17:           keySym = XK_F17; break;
        case WXK_F18:           keySym = XK_F18; break;
        case WXK_F19:           keySym = XK_F19; break;
        case WXK_F20:           keySym = XK_F20; break;
        case WXK_F21:           keySym = XK_F21; break;
        case WXK_F22:           keySym = XK_F22; break;
        case WXK_F23:           keySym = XK_F23; break;
        case WXK_F24:           keySym = XK_F24; break;
        case WXK_NUMLOCK:       keySym = XK_Num_Lock; break;
        case WXK_SCROLL:        keySym = XK_Scroll_Lock; break;
        default:                keySym = id <= 255 ? (KeySym)id : 0;
    }

    return keySym;
}


bool wxGetKeyState(wxKeyCode key)
{
#if defined(__WXX11__)
    Display *pDisplay = (Display*) wxApp::GetDisplay();
#elif defined(__WXGTK__)
    Display *pDisplay = GDK_DISPLAY();
#elif defined(__WXMOTIF__)
    Display *pDisplay = (Display*) (wxTheApp ? wxTheApp->GetInitialDisplay() : NULL);
#else
#error  Add code to get the DISPLAY for this platform
#endif

    int iKey = wxCharCodeWXToX(key);										// Converts e.g. WXK_CANCEL to the key-symbol XK_Cancel
    int          iKeyMask = 0;
    Window       wDummy1, wDummy2;
    int          iDummy3, iDummy4, iDummy5, iDummy6;
    unsigned int iMask;
    KeyCode keyCode = XKeysymToKeycode(pDisplay,iKey);		// Finds the key"code" for the key-symbol: I think this is the position within the map
    if (keyCode == NoSymbol)
        return false;
    if ( IsModifierKey(iKey) )  // If iKey is a modifier key, use a different method
        {	XModifierKeymap* map = XGetModifierMapping(pDisplay);	// Gets a map of the modifier keys
			wxCHECK_MSG( map, false, _T("failed to get X11 modifiers map") );
			for (int i = 0; i < 8; ++i)
                {
                    if ( map->modifiermap[map->max_keypermod * i] == keyCode)
                        {
                            iKeyMask = 1 << i;
                        }
                }

                XQueryPointer(pDisplay, DefaultRootWindow(pDisplay), &wDummy1, &wDummy2,
                                            &iDummy3, &iDummy4, &iDummy5, &iDummy6, &iMask );
                XFreeModifiermap(map);
                return (iMask & iKeyMask) != 0;
        }
    // From the XLib manual:
    // The XQueryKeymap() function returns a bit vector for the logical state of the keyboard, 
    // where each bit set to 1 indicates that the corresponding key is currently pressed down. 
    // The vector is represented as 32 bytes. Byte N (from 0) contains the bits for keys 8N to 8N + 7 
    // with the least-significant bit in the byte representing key 8N.
    char key_vector[32];
    XQueryKeymap(pDisplay, key_vector);
    return key_vector[keyCode >> 3] & (1 << (keyCode & 7)); 
}

#endif




